#include "CircleButton.h"
#include "TimeFunctions.h"
#include <SOIL.h>
#include <iostream>

#define PI (3.14159f)

//how long should we hover a button before it is activated, in milliseconds
long CircleButton::ActivationTimeOut = 1000; 
// global list containing button events
std::deque<std::string> CircleButton::ButtonMessageQueue;


CircleButton::CircleButton(void)
{
	ButtonID = "default";
	loadTexture("../images/default.png");
	XPosition = 0;
	YPosition = 0;
	Radius = 1;
	SqrRadius = Radius * Radius;

	RedChannel = 128;
	GreenChannel = 128;
	BlueChannel = 128;

	CurrentState = B_INACTIVE;
	SetInactiveFlag = false;
	setInvisible();
}

CircleButton::CircleButton(std::string button_id, const char* texturefile, float xpos, float ypos, float radius, unsigned char r, unsigned char g, unsigned char b)
{
	ButtonID = button_id;
	loadTexture(texturefile);
	XPosition = xpos;
	YPosition = ypos;
	Radius = radius;
	SqrRadius = Radius * Radius;

	RedChannel = r;
	GreenChannel = g;
	BlueChannel = b;

	CurrentState = B_INACTIVE;
	SetInactiveFlag = false;
	setInvisible();
}

CircleButton::~CircleButton(void)
{
	glDeleteTextures(1, &BackgroundTextureID);
	glDeleteTextures(1, &SymbolTextureID);
}



void CircleButton::loadTexture(const char* filename)
{
	BackgroundTextureID = SOIL_load_OGL_texture("../images/button_background.png", SOIL_LOAD_AUTO, SOIL_CREATE_NEW_ID, SOIL_FLAG_MIPMAPS);
	SymbolTextureID = SOIL_load_OGL_texture(filename, SOIL_LOAD_AUTO, SOIL_CREATE_NEW_ID, SOIL_FLAG_MIPMAPS);
}


void CircleButton::drawButton()
{
	if(isVisible)
	{
		glPushMatrix();		
		
		glTranslatef(XPosition, YPosition, 0.0f);
		glScalef(Radius, Radius, 1.0);

		unsigned char AlphaChannel = 128; // button is semi-transparent in default state...
		if (CurrentState != B_DEFAULT)
		{
			AlphaChannel = 255; // ... else we will draw it fully opaque
		}

		if(CurrentState == B_HOVERED) // if button is hovered we draw circle around the button that visualizes the elapsed time
		{
			float time_percentage = ((float) (TimeFunctions::getTime() - LastHoverTime)) / ((float) ActivationTimeOut);
			int slices = (int) (time_percentage * 64.0); // we visualize elapsed time in 64 slices of full circle
		
			glBegin(GL_TRIANGLE_FAN);
			glVertex3f(0.0, 0.0, 0.0);
			for (int i = 0; i < slices; i++)
			{
				float current_angle = ((float) i) / ((float) (slices-1)) * time_percentage * 2.0f * PI;
				glColor4f(0.6f, 0.6f, 0.6f, time_percentage); // circle opacity in respect to elapsed time
				glVertex3f( sinf(current_angle) * 1.1f, cosf(current_angle) * 1.1f, 0.0f); // size of circle is 110% of button size
			}
			glEnd();
		}
	
		if(CurrentState == B_CLICKED) // exploding circle after button was clicked
		{
			if ( (TimeFunctions::getTime() - LastHoverTime) < 1.20*ActivationTimeOut)
			{
			
				float time_percentage = ((float) (TimeFunctions::getTime() - LastHoverTime)) / ((float) ActivationTimeOut) - 1.0f;
			
				glBegin(GL_TRIANGLE_FAN);
				glVertex3f(0.0, 0.0, 0.0);		  
				for (int i = 0; i < 64; i++)
				{
					float current_angle =  ((float) i) / 63.0f * 2.0f * PI;
					glColor4f(0.6f, 0.6f, 0.6f, 1.0f - 5*time_percentage);
					glVertex3f( sinf(current_angle) * (1.1f + 3*time_percentage), cosf(current_angle) * (1.1f + 3*time_percentage), 0.0f);
				}
				glEnd();
			}
		}

	
	
		if(CurrentState == B_INACTIVE)  // grey out the button if inactive
		{
			glColor4ub(200, 200, 200, AlphaChannel);
		}
		else
		{
			glColor4ub(RedChannel, GreenChannel, BlueChannel, AlphaChannel);
		}
	
		glEnable(GL_TEXTURE_2D);
		// render button background
		glBindTexture(GL_TEXTURE_2D, BackgroundTextureID);
		glBegin(GL_QUADS);
			glTexCoord2f(0.0, 0.0);		glVertex3f(-1.0, 1.0, 0.0);
			glTexCoord2f(0.0, 1.0);		glVertex3f(-1.0, -1.0, 0.0);
			glTexCoord2f(1.0, 1.0);		glVertex3f(1.0, -1.0, 0.0);
			glTexCoord2f(1.0, 0.0);		glVertex3f(1.0, 1.0, 0.0);
		glEnd();
		// render button symbol
		glColor4ub(255, 255, 255, AlphaChannel);
		glBindTexture(GL_TEXTURE_2D, SymbolTextureID);
		glBegin(GL_QUADS);
			glTexCoord2f(0.0, 0.0);		glVertex3f(-1.0, 1.0, 0.0);
			glTexCoord2f(0.0, 1.0);		glVertex3f(-1.0, -1.0, 0.0);
			glTexCoord2f(1.0, 1.0);		glVertex3f(1.0, -1.0, 0.0);
			glTexCoord2f(1.0, 0.0);		glVertex3f(1.0, 1.0, 0.0);
		glEnd();

	
		glDisable(GL_TEXTURE_2D);
	
		glPopMatrix();
	}
}


bool CircleButton::isHovered(float pointer_x, float pointer_y)
{
	float sqr_dist = (pointer_x - XPosition) * (pointer_x - XPosition) + (pointer_y - YPosition) * (pointer_y - YPosition);
	return (sqr_dist < SqrRadius) ? (true) : (false);
}

	
void CircleButton::updateButtonState(float pointer_x, float pointer_y)
{
	if(SetInactiveFlag)
	{
		// if "set to inactive" flag is set we change the button state to inactive after waiting for the click animation
		if ( (TimeFunctions::getTime() - LastClickTime) > 1.20*ActivationTimeOut)
		{
			CurrentState = B_INACTIVE;
			SetInactiveFlag = false;
		}
	}

	// update the button state
	switch(CurrentState)
	{
	case B_INACTIVE:
		// do nothing if button is inactive
		break;

	case B_DEFAULT:
		// if button is hovered we change from default to hovered state and set the hover time
		if (isHovered(pointer_x, pointer_y))
		{
			LastHoverTime = TimeFunctions::getTime();
			CurrentState = B_HOVERED;
		}
		break;

	case B_HOVERED:
		// if not hovered anymore change state back to default
		if (!isHovered(pointer_x, pointer_y))
		{
			CurrentState = B_DEFAULT;
		}
		else 
		{
			// if hovered long enough we change state to clicked and write the button id to the event list. 
			if ( (TimeFunctions::getTime() - LastHoverTime) > ActivationTimeOut)
			{
				notificate();
				LastClickTime = TimeFunctions::getTime();
				CurrentState = B_CLICKED;
			}
		}
		break;

	case B_CLICKED:
		// in clicked state we change back to default if button is not hovered anymore
		if (!isHovered(pointer_x, pointer_y))
		{
			CurrentState = B_DEFAULT;
		}
		break;
	}
}

void CircleButton::notificate()
{
	ButtonMessageQueue.push_back(ButtonID);
}

void CircleButton::setInactive()
{
	// if button is currently clicked we set the flag (in order to finish a possible animation)
	if(CurrentState == B_CLICKED)
	{
		SetInactiveFlag = true;
	}
	// if button was not clicked we immediately change to inactive state
	else
	{
		CurrentState = B_INACTIVE; 
	}
}
void CircleButton::setActive()
{
	// we always set button state to default and the visibility after activation
	SetInactiveFlag = false;
	CurrentState = B_DEFAULT;
	setVisible();
}

void CircleButton::setPosition(float xpos, float ypos)
{
	XPosition = xpos;
	YPosition = ypos;
}


void CircleButton::setXPosition(float xpos)
{
	XPosition = xpos;
}



void CircleButton::setYPosition(float ypos)
{
	YPosition = ypos;
}


void CircleButton::setRadius(float radius)
{
	Radius = radius;
	SqrRadius = radius * radius;
}

float CircleButton::getXPosition()
{
	return XPosition;
}



float CircleButton::getYPosition()
{
	return YPosition;
}


float CircleButton::getRadius()
{
	return Radius;
}

std::string CircleButton::getButtonID()
{
	return ButtonID;
}

clock_t CircleButton::getLastClickTime()
{
	return LastClickTime;
}

void CircleButton::setColor(unsigned char r, unsigned char g, unsigned char b)
{
	RedChannel = r;
	GreenChannel = g;
	BlueChannel = b;
}


void CircleButton::setVisible()
{
	isVisible = true;
}
void CircleButton::setInvisible()
{
	isVisible = false;
	setInactive();
}


std::string CircleButton::getLastEvent()
{
	std::string ev = ButtonMessageQueue.front();
	ButtonMessageQueue.pop_front();
	return ev;
}