#include "GestureRecognizer.h"
#include "GestureUtility.h"
#include <time.h>
#include "KinectSensor.h"
#include <sstream>
#include <algorithm>
#include <assert.h>


#define INFINITY (10000000.0f)
#define PEARSON_CORRELATION_THRESHOLD (0.5f)
#define PEARSON_CORRELATION_SQUARED_THRESHOLD (0.5f)
#define MODIFIED_PEARSON_THRESHOLD (0.5f)
#define EUCLIDEAN_THRESHOLD (25.0f)
#define MODIFIED_EUCLIDEAN_THRESHOLD (2.4f)
#define DTW_THRESHOLD (35.0f)
#define MODIFIED_DTW_THRESHOLD (30.0f)
#define FM_THRESHOLD (1.0f)

#define DEFAULT_MEASURE (DTW)
#define DEFAULT_POSE_THRESHOLD (20.0f) 
#define DEFAULT_KNN (true)
#define DEFAULT_NR_NEIGHBOURS (3)

#define DEFAULT_LOW_THRESHOLD (4.0f)
#define DEFAULT_HIGH_TRHESHOLD (1.0f)
#define DEFAULT_SIZE_OF_MEAN_FILTER (3)
#define DEFAULT_MAX_LENGTH_OF_GESTURES (150)
#define DEFAULT_MIN_LENGTH_OF_GESTURES (20)
#define DEFAULT_RECOGNITION_WINDOW_SIZE (12)
#define DEFAULT_INACTIVITY_TIME (10)
#define DEFAULT_MINIMUM_RECOGNITION_INTERVAL_LENGTH (8)
#define DEFAULT_DEFAULT_GESTURE_THRESHOLD (28.0f)
#define DEFAULT_RESAMPLING_SIZE (64)

using namespace GestureUtility;

// helper functions
bool compareKNNSet (KNNSet& set1, KNNSet& set2) 
{ 
	return (set1.distance < set2.distance); 
}


int GestureRecognizer::getNextRecInfoID()
{
	RecInfoID = (RecInfoID + 1) % ((int)INFINITY);
	return RecInfoID;
}


int GestureRecognizer::getIndexFromID(int id)
{
	for(unsigned int i = 0; i < RecognitionList.size(); i++)
	{
		if(RecognitionList[i].ID == id)
			return i;
	}
	// if no valid index has been found we return -1
	return -1;
}


void GestureRecognizer::setSimilarityMeasurement(SimilarityMeasureType measure_type, float user_threshold)
{
	SimilarityMeasure = measure_type;
	std::stringstream ss;
	switch(SimilarityMeasure)
	{
	case EUCLIDEAN:
		RecognitionThreshold = EUCLIDEAN_THRESHOLD;
		ss<<"EUCLIDEAN";
		break;
	case PEARSON_CORRELATION:
		RecognitionThreshold = PEARSON_CORRELATION_THRESHOLD;
		ss<<"PEARSON_CORRELATION";
		break;
	case PEARSON_CORRELATION_SQUARED:
		RecognitionThreshold = PEARSON_CORRELATION_SQUARED_THRESHOLD;
		ss<<"PEARSON_CORRELATION_SQUARED";
		break;
	case MODIFIED_EUCLIDEAN:
		RecognitionThreshold = MODIFIED_EUCLIDEAN_THRESHOLD;
		ss<<"MODIFIED_EUCLIDEAN";
		break;
	case MODIFIED_PEARSON:
		RecognitionThreshold = MODIFIED_PEARSON_THRESHOLD;
		ss<<"MODIFIED_PEARSON";
		break;
	case DTW:
		RecognitionThreshold = DTW_THRESHOLD;
		ss<<"DTW";
		break;
	case MODIFIED_DTW:
		RecognitionThreshold = MODIFIED_DTW_THRESHOLD;
		ss<<"MODIFIED_DTW";
		break;

	case FM:
		RecognitionThreshold = FM_THRESHOLD;
		ss<<"FM";
		break;

	case MODIFIED_FM:
		RecognitionThreshold = INFINITY;
		ss<<"MODIFIED_FM";
		break;
	}

	if (user_threshold >= 0)
	{
		RecognitionThreshold = user_threshold;
	}

	//std::cout<<"Measurement set to: "<<ss.str()<<" , RecognitionThreshold set to: "<<RecognitionThreshold<<"\n";
}


void GestureRecognizer::calculateMeanGesture(int gesture_set_index)
{
	Gesture mean_gesture;

	for (unsigned int j = 0; j < TrainedGestureSets[gesture_set_index]->Gestures[0]->getSize(); j++)
	{
		Pose mean;
		for (unsigned int i = 0; i < TrainedGestureSets[gesture_set_index]->Gestures.size(); i++)
		{
			Pose temp = TrainedGestureSets[gesture_set_index]->Gestures[i]->getPose(j);
			mean += temp;
		}
		mean *= 1.0f/((float) TrainedGestureSets[gesture_set_index]->Gestures.size()); 
		mean_gesture.addPose(mean);
	}

	TrainedGestureSets[gesture_set_index]->MeanGesture = mean_gesture;
}


void GestureRecognizer::calculateAllMeanGestures()
{
	for (unsigned int i = 0; i < TrainedGestureSets.size(); i++)
	{
		calculateMeanGesture(i);
	}
}


void GestureRecognizer::setVariances(int index)
{
	// calculate the variances for the gesture set with index i

	// we put all gestures from the set into one big gesture
	Gesture temp;
	
	for (unsigned int j = 0; j < TrainedGestureSets[index]->Gestures.size(); j++)
	{
		for (unsigned int k = 0; k < TrainedGestureSets[index]->Gestures[j]->getSize(); k++)
		{
			temp.addPose(TrainedGestureSets[index]->Gestures[j]->getPose(k));
		}
	}

	// calculate the variance vector
	std::vector<float> variances = GestureUtility::getGestureVariances(temp);
	// assign the variances
	TrainedGestureSets[index]->Variances = variances;
}


void GestureRecognizer::setAllVariances()
{
	// calculate the variances for every set of gestures
	for (unsigned int i = 0; i < TrainedGestureSets.size(); i++)
	{
		setVariances(i);
	}

	// print out all the variances
	std::cout<<"*** VARIANCES FOR ALL GESTURE CLASSES ***\n";
	for (unsigned int i = 0; i < TrainedGestureSets.size(); i++)
	{
		std::cout<<"Set "<<i<<": "<<TrainedGestureSets[i]->Variances[0]<<"\t"<<TrainedGestureSets[i]->Variances[1]<<"\t"<<TrainedGestureSets[i]->Variances[2]<<"\t"<<TrainedGestureSets[i]->Variances[3]<<"\n";
	}
}


void GestureRecognizer::standardizeAllTrainedGestures()
{
	for (unsigned int i = 0; i < TrainedGestureSets.size(); i++)
	{
		for (unsigned int j = 0; j < TrainedGestureSets[i]->Gestures.size(); j++)
		{
			GestureUtility::standardizeGesture(*(TrainedGestureSets[i]->Gestures[j]));
		}
	}
}


float GestureRecognizer::getGestureDistance(Gesture &gest1, Gesture &gest2, int gesture_set_index)
{
	// first we have to distinguish between static and dynamic gestures comparison
	int code = (gest1.getSize() == 1) + 2 * (gest2.getSize() == 1);
	switch (code)
	{
	case 3:
		// if both are static we return their pose distances
		return getPoseDistance(gest1.getFirstPose(), gest2.getFirstPose(), gest1.getGestureType());

	case 1:
	case 2:
		// if one of both gestures is static and the other one is dynamic we abort
		return INFINITY;
	}
	// at this point both gestures have to be dynamic
	
	// we provide a duplicate of gest2 (in case we have to alter its length when using Euclidean or Pearson measurements)
	Gesture copy(gest2);

	switch (SimilarityMeasure)
	{
	case EUCLIDEAN:
		if (gest1.getSize() != gest2.getSize())
		{
			GestureUtility::resamplePoseVector(copy.getPoses(), gest1.getSize());
			return GestureUtility::getEuclideanGestureDistance(gest1, copy);
		}
		else return GestureUtility::getEuclideanGestureDistance(gest1, gest2);

	case PEARSON_CORRELATION:
	case PEARSON_CORRELATION_SQUARED:
	case MODIFIED_PEARSON:
		if (gest1.getSize() != gest2.getSize())
		{
			GestureUtility::resamplePoseVector(copy.getPoses(), gest1.getSize());
			return getPearsonGestureDistance(gest1, copy, gesture_set_index);
		}
		else return getPearsonGestureDistance(gest1, gest2, gesture_set_index);

	case MODIFIED_EUCLIDEAN:
		if (gest1.getSize() != gest2.getSize())
		{
			GestureUtility::resamplePoseVector(copy.getPoses(), gest1.getSize());
			return getModifiedEuclideanGestureDistance(gest1, copy);
		}
		else return getModifiedEuclideanGestureDistance(gest1, gest2);

	case DTW:
		return getDTWGestureDistance(gest1, gest2);

	case MODIFIED_DTW:
		return getModifiedDTWGestureDistance(gest1, gest2);

	case FM:
		return getFMGestureDistance(gest1, gest2);

	case MODIFIED_FM:
		return getModifiedFMGestureDistance(gest2, gesture_set_index);

	default:
		return INFINITY;
	}
}


float GestureRecognizer::getPearsonGestureDistance(Gesture& gest1, Gesture& gest2, int gesture_set_index)
{
	assert(gest1.getSize() == gest2.getSize());

	float pearson_distance = 0.0f;

	// we need to know the dimensions of the correlation vector
	int dimension = GestureUtility::getFloatVectorFromPose(gest1.getPose(0)).size();

	// we calculate the pearson correlation coefficients for every dimension independently
	std::vector<float> sum_of_products;
	sum_of_products.assign(dimension, .0f);

	// get the variances
	std::vector<float> variance1 = GestureUtility::getGestureVariances(gest1);
	std::vector<float> variance2 = GestureUtility::getGestureVariances(gest2);

	// copy the input
	Gesture gest1_copy = gest1;
	Gesture gest2_copy = gest2;

	// center the copies
	GestureUtility::centerGesture(gest1_copy);
	GestureUtility::centerGesture(gest2_copy);

	assert( (variance1.size() == dimension) && (variance2.size() == dimension) );


	for (unsigned int i = 0; i < gest1.getSize(); i ++)
	{
		std::vector<float> temp1 = GestureUtility::getFloatVectorFromPose(gest1_copy.getPose(i));
		std::vector<float> temp2 = GestureUtility::getFloatVectorFromPose(gest2_copy.getPose(i));

		for (int j = 0; j < dimension; j++)
		{
			sum_of_products[j] += (temp1[j] * temp2[j]); // there's no need to subtract the mean since we assume that the data has zero mean 
		}
	}

	// we calculate the pearson distance, the final result is the accumulated distance over all dimensions...
	for (int i = 0; i < dimension; i++)
	{
		// we catch division by zero error
		if ( variance1[i] == 0.0f )
		{
			variance1[i] = 0.0001f;
		}
		if ( variance2[i] == 0.0f )
		{
			variance2[i] = 0.0001f;
		}


		// ...in a manhattan distance manner
		if (SimilarityMeasure == PEARSON_CORRELATION)
		{
			pearson_distance += 1 - sum_of_products[i]/(sqrtf(variance1[i] * variance2[i]));
		}
		// ...in a squared distance manner
		if (SimilarityMeasure == PEARSON_CORRELATION_SQUARED)
		{
			pearson_distance += pow(1 - sum_of_products[i]/(sqrtf(variance1[i] * variance2[i])), 2);
		}
		// ...with variance over all trained data sets
		if (SimilarityMeasure == MODIFIED_PEARSON)
		{
			pearson_distance += 1 - sum_of_products[i]/(sqrtf(variance2[i] * TrainedGestureSets[gesture_set_index]->Variances[i]));
		}
	}
	// we scale along number of data sets
	pearson_distance /= gest1.getSize();
	return pearson_distance;
}


float GestureRecognizer::getModifiedEuclideanGestureDistance(Gesture& gest1, Gesture& gest2)
{
	assert(gest1.getSize() == gest2.getSize());
	
	float euclidean_dist = GestureUtility::getEuclideanGestureDistance(gest1, gest2);

	float complexity_gest1 = GestureUtility::getPathLength(gest1.getPoses()) / ((float) gest1.getSize());
	float complexity_gest2 = GestureUtility::getPathLength(gest2.getPoses()) / ((float) gest2.getSize());

	float correctionfactor = max(complexity_gest1,complexity_gest2) / min(complexity_gest1,complexity_gest2);

	return euclidean_dist * correctionfactor;
}


float GestureRecognizer::getDTWGestureDistance(Gesture &gest1, Gesture &gest2) 
{
	assert(gest1.getSize() != 0);
	assert(gest2.getSize() != 0);

	//clock_t start = TimeFunctions::getTime();
	
	int n = gest1.getSize()-1;
	int m = gest2.getSize()-1;
	int constraint = max(n,m) / 10;
	//constraint = max(constraint, abs(n-m));

	std::vector<std::vector<float>> gamma(n+1, std::vector<float>(m+1, INFINITY));

	gamma[0][0] = 0;
    for (int i = 1; i <= n; ++i) 
	{
        for(int j = max(1, i - constraint); j <= min(m, i + constraint); ++j) 
		{
			gamma[i][j] = GestureUtility::getPoseDistance(gest1.getPose(i), gest2.getPose(j)) + min(min(gamma[i-1][j],gamma[i][j-1]),gamma[i-1][j-1]);      
		}
    }

	// we scale along number of data sets
	//std::cout<<"DTW calculation time: "<<TimeFunctions::getTime() - start<<"\n";
	return (gamma[n][m])/((float) min(n+1, m+1));
}


float GestureRecognizer::getModifiedDTWGestureDistance(Gesture &gest1, Gesture &gest2) 
{
	float dtw_dist = getDTWGestureDistance(gest1, gest2);

	float complexity_gest1 = GestureUtility::getPathLength(gest1.getPoses()) / ((float) gest1.getSize());
	float complexity_gest2 = GestureUtility::getPathLength(gest2.getPoses()) / ((float) gest2.getSize());

	float correctionfactor = max(complexity_gest1,complexity_gest2) / min(complexity_gest1,complexity_gest2);

	return dtw_dist * correctionfactor;
}


// returns the energy of every angular time serie as a vector, used in FMDistance Measure
std::vector<float> getFMEnergyVector(Gesture& gest)
{
	std::vector<float> energy(8,0.0f);
	float size = (float) gest.getSize();
	for (unsigned int i = 1; i < gest.getSize(); i++)
	{
		Pose pose1 = gest.getPose(i);
		Pose pose2 = gest.getPose(i-1);
		std::vector<float> temp1 = GestureUtility::getFloatVectorFromPose(pose1);
		std::vector<float> temp2 = GestureUtility::getFloatVectorFromPose(pose2);

		for(int i = 0; i < 8; i++)
		{
			float v = temp1[i] - temp2[i];
			energy[i] += v*v; 
		}
	}

	// calculate average kinetic energy and logarithmize
	for(int i = 0; i < 8; i++)
	{
		energy[i] /= size; 
		energy[i] = log(energy[i] + 1);
	}

	return energy;
}


float GestureRecognizer::getFMGestureDistance(Gesture& gest1, Gesture& gest2)
{
	//kinetic energy for 1st gesture
	std::vector<float>e1 = getFMEnergyVector(gest1);
	//kinetic energy for 2nd gesture
	std::vector<float>e2 = getFMEnergyVector(gest2);

	// calculate euclidean distance between energy1 and energy2
	float dist = 0.0f;
	for(int i=0; i<8; i++)
	{
		dist += (e1[i] - e2[i]) * (e1[i] - e2[i]); 
	}

	return sqrtf(dist);
}


// some helper functions used in modified fm distance algorithm
std::vector<float> VectorAdd(std::vector<float> v1,std::vector<float> v2)
{
	assert(v1.size() == v2.size());

	std::vector<float> result(v1.size(),0.0f);

	for (unsigned int i = 0; i < v1.size(); i++)
	{
		result[i] = v1[i] + v2[i];
	}

	return result;
}


std::vector<float> VectorSub(std::vector<float> v1,std::vector<float> v2)
{
	assert(v1.size() == v2.size());

	std::vector<float> result(v1.size(),0.0f);

	for (unsigned int i = 0; i < v1.size(); i++)
	{
		result[i] = v1[i] - v2[i];
	}

	return result;
}


std::vector<float> VectorMultScalar(std::vector<float> v, float s)
{
	std::vector<float> result(v.size(),0.0f);

	for (unsigned int i = 0; i < v.size(); i++)
	{
		result[i] = v[i] * s;
	}

	return result;
}


// kernel density estimation
float KDE(float diff, std::vector<float> other_diffs)
{
	int size = other_diffs.size();
	float result = 0;
	float bandwith = 1.0f / pow((float) size, 0.25f); // we need a bandwith estimator here!

	// given a distance we calculate the probability according to the distribution established from the other distances by kernel density estimation with cauchy kernel
	for (int i = 0; i < size; i++)
	{
		float t = (diff - other_diffs[i]) / bandwith;
		result += 1.0f / (3.1415926f * ( 1 + t*t) );
	}
	result *= 1.0f / (bandwith * (float) size); 

	return result;
}


// vector version of kernel density estimation for modified FMDistance measure
std::vector<float> VectorKDE(std::vector<float> diff_vec, std::vector<std::vector<float>> other_diffs_vec)
{
	assert( diff_vec.size() == 8 );
	assert( other_diffs_vec.size() > 0);
	assert( other_diffs_vec[0].size() == 8);

	std::vector<float> result;
	// given a vector of distances we calculate the probability of a single element in this vector according to the distribution established from the elements given from the other vector of distance vectors
	for (int i = 0; i < 8; i ++)
	{
		result.push_back(KDE(diff_vec[i],other_diffs_vec[i]));
	}

	return result;
}


// modification of FMDistance measure with probability model
float GestureRecognizer::getModifiedFMGestureDistance(Gesture& gest, int gesture_set_index, int exclude_gest) // actually not used!
{
	// calculate mean energy vector from gesture set
	std::vector<float> mean_energy_of_gest_set(8,0.0f);

	int size = TrainedGestureSets[gesture_set_index]->Gestures.size();
	int count = 0;
	for (int i = 0; i < size; i++)
	{
		// we exlude a given gesture with index exlude_gest
		if ( i != exclude_gest )
		{
			mean_energy_of_gest_set = VectorAdd( mean_energy_of_gest_set, getFMEnergyVector( *(TrainedGestureSets[gesture_set_index]->Gestures[i]) ) );
			count++;
		}
	}

	mean_energy_of_gest_set = VectorMultScalar(mean_energy_of_gest_set, 1.0f / ((float) count));


	// calculate distance energy vector to mean
	std::vector<float> energy_diff_gest_to_mean = VectorSub( getFMEnergyVector(gest), mean_energy_of_gest_set);

	// calculate energy differences of all gestures in same class 
	std::vector<std::vector<float>> energy_diffs_same_class;
	for (int i = 0; i < size; i++)
	{
		// we exlude a given gesture exlude_gest
		if ( i != exclude_gest )
		{
			energy_diffs_same_class.push_back(VectorSub( getFMEnergyVector( *(TrainedGestureSets[gesture_set_index]->Gestures[i]) ), mean_energy_of_gest_set ) );
		}
	}

	// calculate energy differences of all gestures from other classes 
	std::vector<std::vector<float>> energy_diffs_other_classes;
	for (unsigned int nr_of_set = 0; nr_of_set < TrainedGestureSets.size(); nr_of_set++)
	{
		for (unsigned int nr_of_gest = 0; nr_of_gest < TrainedGestureSets[nr_of_set]->Gestures.size(); nr_of_gest++)
		{
			// collect energy difference if there class index is not the given class index
			if ( nr_of_set != gesture_set_index )
			{
				energy_diffs_other_classes.push_back( VectorSub( getFMEnergyVector( *(TrainedGestureSets[nr_of_set]->Gestures[nr_of_gest]) ), mean_energy_of_gest_set ) );
			}
		}
	}
	

	// calculate probability vector p of being in class gesture_set_index ( calculate distribution with KDE)
	std::vector<float> p = VectorKDE(energy_diff_gest_to_mean, energy_diffs_same_class);

	// calculate probability vector q of not being in class gesture_set_index ( calculate distribution with KDE)
	std::vector<float> q = VectorKDE(energy_diff_gest_to_mean, energy_diffs_other_classes);

	// calculate score for this class
	float score = 0.0f;

	assert( p.size() == 8);
	assert( q.size() == 8);

	for (int i = 0; i < 8; i++)
	{
		score += 2 * log(p[i]+1) - log(q[i]+1);
	}

	return 1.0f / score;
}


GestureRecognizer::GestureRecognizer(KinectSensor* sensor)
{
	NuiSensor = sensor;
	
	initialize(DEFAULT_LOW_THRESHOLD, DEFAULT_HIGH_TRHESHOLD, DEFAULT_POSE_THRESHOLD, DEFAULT_SIZE_OF_MEAN_FILTER, DEFAULT_MAX_LENGTH_OF_GESTURES, 
				  DEFAULT_MIN_LENGTH_OF_GESTURES, DEFAULT_RECOGNITION_WINDOW_SIZE, DEFAULT_INACTIVITY_TIME, DEFAULT_MINIMUM_RECOGNITION_INTERVAL_LENGTH, 
					DEFAULT_DEFAULT_GESTURE_THRESHOLD, DEFAULT_RESAMPLING_SIZE, DEFAULT_MEASURE);

	IS_RECOGNITION_THREAD_RUNNING = false;
	UseKNNAlgorithm = DEFAULT_KNN;
	NrOfNeighbours = DEFAULT_NR_NEIGHBOURS;
}


GestureRecognizer::GestureRecognizer(KinectSensor* sensor, bool use_knn, unsigned int nr_of_neighbours)
{
	NuiSensor = sensor;
	
	initialize(DEFAULT_LOW_THRESHOLD, DEFAULT_HIGH_TRHESHOLD, DEFAULT_POSE_THRESHOLD, DEFAULT_SIZE_OF_MEAN_FILTER, DEFAULT_MAX_LENGTH_OF_GESTURES, 
				  DEFAULT_MIN_LENGTH_OF_GESTURES, DEFAULT_RECOGNITION_WINDOW_SIZE, DEFAULT_INACTIVITY_TIME, DEFAULT_MINIMUM_RECOGNITION_INTERVAL_LENGTH, 
					DEFAULT_DEFAULT_GESTURE_THRESHOLD, DEFAULT_RESAMPLING_SIZE, DEFAULT_MEASURE);
	
	IS_RECOGNITION_THREAD_RUNNING = false;
	UseKNNAlgorithm = use_knn;
	NrOfNeighbours = nr_of_neighbours;
}

 
GestureRecognizer::~GestureRecognizer(void)
{
}


void GestureRecognizer::initialize(float velocity_low_threshold, float velocity_high_threshold, float pose_threshold, int size_of_mean_filter, int max_length_of_gestures, int min_length_of_gestures, unsigned int recognition_window_size,  int inactivity_time, int min_recognition_interval_length, float default_gesture_threshold,  int resampling_size, SimilarityMeasureType measure)
{
	TrainingMode = false;
	ReplaceGesture = false;
	NewGestureSet = false;

	RecInfoID = 0;
	LastRecognitionTime.time = 0;
	LastRecognitionTime.GestureSetID = -1;

	setSimilarityMeasurement(measure);

	//initialize thresholds
	VelocityLowThreshold = velocity_low_threshold;
	VelocityHighThreshold = velocity_high_threshold;
	PoseThreshold = pose_threshold;
	DefaultGestureDistanceThreshold = default_gesture_threshold;
	MaxLengthOfGestures = max_length_of_gestures;
	RecognitionWindowSize = recognition_window_size;
	ResamplingSize = resampling_size;
	MinLengthOfGestures = min_length_of_gestures;

	// initialize dynamic vector of last poses
	LastPoses.clear();
	SizeOfMeanFilter = size_of_mean_filter;

	for(int i = 0; i < SizeOfMeanFilter; i++)
	{
		Pose temp;
		LastPoses.push_back(temp);
	}

	// fill recognition list with dummy data
	RecognitionList.clear();
	for (int i=0; i < MaxLengthOfGestures; i++)
	{
		RecognitionInfo info;
		info.flag = false;
		info.velocity = 0.0f;
		RecognitionList.push_back(info);
	}

	InactivityTime = inactivity_time;
	MinimumRecognitionIntervalLength = min_recognition_interval_length;
	CurrentTrainedGestureSetIndex = 0;
	CurrentTrainedGestureIndex = 0;

	DefaultGesture.clear();
	Pose pose(155, 155, 15, 15, 0, 0, 0, 0, 0, 0, 0, 0, math::Vec3f(0,0,0), math::Vec3f(0,0,0));
	DefaultGesture.addPose(pose);
	GestureUtility::resamplePoseVector(DefaultGesture.getPoses(), 16);
	//standardizeGesture(DefaultGesture);

	// you can manually set the index of a gesture (= IGNORE_INDEX) in a specific gesture set (=IGNORE_SET) to be ignored in the recognition stage
	// used for cross validation of training data. in this case we do not want to match a gesture to itself 
	IGNORE_INDEX = -1;
	IGNORE_SET = -1;
}


void GestureRecognizer::setAllPosesFlags(bool value)
{
	for (unsigned int i = 0; i <	RecognitionList.size(); i++)
	{
		RecognitionList[i].flag = value;
	}
}


void GestureRecognizer::setMinDistances()
{
	// we iterate over all gesture sets in the list and check the mutual distances between every two of them
	unsigned int size = TrainedGestureSets.size();
	if(size > 1)
	{
		for (unsigned int i = 0; i < size; i++)
		{
			for (unsigned int j = i+1; j < size; j++)
			{
				// we only check against gestures with the same gesture type
				if(TrainedGestureSets[i]->MeanGesture.getGestureType() == TrainedGestureSets[j]->MeanGesture.getGestureType())
				{
					float dist1 = getGestureDistance(TrainedGestureSets[i]->MeanGesture, TrainedGestureSets[j]->MeanGesture, i);
					float dist2 = getGestureDistance(TrainedGestureSets[i]->MeanGesture, TrainedGestureSets[j]->MeanGesture, j);

					TrainedGestureSets[i]->MinDistance = (dist1 < TrainedGestureSets[i]->MinDistance) ? (dist1) : (TrainedGestureSets[i]->MinDistance);
					TrainedGestureSets[j]->MinDistance = (dist2 < TrainedGestureSets[j]->MinDistance) ? (dist2) : (TrainedGestureSets[j]->MinDistance);	
				}
			}
		}
	}
}


// prepares a temporarily gesture in the gesture spotting procedure
KNNClassification GestureRecognizer::recognizeInterval(RecognitionInterval& interval)
{
		// the KNNClassification will hold our best matching gesture class
		KNNClassification winner;
		winner.GestureSetID = -1;
		winner.points = -1.0f;

		// create a gesture
		Gesture* temp = new Gesture();

		// we need the current start frame in the recognition list of the given interval
MUTEX.lock();
		int start_index = getIndexFromID(interval.id_start);

		if(start_index == -1)
		{
MUTEX.unlock();	
			return winner;
		}
		else
		{
			//fill the vector																		
			for (int i = start_index; i < start_index + interval.size; i++)
			{
				temp->addPose(RecognitionList[i].pose);
			}	

			// resample if the gesture is too long
			if(temp->getSize() > ResamplingSize)
			{
				GestureUtility::resamplePoseVector(temp->getPoses(), ResamplingSize);
		
				if(temp->getSize() != ResamplingSize)
				{
					//std::cout<<"Incorrect resampling!\n";
MUTEX.unlock();	
					return winner;
				}
			}

			// prepare animation
			SkeletonAnimation* animation = new SkeletonAnimation();

			for (int i = start_index; i < start_index + interval.size; i++)
			{
				animation->addFrame(RecognitionList[i].Skeleton);
			}
	
			// assign animation
			temp->setAnimation(animation);
		}
MUTEX.unlock();	
		

		//******** check out the type of gesture, we distinguish between LEFT, RIGHT and BOTH **********
		std::vector<float> variances = GestureUtility::getGestureVariances(*temp);
		
		// if we find low variances in the vector we assume non-motion in that particular vector element (thus a body part was not moved)
		// if all of the variances for a body side are very low we say that this side was not used (and we zero all the values)

		// we check the left side of body
		bool left_is_still = true;

		// the first 4 elements in the vector contain the variances for the left side of the body
		for(int i = 0; i < 4; i++)
		{
			if(variances[i] > 20.0f)
			{
				left_is_still = false;
				break;
			}
		}
		
		// we check the right side of body
		bool right_is_still = true;

		// the last 4 elements in the vector contain the variances for the right side of the body
		for(int i = 4; i < 8; i++)
		{
			if(variances[i] > 20.0f)
			{
				right_is_still = false;
				break;
			}
		}

		// since we know the non-motion of left and right we can distinguish
		// a) left and right side of body had no motion -> we have recorded a static gesture. then we check against the default gesture
		// b) only one side had no motion -> we have recorded a dynamic gesture. then we will center the gesture

		//case a)
		if(left_is_still && right_is_still)
		{
			// we reduce the gesture size to one due to performance
			Pose mean = GestureUtility::getMeanPose(*temp);
			temp->clear();
			temp->addPose(mean);

			// now we check if one or both sides of the body correspond to the default (static) gesture
			int code = (int) (GestureUtility::getPoseDistance(temp->getFirstPose(), DefaultGesture.getFirstPose(), LEFT) < PoseThreshold);  // left side of body is default
			code += 2 * (int) (GestureUtility::getPoseDistance(temp->getFirstPose(), DefaultGesture.getFirstPose(), RIGHT) < PoseThreshold); // right side of body is default

			switch(code)
			{
			case 0:
				// none of the body sides corresponds to the default gesture. hence we assume a both sided (static) gesture
				temp->setGestureType(BOTH);
				break;

			case 1:
				// only the left side of the body corresponds to the default gesture. so we have a right sided gesture
				temp->setGestureType(RIGHT);
				GestureUtility::zeroLeftSideOfBody(*temp);
				break;

			case 2:
				// only the right side of the body corresponds to the default gesture. so we have a left sided gesture
				temp->setGestureType(LEFT);
				GestureUtility::zeroRightSideOfBody(*temp);
				break;
			case 3:
				// both sides of the body correspond to the default gesture. we immediately return the default gesture id
				winner.GestureSetID = -2;
				winner.points = INFINITY;
				return winner;
			}
		}
		else
		{
			//case b)
			//centerGesture(*temp);
			temp->setGestureType(BOTH);
			if(left_is_still)
			{
				GestureUtility::zeroLeftSideOfBody(*temp);
				temp->setGestureType(RIGHT);
			}
			if(right_is_still)
			{
				GestureUtility::zeroRightSideOfBody(*temp);
				temp->setGestureType(LEFT);
			}
		}
		
		//GestureType type = temp->getGestureType(); std::cout<<GestureTypeToString(type)<<"\n";

		// if we do not want to train a gesture...
		if(!TrainingMode)
		{
			// ... we check the newly created gesture against all the trained gestures
			//std::cout<<TimeFunctions::getTime()<<" Trying to validate a gesture!\n";
			recognizeGesture(*temp, winner);
			// free up memory
			delete temp;
		}	

		// in training mode we simply save the gesture and the corresponding skeleton animation
		else 
		{
			bool success = true;
			// if we don't use KNN or (mod.)DTW/(mod.)FM all the trained data must have the same length
			bool resample = (SimilarityMeasure == EUCLIDEAN) ||
							(SimilarityMeasure == PEARSON_CORRELATION) ||
							(SimilarityMeasure == PEARSON_CORRELATION_SQUARED) ||
							(SimilarityMeasure == MODIFIED_PEARSON) ||
							(SimilarityMeasure == MODIFIED_EUCLIDEAN) ||
							(!UseKNNAlgorithm);
			if( resample ) 
			{
				GestureUtility::resamplePoseVector(temp->getPoses(), ResamplingSize);

				if (temp->getSize() != ResamplingSize)
				{
					success = false;
					std::cout<<"Resampling error!\n";
				}
			}

			// if the user wants to create a new set
			if (NewGestureSet)
			{
				// we instantiate a new set
				GestureSet* set = new GestureSet();
				// and push back the gesture
				set->Gestures.push_back(temp);
				set->MinDistance = INFINITY;
				set->SideOfBody = temp->getGestureType();
				// append the new set
				TrainedGestureSets.push_back(set);
				// we set the indices to our newly created gesture
				CurrentTrainedGestureSetIndex = TrainedGestureSets.size()-1;
				CurrentTrainedGestureIndex = 0;
MUTEX.lock();
				GestureMessageQueue.push_back("NEW_GESTURE_SET");
				GestureMessageQueue.push_back("GESTURE_TRAINED");
MUTEX.unlock();
				std::cout<<"Successfully created a new "<<GestureTypeToString(set->SideOfBody)<<" gesture set!\n";
			}
			else
			{
				
				// we push the gesture into the list of gestures of the current set
				
				// ...we check if the new gesture is from the same type as the current set
				if(temp->getGestureType() == TrainedGestureSets[CurrentTrainedGestureSetIndex]->SideOfBody)
				{
					// if the replace flag is true...
					if (ReplaceGesture)
					{
						{
							//...we update the current gesture
							delete TrainedGestureSets[CurrentTrainedGestureSetIndex]->Gestures[CurrentTrainedGestureIndex];
							TrainedGestureSets[CurrentTrainedGestureSetIndex]->Gestures[CurrentTrainedGestureIndex] = temp;
							std::cout<<"Successfully replaced a gesture!\n";
						}	
					
					}
					// else we push back the newly created gesture at the end of the gesture stack
					else
					{
						TrainedGestureSets[CurrentTrainedGestureSetIndex]->Gestures.push_back(temp);
						// set the current index to our new gesture
						CurrentTrainedGestureIndex = TrainedGestureSets[CurrentTrainedGestureSetIndex]->Gestures.size() - 1;	
						std::cout<<"Successfully trained a gesture!\n";
					}
MUTEX.lock();
					GestureMessageQueue.push_back("GESTURE_TRAINED");
MUTEX.unlock();	
				}
				else
				{
					success = false;
MUTEX.lock();
					GestureMessageQueue.push_back("GESTURE_NOT_TRAINED");
MUTEX.unlock();
				}
			}
		
			if (success)
			{
				// we recalculate the mean gesture of the altered gesture set
				if(!UseKNNAlgorithm)
					calculateMeanGesture(CurrentTrainedGestureSetIndex);

				// then we set the mutual minimal distances between the sets   <--- do we still need that?
				//setMinDistances();
			

				// calculate the variances of the modified gesture set
				if ( (SimilarityMeasure == MODIFIED_PEARSON) )
				{
					setVariances(CurrentTrainedGestureSetIndex);
				}
			}
			
			TrainingMode = false;
		}

		return winner;
}


bool GestureRecognizer::isDefaultGesture(Gesture& gest)
{
	if(gest.getSize() != DefaultGesture.getSize())
	{
		// if the gestures sizes don't match we resample a copy of the vector
		Gesture copy(gest);
		GestureUtility::resamplePoseVector(copy.getPoses(),DefaultGesture.getSize());
		return ( GestureUtility::getEuclideanGestureDistance(copy, DefaultGesture) < DefaultGestureDistanceThreshold);
	}

	return ( GestureUtility::getEuclideanGestureDistance(gest, DefaultGesture) < DefaultGestureDistanceThreshold);
}


// classifies a temporarily gesture from the sliding window
void GestureRecognizer::recognizeGesture(Gesture& gesture, KNNClassification &winner)
{
	float scaling = 1.0f;
	if(gesture.getGestureType()==BOTH)
	{
		//if we have a BOTH gesture we raise the threshold
		scaling *= 1.5;
	}


	bool gest_is_static = gesture.getSize() == 1;

	// if the use_means flag is set we only check against the means of the classes
	if(!UseKNNAlgorithm)
	{
		float best_distance = INFINITY;
		int best_id = -1;
		
		for(unsigned int i = 0; i < TrainedGestureSets.size(); i++)
		{
			// only check against training data with same type
			if(gesture.getGestureType() == TrainedGestureSets[i]->SideOfBody)
			{
				// speed-up: for dynamic gesture we first compare start and end poses. if they don't match, we do no further calculation
				//if(!gest_is_static)
				//{
				//	float distance_start = GestureUtility::getPoseDistance(gesture.getFirstPose(), TrainedGestureSets[i]->MeanGesture.getFirstPose());
				//	float distance_end = GestureUtility::getPoseDistance(gesture.getLastPose(), TrainedGestureSets[i]->MeanGesture.getLastPose());
				//	if( (distance_start > 3 * PoseThreshold) || (distance_end > 3 * PoseThreshold)) //<- we relax the threshold a little
				//	{
				//		continue;
				//	}
				//}

				// we set the distance
				float current_distance = getGestureDistance(TrainedGestureSets[i]->MeanGesture, gesture, i);

				if ( (gest_is_static && (current_distance < scaling * PoseThreshold))  || (!gest_is_static && (current_distance < scaling * RecognitionThreshold)) )
				{
					if (current_distance < best_distance)
					{
						best_distance = current_distance;
						best_id = i;
					}
				}
			}
		}

		winner.GestureSetID = best_id;
		winner.points = 1.0f/(best_distance);

		return;
	}

	// if we use the modified fm distance metric
	if(SimilarityMeasure == MODIFIED_FM)
	{
		float best_distance = INFINITY;
		int best_id = -1;
		
		for(unsigned int i = 0; i < TrainedGestureSets.size(); i++)
		{
			// only check against training data with same type
			if(gesture.getGestureType() == TrainedGestureSets[i]->SideOfBody)
			{
				// speed-up: for dynamic gesture we first compare start and end poses. if they don't match, we do no further calculation
				//if(!gest_is_static)
				//{
				//	float distance_start = GestureUtility::getPoseDistance(gesture.getFirstPose(), TrainedGestureSets[i]->MeanGesture.getFirstPose());
				//	float distance_end = GestureUtility::getPoseDistance(gesture.getLastPose(), TrainedGestureSets[i]->MeanGesture.getLastPose());
				//	if( (distance_start > 3 * PoseThreshold) || (distance_end > 3 * PoseThreshold)) //<- we relax the threshold a little
				//	{
				//		continue;
				//	}
				//}

				// we set the distance
				
				float current_distance = getGestureDistance(TrainedGestureSets[i]->MeanGesture, gesture, i);

				if ( (gest_is_static && (current_distance < scaling * PoseThreshold)) || !gest_is_static )
				{
					if (current_distance < best_distance)
					{
						best_distance = best_distance;
						best_id = i;
					}
				}
			}
		}

		winner.GestureSetID = best_id;
		winner.points = 1.0f/(best_distance);

		return;
	}

	// standard k-Nearest-Neighbour algorithm
	// we validate a new gesture against all the already trained gestures
	if(TrainedGestureSets.size() > 0)
	{
		// we insert all valid gestures into a flat vector
		std::vector<KNNSet> valid_gestures;

		for(unsigned int i = 0; i < TrainedGestureSets.size(); i++)
		{
			// we only check against training data with same type
			if(gesture.getGestureType() == TrainedGestureSets[i]->SideOfBody)
			{
				for(unsigned int j = 0; j < TrainedGestureSets[i]->Gestures.size(); j++)
				{
					if( (j==IGNORE_INDEX) && (i==IGNORE_SET) ) continue;

						KNNSet new_set;
						new_set.ID = i;

						// speed-up: for dynamic gesture we first compare start and end poses. if they don't match, we do no further calculation
						//if(!gest_is_static)
						//{
						//	float distance_start = GestureUtility::getPoseDistance(gesture.getFirstPose(), TrainedGestureSets[i]->Gestures[j]->getFirstPose());
						//	float distance_end = GestureUtility::getPoseDistance(gesture.getLastPose(), TrainedGestureSets[i]->Gestures[j]->getLastPose());
						//	if( (distance_start > 3 * PoseThreshold) || (distance_end > 3 * PoseThreshold)) //<- we relax the threshold a little
						//	{
						//		continue;
						//	}
						//}
					
						// we set the distance of the KNN set
						new_set.distance = getGestureDistance(*(TrainedGestureSets[i]->Gestures[j]), gesture, i);
						
						

						// if distance lies below the threshold we push it on the stack
						if ( (gest_is_static && (new_set.distance < scaling * PoseThreshold))  || (!gest_is_static && (new_set.distance < scaling * RecognitionThreshold)) )
						{
							valid_gestures.push_back(new_set);
						}
				}
			}
			
		}

		// now, if there are at least "number_of_neighbours" gestures trained we will start the algorithm
		if ( valid_gestures.size() >= NrOfNeighbours)
		{
			// we sort the vector
			std::sort(valid_gestures.begin(), valid_gestures.end(), compareKNNSet);
			
			// the first "number_of_neighbours" neighbours are the nearest ones, we categorize them into classes and give points depending on their distances ( weight = 1/distance)
			std::vector<KNNClassification> knn_classes;
			bool class_already_exists;
		
			for (unsigned int i = 0; i < NrOfNeighbours; i++)
			{	
				class_already_exists = false;
				// ... we have to check the current set ID against all the others on the classes stack
				for (unsigned int j = 0; j < knn_classes.size(); j++)
				{
					// if we find the same class...
					if (valid_gestures[i].ID == knn_classes[j].GestureSetID)
					{
						//...we add the weighted distance points
						knn_classes[j].points += 1.0f / (valid_gestures[i].distance);	
						class_already_exists = true;
						break;
					}
				}
				// if class does not exist we generate a new one and push it on the classes stack
				if(!class_already_exists)
				{
					KNNClassification temp;
					temp.GestureSetID = valid_gestures[i].ID;
					temp.points = 1.0f / (valid_gestures[i].distance);
					knn_classes.push_back(temp);
				}
			}

			// now we have a number of classes. let's find the one with the most points
			for (unsigned int i = 0; i < knn_classes.size(); i++)
			{
				if(knn_classes[i].points > winner.points)
				{
					winner.GestureSetID = knn_classes[i].GestureSetID;
					winner.points = knn_classes[i].points;
				}
			}
		}
		else
		{
			//std::cout<<"Gesture not recognized!\n";
		}
	}
}


Pose GestureRecognizer::update(bool spot_gesture)
{
	NUI_SKELETON_DATA skeleton = NuiSensor->getCurrentSkeletonData();
	return update(skeleton, spot_gesture);
}


Pose GestureRecognizer::update(NUI_SKELETON_DATA& skeleton, bool spot_gesture)
{
	// set up a pose from the current skeleton input
	Pose pose;
	GestureUtility::setPoseFromSkeletonData(skeleton, pose);

	// push the pose on the stack
	LastPoses.push_back(pose);
	// pop the oldest stack element
	LastPoses.pop_front();

	// we smooth the current pose
	Pose current_pose;
	for (unsigned int i = 0; i < LastPoses.size(); i++)
	{
		current_pose += LastPoses[i];
	}
	current_pose *= 1.0f/((float) LastPoses.size());


	//std::cout<<GestureUtility::getPoseDistance(current_pose, DefaultGesture.getFirstPose())<<"        ";
	//std::cout<<GestureUtility::getPoseDistance(current_pose, DefaultGesture.getFirstPose(), LEFT)<<"\n";


	//velocity calculation
	float last_velocity = RecognitionList.back().velocity;
	float cur_velo = max(0,GestureUtility::getPoseDistance(current_pose, LastPose));
	float velocity = 0.5f * cur_velo + 0.5f * last_velocity;

	// only used for data mining
	CURRENT_VELOCITY = velocity;


	// save the current_pose for the next time step
	LastPose = current_pose;

	RecognitionInfo info;
	info.pose = current_pose;
	info.Skeleton = skeleton;
	info.flag = false;
	info.velocity = velocity;
	info.ID  = getNextRecInfoID();


MUTEX.lock();
	if (!spot_gesture)
	{
		RecognitionList.back().flag = true;
		// in training mode we return immediately else we let recognize the default pose as an abort command
		if(!TrainingMode)
		{
			if (GestureUtility::getPoseDistance(current_pose, DefaultGesture.getFirstPose()) < PoseThreshold)
			{
				clock_t time = TimeFunctions::getTime();

				if( ! ( (LastRecognitionTime.GestureSetID == -2) && ((time - LastRecognitionTime.time) < 500) ) )
				{
					GestureMessageQueue.push_back("DEFAULT_GESTURE");
					std::cout<<"gesture recognized as: DEFAULT_GESTURE "<<std::endl;
					LastRecognitionTime.GestureSetID = -2;
					LastRecognitionTime.time = time;
				}
			}
		}
	}
	else
	{
		// GESTURE SPOTTING:
		// First stage: we tag special candidate frames. these could be possible start or end frames for gestures
		// we separate into different cases with descending relevance:
		// a) we have a velocity change from above to below the low threshold
		// b) we have a velocity change from below to above the high threshold
		// c) the velocity is below the low threshold or above the high threshold and there was no tag for a long time (i.e. for MinimumRecognitionIntervalLength ticks)
		// d) a pose corresponds to a start or end point of an already trained gesture if there was no tag for long time (this will happen in motion phases with a velocity between the two thresholds)
	
		bool below  = (velocity < VelocityLowThreshold) && (last_velocity > VelocityLowThreshold); 
		bool above  = (velocity > VelocityHighThreshold) && (last_velocity < VelocityHighThreshold);
	
		if(!TrainingMode) 
		{
			// cases a) and b)
			if(below || above)
			{
				info.flag = true;
			}
			else 
			{
				// we check the last flags
				bool check = true;
				for(int i=0; i <= MinimumRecognitionIntervalLength; i++)
				{
					if(RecognitionList[RecognitionList.size() - 1 - i].flag)
						check = false;
				}

				// case c)
				if( (velocity < VelocityLowThreshold) || (velocity > VelocityHighThreshold))
				{
					info.flag = check;
				}
				// case d)
				else
				{	
					info.flag = check && checkPose(info);
				}
			}

			if(info.flag)
			{
				// we hold the flagged IDs on a special "moving window"
				MovingWindow.push_back(info.ID);
			}
		}
	}

	// push the recognition info on the deque
	RecognitionList.push_back(info);
	// take out the first element due to maximal length of gesture
	RecognitionList.pop_front();

	// update the moving window
	updateMovingWindow();

	CURRENT_SIZE_OF_MOVING_WINDOW = MovingWindow.size();

	// we examine a snap shot of the moving window inside a new thread
	std::deque<int> moving_window_copy = MovingWindow;

MUTEX.unlock();

	// if we are in training mode (don't put it into the else branch above due to possible dead locks)
	if(TrainingMode) 
	{
		findTrainingInterval();
	}


	if(spot_gesture && !TrainingMode)
	{
MUTEX.lock();
		if(!IS_RECOGNITION_THREAD_RUNNING)
		{
			boost::thread RecognitionThread(&GestureRecognizer::spotGesture, this, moving_window_copy);
		}
MUTEX.unlock();
	}
	return pose;
}


void GestureRecognizer::updateMovingWindow()
{
	// we pop the first element if it's id is invalid
	//std::cout<<"Window Size = "<<MovingWindow.size()<<"\n";
	if(MovingWindow.size() > 0)
	{
		if(getIndexFromID(MovingWindow.front()) == -1)
			MovingWindow.pop_front();
	}

	// if the window size is too large we cut
	while(MovingWindow.size() > RecognitionWindowSize)
	{
		int i = MovingWindow.front();
		RecognitionList[getIndexFromID(i)].flag = false;
		MovingWindow.pop_front();
	}
}


bool GestureRecognizer::checkPose(RecognitionInfo &info)
{
	// we step through all gesture sets and try to find corresponding start or end poses
	for (unsigned int i = 0; i < TrainedGestureSets.size(); i++)
	{
		
		if( (!UseKNNAlgorithm) )
		{
			// we only check against the mean of the gesture classes
			float dist_to_first = GestureUtility::getPoseDistance(info.pose, TrainedGestureSets[i]->MeanGesture.getFirstPose(), TrainedGestureSets[i]->SideOfBody);
			float dist_to_last  = GestureUtility::getPoseDistance(info.pose, TrainedGestureSets[i]->MeanGesture.getLastPose(), TrainedGestureSets[i]->SideOfBody);
			// we only need to know if there is a corresponding pose, but we don't need to find the best match
			return ( (dist_to_first < PoseThreshold) || (dist_to_last < PoseThreshold) );
		}
		else
		{
			// we check against all the gestures
			for(unsigned int j = 0; j < TrainedGestureSets[i]->Gestures.size(); j++)
			{
				float dist_to_first = GestureUtility::getPoseDistance(info.pose, TrainedGestureSets[i]->Gestures[j]->getFirstPose(), TrainedGestureSets[i]->SideOfBody);
				float dist_to_last  = GestureUtility::getPoseDistance(info.pose, TrainedGestureSets[i]->Gestures[j]->getLastPose(), TrainedGestureSets[i]->SideOfBody);
				// we only need to know if there is a corresponding pose, but we don't need to find the best match
				return ( (dist_to_first < PoseThreshold) || (dist_to_last < PoseThreshold) );
			}
		}
	}
	return false;
}


void GestureRecognizer::spotGesture(std::deque<int> moving_window)
{
	//clock_t start = TimeFunctions::getTime();

MUTEX.lock();
	IS_RECOGNITION_THREAD_RUNNING = true;
MUTEX.unlock();
	
	if(moving_window.size() < RecognitionWindowSize)
	{
MUTEX.lock();
		IS_RECOGNITION_THREAD_RUNNING = false;
MUTEX.unlock();
		return;
	}

	// first we search the intervals of interest
	std::vector<RecognitionInterval> LastRecognitionIntervals = findRecognitionIntervals(moving_window);

	// now we have all the possible intervals in the LastRecognitionIntervals list
	// we check all of them for a minimum length and copy the valid intervals
	if(LastRecognitionIntervals.size() == 0)
	{
MUTEX.lock();
		IS_RECOGNITION_THREAD_RUNNING = false;
MUTEX.unlock();
		return;
	}

	std::vector<RecognitionInterval> swap;

	for (unsigned int i = 0; i < LastRecognitionIntervals.size(); i++)
	{
		// the intervals must have a minimum length
		if( LastRecognitionIntervals[i].size > MinLengthOfGestures )
		{
			swap.push_back(LastRecognitionIntervals[i]);
		}
	}

	// swap the vectors, all valid intervals with minimum length are in "LastRecognitionIntervals" now
	swap.swap(LastRecognitionIntervals);

	// DATA MINING average calculation time
	//clock_t average_time_start = TimeFunctions::getTime();

	// now we check all the intervals and find the best distance match
	std::vector<KNNClassification> winners;

	for (unsigned int i = 0; i < LastRecognitionIntervals.size(); i++)
	{
		KNNClassification winner = recognizeInterval(LastRecognitionIntervals[i]);

		if (winner.GestureSetID != -1)
		{
			winner.id_start = LastRecognitionIntervals[i].id_start;
			winner.id_end = LastRecognitionIntervals[i].id_end;
			winner.size = LastRecognitionIntervals[i].size;
			winners.push_back(winner);
		}
	}

	// DATA MINING average calculation time
	//std::cout<<TimeFunctions::getTime() - average_time_start<<"\n";

	// maybe we found more than one winner. than we take the one with the biggest interval length to due to co-articulation
	int max_interval_length = 0;
	KNNClassification best;

	if (winners.size() > 0)
	{
		for (unsigned int i = 0; i < winners.size(); i++)
		{
			int current_interval_length = winners[i].size;
			if ( current_interval_length > max_interval_length)
			{
				max_interval_length = current_interval_length;
				best = winners[i];
			}
		}
		// we have found the class of the gesture! let's push its ID on the event queue
		std::stringstream ss;
		if(best.GestureSetID == -2)
		{
			ss<<"DEFAULT_GESTURE";
		}
		else
		{
			ss<<best.GestureSetID;
		}

MUTEX.lock();
		// we have to check if the same gesture class was currently recognized to avoid duplicates
		clock_t time = TimeFunctions::getTime();
		if( ! ( (LastRecognitionTime.GestureSetID == best.GestureSetID) && ((time - LastRecognitionTime.time) < 2000) ) ) // we have a fixed repetition rate of 2000 ms
		{
			GestureMessageQueue.push_back(ss.str());
			std::cout<<"gesture recognized as: "<<best.GestureSetID<<", points: "<<best.points<<std::endl;
			std::cout<<time - LastRecognitionTime.time<<"\n";

			LastRecognitionTime.GestureSetID = best.GestureSetID;
			LastRecognitionTime.time = time;
		}

		//we delete all entries in the moving window before the winner element's id_end (hence this frame can be a new starting point for the next gesture)
		int index = -1;
		//check if the element is in the list (it could be popped in the updateMovingWindow function)
		for (unsigned int i = 0; i < MovingWindow.size(); i++)
		{
			if(MovingWindow[i] == best.id_end)
					index = i;
		}
		
		for (int i = 0; i < index; i++)
		{
			int j = getIndexFromID(MovingWindow.front());
			if(j != -1)
				RecognitionList[j].flag = false;

			MovingWindow.pop_front();
		}

MUTEX.unlock();
	}
	else
	{
MUTEX.lock();
		// if there is no winner we pop the first element of the moving window
		/*if(MovingWindow.size() > 0)											<-- already done in updateMovingWindow()
		{
		MovingWindow.pop_front();
		}*/
		
MUTEX.unlock();
	}

// close the thread
MUTEX.lock();
	IS_RECOGNITION_THREAD_RUNNING = false;
MUTEX.unlock();

//std::cout<<"Window size: "<<RecognitionWindowSize<<" ,  calculation Time: "<<TimeFunctions::getTime() - start<<"\n";
}


// find the movement interval boundaries in training mode
void GestureRecognizer::findTrainingInterval()
{
	// if the last pose is flagged we don't recognize
	if(RecognitionList[RecognitionList.size()-2].flag)
		return;

	// we look back in time and try to find the interval boundaries for a  gesture
	// there are 2 possibilities:
	// a) a static gesture hence we have no motion at all, all the values are near zero
	// b) a dynamic motion, we have to look for an interval between two periods of no motion

	// a) check for non-motion at the end of the recognition list 
	bool search = true;
	for ( int i = 1; i <= InactivityTime; i++ )
	{
		if (RecognitionList[RecognitionList.size() - 1 - i].velocity > VelocityLowThreshold)
		{
			search = false;
		}
	}
	
	int begin = -1;
	int end = 0;

	// let's start a search
	if (search)
	{
		// we try to find an interval of activity between two periods of inactivity
		for ( int i = RecognitionList.size() - InactivityTime; i >= 0; i--)
		{
			// if we find a flagged pose we don't search any longer
			if(RecognitionList[i].flag)
			{
				end = 0;
				return;
			}

			// search for first velocity which is above the low threshold, this is our interval end
			if (RecognitionList[i].velocity > VelocityLowThreshold)
			{
				end = i;
				break;
			}
			// if we do not find a velocity above the threshold after half the recognition list size we assume a static gesture
			if (i == RecognitionList.size()/3)
			{
				// we train the interval and return
				RecognitionInterval result;
				result.size = RecognitionList.size() - InactivityTime - i + 1;
				result.id_start = RecognitionList[i].ID;
				result.id_end = RecognitionList[i + result.size - 1].ID;
				recognizeInterval(result);
				return;
			}
		}
	
		// if there was a user motion we search for the interval start 
		// we have to count the contiguous values below the high threshold, all the values above this threshold describe the movement
		int count = 0;
		int temp_begin;

		for (int i = end - 1; i >= 0; i--)
		{
			if ( RecognitionList[i].velocity > VelocityHighThreshold )
			{
				count = 0;
			}	
			else
			{
				//first velocity which is below the threshold
				if (count == 0) temp_begin = i;
				//every time we find another value below the threshold we increment count
				count++;
				// if we get "InactivityTime" contiguous values in row we can set the interval start (i.e. a short break within the movement is accepted as long as its duration is not larger than "InactivityTime" ticks)
				if (count == InactivityTime)
				{
					// we found the interval begin and break the loop
					begin = temp_begin;
					break; 
				}
			}
		}

		// if there's a valid interval with a minimum length
		if ( (begin != -1) && ((end - begin) >= MinLengthOfGestures) )
		{
			// we train the interval
			RecognitionInterval result;
			result.id_start = RecognitionList[begin].ID;
			result.id_end = RecognitionList[end].ID;
			result.size = end - begin + 1;
			recognizeInterval(result);
		}
	}
}


std::vector<RecognitionInterval> GestureRecognizer::findRecognitionIntervals(std::deque<int> &moving_window)
{
	std::vector<RecognitionInterval> intervals;
	// find current indices in the Recognition List only to get their distances (they will not change over time)
	std::vector<int> indices;

MUTEX.lock();
	for(unsigned int i = 0; i < moving_window.size(); i++)  // we only check the first "RecognitionWindowSize" elements
	{
		indices.push_back(getIndexFromID(moving_window[i]));
	}
MUTEX.unlock();

	// we check the the moving window and write all possible intervals from the candidate frames into recognition interval structs
	for(unsigned int i = 0; i < moving_window.size(); i++)
	{
		for(unsigned int j = i + 1; j < moving_window.size(); j++)
		{
			//if both indices are valid
			if( (indices[i] != -1) && (indices[j] != -1) )
			{
				RecognitionInterval interval;
				interval.id_start = moving_window[i];
				interval.id_end = moving_window[j];
				interval.size = indices[j] - indices[i] + 1;
				intervals.push_back(interval);
			}
		}
	}
		
	return intervals;
}



float GestureRecognizer::getCurrentVelocity()
{
	return RecognitionList.back().velocity;
}


void GestureRecognizer::setTrainingMode(bool new_set, bool replace)
{
	bool go_on = false;

	while(!go_on)
	{
MUTEX.lock();
		// we have to assure that the recognition thread is not still running
		go_on = !IS_RECOGNITION_THREAD_RUNNING;
MUTEX.unlock();
	}

	TrainingMode = true;
	ReplaceGesture = replace;
	NewGestureSet = new_set;
	MovingWindow.clear();
	// we flag all the poses
	setAllPosesFlags(true);
}


bool GestureRecognizer::isTrainingMode()
{
	return TrainingMode;
}


Gesture* GestureRecognizer::getCurrentTrainedGesture()
{
	return TrainedGestureSets[CurrentTrainedGestureSetIndex]->Gestures[CurrentTrainedGestureIndex];
}


int GestureRecognizer::removeCurrentTrainedGestureSet()
{
	if ( TrainedGestureSets.size() > 0)
	{
		// check if there are still gestures in the set
		if ( TrainedGestureSets[CurrentTrainedGestureSetIndex]->Gestures.size() > 0)
		{
			// remove all gestures from that set
			for (unsigned int i = 0; i < TrainedGestureSets[CurrentTrainedGestureSetIndex]->Gestures.size(); i++)
			{
				delete TrainedGestureSets[CurrentTrainedGestureSetIndex]->Gestures[i];
			}
			TrainedGestureSets[CurrentTrainedGestureSetIndex]->Gestures.clear();
		}
		// remove the current gesture set
		delete TrainedGestureSets[CurrentTrainedGestureSetIndex];
		TrainedGestureSets.erase(TrainedGestureSets.begin() + CurrentTrainedGestureSetIndex);
		
		//if( (SimilarityMeasure != DTW) || (!UseKNNAlgorithm) )
		//{
		//	// recalculate min distances
		//	setMinDistances(); <- do we still need that?
		//}

		// set the gesture set index to the previous one
		return setIndexToPreviousTrainedGestureSet();
	}
	else return -1;
}


int GestureRecognizer::removeCurrentTrainedGesture()
{
	if ( TrainedGestureSets[CurrentTrainedGestureSetIndex]->Gestures.size() > 0)
	{
		// we remove the current gesture from the current set
		delete TrainedGestureSets[CurrentTrainedGestureSetIndex]->Gestures[CurrentTrainedGestureIndex];
		TrainedGestureSets[CurrentTrainedGestureSetIndex]->Gestures.erase(TrainedGestureSets[CurrentTrainedGestureSetIndex]->Gestures.begin() + CurrentTrainedGestureIndex);

		if( (!UseKNNAlgorithm) )
		{
			// recalculate mean gesture of current set
			calculateMeanGesture(CurrentTrainedGestureSetIndex);
			// recalculate min distances
			//setMinDistances(); <- do we still need that
		}

		// now, we set the gesture index to the previous one (if the set has no gestures anymore we erase it too)
		if (setIndexToPreviousTrainedGesture() == -1)
		{
			CurrentTrainedGestureIndex = 0;
			return removeCurrentTrainedGestureSet();
		}
		else return CurrentTrainedGestureIndex;
	}
	else return -1;
}


int GestureRecognizer::setIndexToNextTrainedGesture()
{
	int size = TrainedGestureSets[CurrentTrainedGestureSetIndex]->Gestures.size();
	if( size != 0 )
	{
		CurrentTrainedGestureIndex = (CurrentTrainedGestureIndex + 1) % size;
		return CurrentTrainedGestureIndex;
	}
	else return -1;
}


int GestureRecognizer::setIndexToPreviousTrainedGesture()
{
	int size = TrainedGestureSets[CurrentTrainedGestureSetIndex]->Gestures.size();
	if( size != 0 )
	{
		CurrentTrainedGestureIndex = (CurrentTrainedGestureIndex - 1 + size) % size;
		return CurrentTrainedGestureIndex;
	}
	else return -1;
}


int GestureRecognizer::setIndexToNextTrainedGestureSet()
{
	if( TrainedGestureSets.size() != 0 )
	{
		CurrentTrainedGestureSetIndex = (CurrentTrainedGestureSetIndex + 1) % TrainedGestureSets.size();
		CurrentTrainedGestureIndex = 0;
		return CurrentTrainedGestureSetIndex;
	}
	else return -1;
}


int GestureRecognizer::setIndexToPreviousTrainedGestureSet()
{
	if( TrainedGestureSets.size() != 0 )
	{
		CurrentTrainedGestureSetIndex = (CurrentTrainedGestureSetIndex - 1 + TrainedGestureSets.size()) % TrainedGestureSets.size();
		CurrentTrainedGestureIndex = 0;
		return CurrentTrainedGestureSetIndex;
	}
	else return -1;
}



int GestureRecognizer::getMaxLengthOfGestures()
{
	return MaxLengthOfGestures;
}


float GestureRecognizer::getVelocityLowThreshold()
{
	return VelocityLowThreshold;
}


float GestureRecognizer::getVelocityHighThreshold()
{
	return VelocityHighThreshold;
}


std::string GestureRecognizer::getLastEvent()
{
MUTEX.lock();
	std::string ev = GestureMessageQueue.front();
	GestureMessageQueue.pop_front();
MUTEX.unlock();
	return ev;
}


// file handling
bool GestureRecognizer::saveTrainedGestures(std::string filename)
{
	int nr_of_gesture_sets = TrainedGestureSets.size();
	int count = 0;
	if (nr_of_gesture_sets == 0)
	{
		std::cout<<"Nothing to save!"<<std::endl;
		return false;
	}

	std::ofstream file(filename.c_str(), std::ios::out | std::ios::binary | std::ios::trunc);

	if(file.is_open())
	{
		// first we write the number of trained gesture sets
		file.write( (char*) &nr_of_gesture_sets, sizeof(int));

		// then we successively write all the gesture data
		for (int i = 0; i < nr_of_gesture_sets; i++)
		{
			// we need the number of gestures per set
			int nr_of_gestures = TrainedGestureSets[i]->Gestures.size();
			file.write( (char*) &nr_of_gestures, sizeof(int));
			for (int j = 0; j < nr_of_gestures; j++)
			{
				TrainedGestureSets[i]->Gestures[j]->saveGesture(file);
				count++;
			}
		}

		file.close();		
		std::cout<<"Saved "<<count<<" gestures in "<< nr_of_gesture_sets <<" sets to disk!\n";
		return true;
	}
	else
	{
		std::cout<<"Cannot write to disk!"<<std::endl;
		return false;
	}
}

bool GestureRecognizer::loadTrainedGestures(std::string filename)
{
	std::ifstream file(filename.c_str(), std::ios::in | std::ios::binary);

	if(file.is_open())
	{
		// clear all the current gesture sets
		CurrentTrainedGestureSetIndex = 0;
		while (TrainedGestureSets.size() > 0)
		{
			removeCurrentTrainedGestureSet();
		}
		
		// we check the size of TrainedGestureSets
		assert(TrainedGestureSets.size() == 0);

		
		int nr_of_gesture_sets;
		// we count the total number of loaded gestures
		int count = 0;

		// first we load the number of trained gesture sets
		file.read( (char*) &nr_of_gesture_sets, sizeof(int));

		
		bool resample = (SimilarityMeasure == EUCLIDEAN) ||
						(SimilarityMeasure == PEARSON_CORRELATION) ||
						(SimilarityMeasure == PEARSON_CORRELATION_SQUARED) ||
						(SimilarityMeasure == MODIFIED_EUCLIDEAN) ||
						(SimilarityMeasure == MODIFIED_PEARSON) ||
						(!UseKNNAlgorithm);

		// then we load all the gesture data
		for (int i = 0; i < nr_of_gesture_sets; i++)
		{
			int nr_of_gestures;
			file.read( (char*) &nr_of_gestures, sizeof(int));

			GestureSet* temp_set = new GestureSet();

			for(int j = 0; j < nr_of_gestures; j++)
			{
				Gesture* temp_gest = new Gesture();
				temp_gest->loadGesture(file);

				
				if( (resample) )
				{
					if(temp_gest->getSize() != ResamplingSize)
						GestureUtility::resamplePoseVector(temp_gest->getPoses(), ResamplingSize);
				}
				
				temp_set->Gestures.push_back(temp_gest);
			
				count++;
			}
			temp_set->MinDistance = INFINITY;
			temp_set->SideOfBody = temp_set->Gestures[0]->getGestureType();
			TrainedGestureSets.push_back(temp_set);
		}

		CurrentTrainedGestureSetIndex = 0;
		CurrentTrainedGestureIndex = 0;


		if( (resample) )
		{
			setAllVariances();
			calculateAllMeanGestures();
			//setMinDistances(); <- do we still need that?
		}

		file.close();	
		std::cout<<"Loaded "<<count<<" gestures in "<<nr_of_gesture_sets<<" sets from disk!\n";
		return true;
	}
	else
	{
		std::cout<<"Cannot load from disk!"<<std::endl;
		return false;
	}
}



//************** statistical analysis ***************
// print out trained gestures for statistical reasons
void GestureRecognizer::stat_printGestures()
{
	for (unsigned int i = 0; i < TrainedGestureSets.size(); i++)
	{
		for (unsigned int j = 0; j < TrainedGestureSets[i]->Gestures.size(); j++)
		{
			std::stringstream ss;
			ss<<"gesture_"<<i;
			ss<<"_"<<j;
			ss<<".txt";
			writeGestureToDisk(TrainedGestureSets[i]->Gestures[j], ss.str().c_str(), false);
		}

		std::stringstream ss;
		ss<<"gesture_set_"<<i;
		ss<<"_mean_gesture.txt";
		writeGestureToDisk(&(TrainedGestureSets[i]->MeanGesture), ss.str().c_str(), false);
	}
}

// validate gesture sets against training data in respect to k of "k"-nearest neighbour algorithm, we asume that the gesture sets of test and training data correspond to each other
void GestureRecognizer::stat_printValidateGestures(std::vector<std::vector<Gesture*>> gestures_to_validate, int minimum_neighbours, int maximum_neighbours, SimilarityMeasureType measure_type, float user_threshold, const char* filename)
{
	// for every gesture in the list we do a k-nearest-neighbour validation over an interval of [minimum_neighbours, maximum_neighbours] using the given similarity measure
	// and print out to file

	UseKNNAlgorithm = true;

	int min_neighbours = minimum_neighbours;
	int max_neighbours = maximum_neighbours;

	setSimilarityMeasurement(measure_type, user_threshold);

	// MODIFIED FM MEASURE USES ANOTHER ALGORITHM! hence we reset the neighbours to one
	if(measure_type == MODIFIED_FM)
	{
		min_neighbours = 1;
		max_neighbours = 1;
	}

	std::stringstream ss;
	ss<<"Gesture Class";
	for(int j = min_neighbours; j <= max_neighbours; j++)
	{
		ss<<"\tnr. of neighbours"<<"\trecognized right"<<"\trecognized wrong"<<"\tnot recognized"<<"\tcalc. time";
	}


	for (unsigned int i = 0; i < gestures_to_validate.size(); i++)
	{
		ss<<"\n"<<i;
	
		std::cout<<"\r Validation of gesture class "<<i;

		for(int j = min_neighbours; j <= max_neighbours; j++)
		{
			NrOfNeighbours = j;
		
			int recognized_right = 0;
			int recognized_wrong = 0;
			int not_recognized = 0;
			clock_t start = TimeFunctions::getTime();

		

			for (unsigned int k = 0; k < gestures_to_validate[i].size(); k++)
			{
				KNNClassification winner;
				winner.GestureSetID = -1;
				winner.points = -1;

				recognizeGesture(*(gestures_to_validate[i][k]), winner);
				if(winner.GestureSetID == i)
				{
					recognized_right++;
				}
				else
				{
					if(winner.GestureSetID == -1)
					{
						not_recognized++;
					}
					else
					{
						recognized_wrong++;
					}
				}
			}	
		
			ss<<"\t"<<j<<"\t"<<recognized_right<<"\t"<<recognized_wrong<<"\t"<<not_recognized<<"\t"<<(TimeFunctions::getTime() - start);
		}
	}

	// we open a file
	std::ofstream file(filename, std::ios::trunc);
	
	if (file.is_open())
	{
		// write data into file
		file<<ss.str();
		file.close();
	}
	else std::cout << "Unable to open file";
}
	
void GestureRecognizer::stat_printCalculationOfMeanGestures()
{
	// we concatenate all the gestures from a set to on big vector
	calculateAllMeanGestures();

	for (unsigned int i = 0; i < TrainedGestureSets.size(); i++)
	{
		for (unsigned int j = 0; j < TrainedGestureSets[i]->Gestures.size(); j++)
		{
			std::stringstream ss;
			ss<<"gesture_set_"<<i;
			ss<<"_gestures.txt";
			writeGestureToDisk(TrainedGestureSets[i]->Gestures[j], ss.str().c_str(), true);
		}

		std::stringstream ss;
		ss<<"gesture_set_"<<i;
		ss<<"_mean_gesture.txt";
		writeGestureToDisk(&(TrainedGestureSets[i]->MeanGesture), ss.str().c_str(), false);
	}
}