/*
Class providing recognition of two hand shapes: open and closed hand.
We cut out the region in the depth map where we assume the hand (via a bounding 
box algorithm) and perform a scale invariant pixel-based search along rays from 
the hand wrist to the fingers under particular angles to find the contour line of the hand.

The HandRecognizer provides own event list with the following event ids:
LEFT_HAND_CLOSED
LEFT_HAND_OPENED
RIGHT_HAND_CLOSED
RIGHT_HAND_OPENED

*/
#pragma once

#include "KinectSensor.h"
#include <deque>
#include <vector>
#include <string>
#include "math\Vec3f.h"
#include "math\Vec2f.h"

// struct containing pixel data of hands
struct HandImage
{
	unsigned int width;
	unsigned int height;
	BYTE* imagedata;
};

// the two hand shapes
enum HandState{H_OPEN, H_CLOSED};
// side of body
enum HandType{H_LEFT, H_RIGHT};

class HandRecognizer
{
public:

	HandRecognizer(KinectSensor* sensor, HandType which_hand = H_RIGHT);
	~HandRecognizer(void);

	// invokes hand shape recognition and updates the hand state (should be called each time in the main program loop)
	void update();
	// returns the distance vector of the hand contour pixels
	std::vector<float> & getContourPixelsDistances(){ return ContourPixelsDistances;}
	// returns the mean value of the contour pixel distances
	float getContourPixelsDistancesMean(){return ContourPixelsDistancesMean;}
	// returns the maximum distances between wrist and a contour pixel
	std::vector<int> & getMaximaOfContourPixelsDistances(){ return MaximaOfContourPixelsDistances;}
	// returns size of the neighbourhood in the contour line algorithm
	unsigned int getSizeOfMaximaNeighbourHood(){return SizeOfMaximaNeighbourHood;};
	
	// returns the last image of hand in bounding box
	HandImage* getLastHandImageData();
	// returns the current state
	HandState getHandState();

	// event handling
	static bool isEventWaiting(){return (HandMessageQueue.size() > 0);};
	// returns the last event and pops it from the event list
	static std::string getLastEvent();

	// threshold to accept a pixels depth
	unsigned int DepthThreshold;
	// size of the neighbourhood in the contour line algorithm
	unsigned int SizeOfMaximaNeighbourHood;
	// radius of mean filter in the contour line algorithm
	unsigned int SizeOfContourFilter;
	// size of gradient in the contour line algorithm
	float ContourGradient;






private:
	// updates the current state of hand
	void updateState();
	// sets the bounding box boundaries
	void setBoundingBox(const USHORT* depth_map, int depth, int hand_x, int hand_y, float angle_val, int &left_boundary, int &right_boundary, int &top_boundary, int &bottom_boundary);


	// list containing all hand shape events
	static std::deque<std::string> HandMessageQueue;
	// the type of the hand
	HandType SideOfBody;
	// indices of involved joints
	unsigned int WristSkeletonIndex;
	unsigned int ElbowSkeletonIndex;
	// assigned kinect sensor
	KinectSensor* NuiSensor;
	// information about hand position 
	bool HandIsInFront;
	// vector containing the distances from center (=wrist) to contour pixels
	std::vector<float> ContourPixelsDistances;
	// mean value of "ContourPixelsDistances"
	float ContourPixelsDistancesMean;
	// indices of local maxima in the contour line corresponding to finger positions
	std::vector<int> MaximaOfContourPixelsDistances;
	// small deque of flags that specify if hand was opened or closed the last x frames ("closed" is equivalent to "true").
	std::deque<bool> HandFlags;
	// last pixel image of hand
	HandImage LastHandImage;
	// current state of hand
	HandState State;





public:
	//************** statistical analysis and data mining ***************
	// list containing classification infos about closed hand recognition
	std::vector<math::Vec3f> stat_ClosedHandParams;
	// list containing classification infos about open hand recognition
	std::vector<math::Vec3f> stat_OpenHandParams;
	// reference state
	bool stat_HandClosed;
	// flag to decide if we shall capture classification results
	bool stat_captureParams;
	// sets "stat_captureParams" and "stat_HandClosed" 
	void stat_setCapture(bool trigger, bool hand_closed);
	// write classification analysis to file
	void stat_printStats();
	//*******************************************************************
};

