/*

This tool provivides the visualization of system parameters and recognition results as well as validation methods.
There is also a sub program to generate test data and a little easter egg application to test the recognition system practically (press 'p'). 
For the sake of convinience (and because glut forces static callbacks) all the functionality has been put into this one file. 
The actual demo application ("Kinect Terrain Modeler") has a much better software design.

*/
#include <windows.h>

#include <GL/freeglut.h>
#include <glui.h>
#include <iostream>
#include <sstream>
#include <iomanip>
#include <cstdlib>
#include <SOIL.h>
#include <deque>

#include "math/Math.h"

#include "WindowProperties.h"
#include "FileHandling.h"
#include "TimeFunctions.h"

#include "KinectSensor.h"
#include "Gesture.h"
#include "SkeletonAnimation.h"

#include "GestureUtility.h"
#include "GestureRecognizer.h"
#include "HandRecognizer.h"
#include "KinectPointer.h"

#include "Countdown.h"
#include "CircleButton.h"
#include "TextureTextRenderer.h"

#include "Tools\Camera.h"
#include "Tools\OrbitManipulator.h"
#include "Tools\Tool.h"
#include "Tools\Colorpicker.h"
#include "Tools\NullTool.h"
#include "Tools\Manipulator2D.h"
#include "Tools\Brush2D.h"
#include "Tools\StraightLine2D.h"
#include "Tools\StrokeEraser2D.h"
#include "Tools\Ink2D.h"
#include "Tools\Rubber2D.h"
#include "Tools\UndoRedoManager.h"

#include <assert.h>


using namespace GestureUtility;

// ********  DEBUGGING & STATISTICS ************
// recognition test  
void stat_validateRecognition(const char* test_file, const char* training_file);
// writes gathered poses to file
void stat_printPoses();
// cross validation test of gestures in given test file for all similarity measures
void stat_crossValidateGestures(const char* training_file);

std::ofstream of_velocites;
bool write_velocities_to_file = false;
int FRAME_COUNT = 0;
int COUNT = 50;
std::vector<Pose> POSES; // <- used for writing poses vector into file
std::vector<math::Vec2f> RightHandCoordsWithoutSmoothing;
std::vector<math::Vec2f> RightHandCoordsWithSmoothing;
// --------------------------------------------------------------------------------------------


// window informations
WindowProperties WindowProps;
int MainWindowID;

// TextureID for VideoFrame
GLuint VideoTexID;

//Kinect Wrapper
KinectSensor* NuiSensor;

// Gesture Reognizer
GestureRecognizer* GRecognizer;
bool HandleGestures = true;

// Hand Recognizers
HandRecognizer* LeftHandRecognizer;
HandRecognizer* RightHandRecognizer;

// virtual hand pointers
KinectPointer* RightHandPointer;
KinectPointer* LeftHandPointer;

// font rendering module
TextureTextRenderer* InfoText;
TextureTextRenderer* RecognizerStateText;

// app mode (0 = depth map view, 1 = training menu, 2 = easter egg app)
int app_mode = 0;

// renderable countdown
Countdown* Timer;

// current pose and pose velocity (glui live variable) 
Pose CurrentPose;
float PoseDifference;

// vector to hold all buttons
std::vector<CircleButton*> ButtonList;
std::vector<CircleButton*> MenuTrainingButtonList;

// Text module for gesture menu
TextureTextRenderer* GestureMenuText;

// background texture for gesture menu
GLuint TrainingBGTexID;

// list box
GLUI_Listbox* SMListBox;



//***** EASTER EGG ZONE - PAINT APP *************************************************************************
//Texture frame for current color
GLuint TextureFrameID;
// list of tools
std::vector<Tool*> ToolList;
Tool* ActiveTool;
// list containing all strokes
std::vector<Stroke> Strokes;
int SelectedStrokeID = -1;

Manipulator2D* PaintManipulator;


// generate id from rgb color
int RGBtoIndex(unsigned char R, unsigned char G, unsigned char B)
{
	return 65536 * R + 256 * G + B;
}

// generate rgb color from id
std::vector<unsigned char> IndexToRGB(unsigned int index)
{
	unsigned char R = index / 65536;
	index -= R * 65536;
	unsigned char G = index / 256;
	index -= G * 256;
	unsigned char B = index;

	std::vector<unsigned char> result;
	result.push_back(R);
	result.push_back(G);
	result.push_back(B);

	return result;
}

// provides tool handling
void updateActiveTool(Pose& pose)
{
	bool left_active = (LeftHandRecognizer->getHandState() == H_CLOSED) && (LeftHandPointer->isActive());
	bool right_active = (RightHandRecognizer->getHandState() == H_CLOSED) && (RightHandPointer->isActive());

	int code = left_active + 2 * right_active;
	
	switch(code)
	{
	case 1:
		ActiveTool->updateLeftHandInput(pose);
		break;
	case 2:
		ActiveTool->updateRightHandInput(pose);
		break;
	case 3:
		ActiveTool->updateBothHandInput(pose);
		break;
	default:
		ActiveTool->update(pose);
	}
}

//helper function for paint app
void drawStroke(int i)
{
	glLineWidth(Strokes[i].LineWidth);
	glBegin(GL_LINE_STRIP);
	for (unsigned int j = 0; j < Strokes[i].Points.size(); j++)
	{
		glVertex2fv(Strokes[i].Points[j].v);			
	}
	glEnd();
	glPointSize(Strokes[i].LineWidth);
	glBegin(GL_POINTS);
	for (unsigned int j = 0; j < Strokes[i].Points.size(); j++)
	{
		glVertex2fv(Strokes[i].Points[j].v);			
	}
	glEnd();
}

// set the Stroke ID from the currently selected stroke
bool setSelectedStrokeID()
{
	// we try to find the coordinates of the pointer in object space with colorpicking
	glClearColor(1.0, 1.0, 1.0, 1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glDisable(GL_LIGHTING);
	glDisable(GL_LINE_SMOOTH);
	glDisable(GL_POINT_SMOOTH);

	for (unsigned int i = 0; i < Strokes.size(); i++)
	{
		glColor3ub(IndexToRGB(i)[0], IndexToRGB(i)[1], IndexToRGB(i)[2]);
		drawStroke(i);
	}

	math::Vec2f pointer_coords = RightHandPointer->getWindowCoordinates();
	
	//pointer_coords.y = min(pointer_coords.y, WindowProps.WindowHeight-1); // we have a small region with a wrong background color at the very top of the viewport area????
	
	unsigned char color[3];
	glReadPixels((int) pointer_coords.x, (int) pointer_coords.y, 1, 1, GL_RGB, GL_UNSIGNED_BYTE, color);

	//std::cout<<(int) color[0]<<"   "<<(int) color[1]<<"   "<<(int) color[2]<<"\n";

	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_POINT_SMOOTH);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	if( (color[0] == 255) && (color[1] == 255) && (color[2] == 255) )
	{
		// this is the background color. hence we didn't find any stroke
		SelectedStrokeID = -1;
		return false;
	}

	// we found a stroke, set the SelectedStrokeID
	int selected_index = RGBtoIndex(color[0], color[1], color[2]);
	SelectedStrokeID = Strokes[selected_index].ID;

	return true;
}

// draw all of the strokes in our list
void paint_drawPainting()
{
	glPushMatrix();
		
	glLoadMatrixf(PaintManipulator->getTransformationMatrix().ma);

	// get the selected stroke with colorpicking
	setSelectedStrokeID();

	// render strokes
	for (unsigned int i=0; i<Strokes.size(); i++)
	{
		/*if(Strokes[i].ID == SelectedStrokeID)
			glColor3f(0.8, 0.8, 0.8);
		else*/
			glColor3fv(Strokes[i].Color.v);
		drawStroke(i);
	}
		
	glPopMatrix();

	// show tool on the top right side of the window
	glPushMatrix();
	glTranslatef((float) WindowProps.WindowWidth - 148.0f, (float) WindowProps.WindowHeight - 148.0f, 0.0f);
	glColor4d(1.0, 1.0, 1.0, 1.0);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, ActiveTool->getTextureID());
	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 1.0);		glVertex3f(0.0, 0.0, 0.0);
		glTexCoord2f(1.0, 1.0);		glVertex3f(128.0, 0.0, 0.0);
		glTexCoord2f(1.0, 0.0);		glVertex3f(128.0, 128.0, 0.0);
		glTexCoord2f(0.0, 0.0);		glVertex3f(0.0, 128.0, 0.0);
	glEnd();
	glPopMatrix();
	
	// show current color in a frame on the bottom right side of the window
	glPushMatrix();
	glTranslatef((float) WindowProps.WindowWidth - 148.0f, 20.0f, 0.0f);
	glColor3fv(Colorpicker::CurrentColor.v);
	glBindTexture(GL_TEXTURE_2D, TextureFrameID);
	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 1.0);		glVertex3f(0.0, 0.0, 0.0);
		glTexCoord2f(1.0, 1.0);		glVertex3f(128.0, 0.0, 0.0);
		glTexCoord2f(1.0, 0.0);		glVertex3f(128.0, 128.0, 0.0);
		glTexCoord2f(0.0, 0.0);		glVertex3f(0.0, 128.0, 0.0);
	glEnd();
	glDisable(GL_TEXTURE_2D);
	glPopMatrix();
}
//***************************************************************************************************************************





// callback function for "similarity measure" listbox
void ListBoxCB(int id)
{
	int option = SMListBox->get_int_val();

	switch(option)
	{
	case 0:
		GRecognizer->setSimilarityMeasurement(EUCLIDEAN);
		break;
	case 1:
		GRecognizer->setSimilarityMeasurement(MODIFIED_EUCLIDEAN);
		break;
	case 2:
		GRecognizer->setSimilarityMeasurement(PEARSON_CORRELATION);
		break;
	case 3:
		GRecognizer->setSimilarityMeasurement(PEARSON_CORRELATION_SQUARED);
		break;
	case 4:
		GRecognizer->setSimilarityMeasurement(DTW);
		break;
	case 5:
		GRecognizer->setSimilarityMeasurement(FM);
		break;
	}
}

// mouse move function (for debugging purposes)
void PassiveMotion(int x, int y)
{
	//std::cout<<x<<"  "<<WindowHeight - y<<"\n";
	for (unsigned int i = 0; i < ButtonList.size(); i++)
	{
		ButtonList[i]->updateButtonState((float) x, (float) (WindowProps.WindowHeight - y));
	}

	for (unsigned int i = 0; i < MenuTrainingButtonList.size(); i++)
	{
		MenuTrainingButtonList[i]->updateButtonState((float) x, (float) (WindowProps.WindowHeight - y));
	}
}


// helper function, searches for button index in a particular list with the help of its id
int getButtonIndex(std::vector<CircleButton*>& list, std::string id)
{
	for (unsigned int i = 0; i < list.size(); i++)
	{
		if(list[i]->getButtonID() == id)
		{
			return i;
		}		
	}
	return -1;
}


// updates all the buttons states in a particular list
void updateButtons(std::vector<CircleButton*>& list, math::Vec2f& coords)
{
	// update buttons
	for (unsigned int i = 0; i < list.size(); i++)
	{
		list[i]->updateButtonState(coords.x, coords.y);
	}
}


// sets the training menu buttons to their correct states 
void initMenuTrainingButtons()
{

	// we set all the training buttons to visible but inactive
	for (unsigned int i = 0; i < MenuTrainingButtonList.size(); i++)
	{
		MenuTrainingButtonList[i]->updateButtonState(0.0f, 0.0f);
		MenuTrainingButtonList[i]->setVisible();
		MenuTrainingButtonList[i]->setInactive();
	}

	// we activate some of the buttons
	MenuTrainingButtonList[getButtonIndex(MenuTrainingButtonList, "BUTTON_NEW_SET")]->setActive();     
																				  
	if(GRecognizer->TrainedGestureSets.size() > 0) 													 
	{
		MenuTrainingButtonList[getButtonIndex(MenuTrainingButtonList, "BUTTON_REMOVE_SET")]->setActive();
		MenuTrainingButtonList[getButtonIndex(MenuTrainingButtonList, "BUTTON_PREV_SET")]->setActive();
		MenuTrainingButtonList[getButtonIndex(MenuTrainingButtonList, "BUTTON_NEXT_SET")]->setActive();
		MenuTrainingButtonList[getButtonIndex(MenuTrainingButtonList, "BUTTON_NEW_GEST")]->setActive();
		MenuTrainingButtonList[getButtonIndex(MenuTrainingButtonList, "BUTTON_REMOVE_GEST")]->setActive();
		MenuTrainingButtonList[getButtonIndex(MenuTrainingButtonList, "BUTTON_REPLACE_GEST")]->setActive();
		MenuTrainingButtonList[getButtonIndex(MenuTrainingButtonList, "BUTTON_PREV_GEST")]->setActive();
		MenuTrainingButtonList[getButtonIndex(MenuTrainingButtonList, "BUTTON_NEXT_GEST")]->setActive();
	}

	MenuTrainingButtonList[getButtonIndex(MenuTrainingButtonList, "BUTTON_EXIT_GEST")]->setActive();
}


// sets the textual informations about trained gesture sets in the training window
void setGestureMenuText()
{
	if(GRecognizer->TrainedGestureSets.size() > 0)
	{
		std::stringstream ss;
		ss<<"Gesture Set: "<<(GRecognizer->CurrentTrainedGestureSetIndex)+1<<" / "<<GRecognizer->TrainedGestureSets.size();
		ss<<" - Gesture: "<<(GRecognizer->CurrentTrainedGestureIndex)+1<<" / "<<GRecognizer->TrainedGestureSets[GRecognizer->CurrentTrainedGestureSetIndex]->Gestures.size();
		GestureType current_type = GRecognizer->TrainedGestureSets[GRecognizer->CurrentTrainedGestureSetIndex]->SideOfBody;
		ss<<" - "<<GestureUtility::GestureTypeToString(current_type);
		GestureMenuText->setLineOfText(ss.str(), 24);
		GestureMenuText->setTextColor(0,0,0,255);
		GestureMenuText->setDuration(-1);
		GestureMenuText->activate();
	}
}

// event handling for all the buttons
void HandleButtonEvent(std::string ev)
{
	//std::cout<<ev<<"\n";

	if (ev == "BUTTON_TRAINING")
	{
		// we change the app_mode to training menu
		app_mode = 1;
		ButtonList[getButtonIndex(ButtonList, "BUTTON_TRAINING")]->setInvisible();
		// set navigation buttons and menu text
		initMenuTrainingButtons();
		setGestureMenuText();
	}

	// switch gesture recognizer into training mode
	if( (ev == "BUTTON_NEW_SET") || (ev == "BUTTON_NEW_GEST") || (ev == "BUTTON_REPLACE_GEST"))
	{
		if( ev == "BUTTON_NEW_SET" )
		{
			GRecognizer->setTrainingMode(true);
		}
		if( ev == "BUTTON_NEW_GEST" )
		{
			GRecognizer->setTrainingMode();
		}
		if (ev == "BUTTON_REPLACE_GEST")
		{
		GRecognizer->setTrainingMode(false, true);
		}

		// we set all the training buttons to inactive
		for (unsigned int i = 0; i < MenuTrainingButtonList.size(); i++)
		{
			MenuTrainingButtonList[i]->setInactive();
		}

		// set the countdown timer
		Timer->start(4000);

		// set the flag to spot the incoming gesture data
		HandleGestures = true;
	}

	// remove a whole gesture set
	if (ev == "BUTTON_REMOVE_SET")
	{
		if( GRecognizer->removeCurrentTrainedGestureSet() == -1)
		{
			initMenuTrainingButtons();
		}

		setGestureMenuText();
		InfoText->setLineOfText("Gesture Set removed!", 30, true, 5);
		InfoText->activate();
	}

	// navigate through the current gestures and sets
	if (ev == "BUTTON_PREV_SET")
	{
		GRecognizer->setIndexToPreviousTrainedGestureSet();
		setGestureMenuText();
	}

	if (ev == "BUTTON_NEXT_SET")
	{
		GRecognizer->setIndexToNextTrainedGestureSet();
		setGestureMenuText();
	}

	if (ev == "BUTTON_PREV_GEST")
	{
		GRecognizer->setIndexToPreviousTrainedGesture();
		setGestureMenuText();
	}

	if (ev == "BUTTON_NEXT_GEST")
	{
		GRecognizer->setIndexToNextTrainedGesture();
		setGestureMenuText();
	}
	
	// remove a gesture from the current set
	if (ev == "BUTTON_REMOVE_GEST")
	{
		if( GRecognizer->removeCurrentTrainedGesture() == -1)
		{
			initMenuTrainingButtons();
		}
		setGestureMenuText();
		InfoText->setLineOfText("Gesture removed!", 30, true, 5);
		InfoText->activate();
	}

	// exit the training program
	if (ev == "BUTTON_EXIT_GEST")
	{
		app_mode = 0;
		
		// we set all the training buttons to invisible
		for (unsigned int i = 0; i < MenuTrainingButtonList.size(); i++)
		{
			MenuTrainingButtonList[i]->setInvisible();
		}

		// we activate the training button
		ButtonList[getButtonIndex(ButtonList, "BUTTON_TRAINING")]->setActive();
	}	
}


// handling for hand gesture events
void HandleHandEvent(std::string ev)
{
}


// event handling for recognized gestures
void HandleGestureEvent(std::string ev)
{
	// if we do not want to recognize gesture input (e.g. in tool mode)
	if(!HandleGestures)
	{
		// in the special case of the DEFAULT_GESTURE we reset the tool and go back to gesture mode
		if(ev == "DEFAULT_GESTURE")
		{
			ActiveTool = ToolList[0];
			HandleGestures = true;
		}
		return;
	}

	// prepare string for text module
	std::stringstream ss;
	if( ev== "DEFAULT_GESTURE")
	{
		ss<<"Default Gesture recognized";
	}
	else
	{
		// write gesture id to string stream
		// all the trained data has zero based gesture index numbers. so we increment these numbers by one.
		ss<<"Gesture "<<(atoi(ev.c_str()) + 1)<<" recognized";
	}

	// set the info text
	InfoText->setLineOfText(ss.str(), 30, true, 5);

	// gesture training events
	if(ev == "GESTURE_NOT_TRAINED")
	{
		initMenuTrainingButtons();
		InfoText->setLineOfText("Error! Gesture not trained!", 30, true, 5);
		setGestureMenuText();
	}

	if(ev == "GESTURE_TRAINED")
	{
		initMenuTrainingButtons();
		InfoText->setLineOfText("Success! Gesture trained!", 30, true, 5);
		setGestureMenuText();

		// auto save
		//GRecognizer->saveTrainedGestures("trained_gestures.bin");
		//COUNT--;
		//if (COUNT > 0)
		//	HandleButtonEvent("BUTTON_NEW_GEST");
	}

	InfoText->activate();


	// event handling for easter egg app
	if(ev == "0")
	{
		// camera tool
		ActiveTool = ToolList[1];
		if(app_mode == 2) 
			HandleGestures = false;		
	}


	if(ev == "1")
	{
		// color picker
		ActiveTool = ToolList[2];
		if(app_mode == 2) 
			HandleGestures = false;
	}

	if(ev == "2")
	{
		// brush tool
		ActiveTool = ToolList[3];
		if(app_mode == 2) 
			HandleGestures = false;
	}

	if(ev == "3")
	{
		// line tool
		ActiveTool = ToolList[4];
		if(app_mode == 2) 
			HandleGestures = false;
	}

	if(ev == "4")
	{
		// ink tool
		ActiveTool = ToolList[5];
		if(app_mode == 2) 
			HandleGestures = false;
	}

	if(ev == "5")
	{
		// rubber tool
		ActiveTool = ToolList[6];
		if(app_mode == 2) 
			HandleGestures = false;
	}

	if(ev == "6")
	{
		// eraser tool
		ActiveTool = ToolList[7];
		if(app_mode == 2) 
			HandleGestures = false;
	}

	if(ev == "7")
	{
		// undo
		if(app_mode == 2) 
			UndoRedoManager::undo();
	}

	if(ev == "8")
	{
		// redo
		if(app_mode == 2) 
			UndoRedoManager::redo();
	}

	
}


// glut callback function when changing the window's position or size
void reshapeWindow(int width, int height)
{
	int tx, ty, tw, th;
	glutSetWindow(MainWindowID);
	GLUI_Master.get_viewport_area( &tx, &ty, &tw, &th );
	WindowProps.WindowWidth = tw;
	WindowProps.WindowHeight = th;
	WindowProps.WindowAspect = ((float) WindowProps.WindowWidth) / ((float) WindowProps.WindowHeight);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glViewport(0, 0, WindowProps.WindowWidth, WindowProps.WindowHeight);
	gluPerspective(45, WindowProps.WindowAspect, 0.1, 100.0);
	glMatrixMode(GL_MODELVIEW);

	int index = getButtonIndex(ButtonList, "BUTTON_TRAINING");
	float radius = ButtonList[index]->getRadius();
	ButtonList[index]->setXPosition(1.2f * radius);
	ButtonList[index]->setYPosition((float) WindowProps.WindowHeight - 1.2f*radius);
}


// glut idle callback function, we update the recognition system's state ( glutPostReDisplay implicitely calls update() )
void idle()
{
	glutSetWindow(MainWindowID); // glui occasionally changes window handle
	glutPostRedisplay();
}


// we use the cursor keys up & down to change the sensor's angle in 2 degree steps, the use of ALT changes the increment steps to 10 degree
void handleSpecialKeyEvent(int key, int x, int y)
{
	int angle_inc = 0;
	
	if(key == GLUT_KEY_UP || key == GLUT_KEY_DOWN)
		angle_inc = (glutGetModifiers() == GLUT_ACTIVE_CTRL)?(10):(2);

	if(key == GLUT_KEY_DOWN)
		angle_inc *= -1;
	
	LONG angle = NuiSensor->getCameraAngle();
	NuiSensor->setCameraAngle(angle + angle_inc);	
}


// glut event handling for key events
void handleKeyEvent(unsigned char key, int x, int y)
{
	// only used for data mining, user can trigger the start or end of a gesture via spacebar
	//if ( write_velocities_to_file && (key == 32) )
	//{
	//	  of_velocites<<"\t100";
	//}

	// string stream and file for loading/saving training data
	std::stringstream ss;
	std::string file;
	// ONLY USED FOR DATA MINING
	//std::ofstream of;
	// ----------------------------

	switch(key)
	{
	// exit the program
	case 27:			//ESC

			// ONLY USED FOR DATA MINING --------------------------------------
			//RightHandRecognizer->stat_printStats();
			//GRecognizer->stat_printGestures();
			//stat_printPoses();

			/*of.open("hand_coords_smoothing", std::ios::out | std::ios::trunc);
			for(unsigned int i=0; i < RightHandCoordsWithoutSmoothing.size(); i++)
			{
				of<<RightHandCoordsWithoutSmoothing[i].x<<"\t"<<
					RightHandCoordsWithoutSmoothing[i].y<<"\t"<<
					RightHandCoordsWithSmoothing[i].x<<"\t"<<
					RightHandCoordsWithSmoothing[i].y<<"\n";
			}
			of.close();*/
		    // -------------------------------------------------------------
			
			exit(0);
			break;


	// save the current training data into a file
	case 's':
			// open a save dialog box
			file = FileHandling::save(".bin", WindowProps.WindowTitle);
			if(GRecognizer->saveTrainedGestures(file))
			{
				ss<<"Saved "<<GRecognizer->getSizeOfGestureSets()<<" gesture sets to disk...";
				InfoText->setLineOfText(ss.str(), 30, true, 5);
				InfoText->activate();
			}
			break;
		
	// load training data from a file
	case 'l':
			// open a open dialog box
			file = FileHandling::open("*.bin", WindowProps.WindowTitle);
			if(GRecognizer->loadTrainedGestures(file))
			{
				ss<<"Loaded "<<GRecognizer->getSizeOfGestureSets()<<" gesture sets from disk...";
				InfoText->setLineOfText(ss.str(), 30, true, 5);
				InfoText->activate();
				initMenuTrainingButtons();
			}
			break;



	// key events for easter egg app *********************************

	case 'p':
		// change app mode to easter egg app
		if(app_mode == 2)
		{
			app_mode = 0;
		}
		else 
			app_mode = 2;
		break;

	case '1':
		// camera
		ActiveTool = ToolList[1];
		break;
	
	case '2':
		// color chooser
		ActiveTool = ToolList[2];
		break;

	case '3':
		// brush tool
		ActiveTool = ToolList[3];
		break;

	case '4':
		// straight line tool
		ActiveTool = ToolList[4];
		break;

	case '5':
		// ink tool
		ActiveTool = ToolList[5];
		break;

	case '6':
		// rubber tool
		ActiveTool = ToolList[6];
		break;

	case '7':
		// eraser tool
		ActiveTool = ToolList[7];
		break;

	case '8':
		// undo
		UndoRedoManager::undo();
		break;

	case '9':
		// redo
		UndoRedoManager::redo();
		break;
	// *************************************************************************
	
	
	// events for empirical analysis of hand recognition params
	/*
	case 'e':
		RightHandRecognizer->DepthThreshold++;
		break;
	case 'r':
		RightHandRecognizer->DepthThreshold = max(RightHandRecognizer->DepthThreshold-1,0);
		break;
	case 't':
		RightHandRecognizer->SizeOfContourFilter++;
		break;
	case 'z':
		RightHandRecognizer->SizeOfContourFilter = max(RightHandRecognizer->SizeOfContourFilter-1,0);
		break;
	case 'd':
		RightHandRecognizer->SizeOfMaximaNeighbourHood++;
		break;
	case 'f':
		RightHandRecognizer->SizeOfMaximaNeighbourHood = max(RightHandRecognizer->SizeOfMaximaNeighbourHood-1,0);
		break;
	case 'g':
		RightHandRecognizer->ContourGradient += 0.05;
		break;
	case 'h':
		RightHandRecognizer->ContourGradient -= 0.05;
		break;

	case 'q':
		RightHandRecognizer->stat_setCapture(true, true);
		InfoText->setLineOfText("Capturing closed hand...", 30, true, 5);
		InfoText->activate();
		break;
	
	case 'w':
		RightHandRecognizer->stat_setCapture(true, false);
		InfoText->setLineOfText("Capturing open hand...", 30, true, 5);
		InfoText->activate();
		break;


	case 'a':
		std::cout<<"Stopped Capturing\n";
		RightHandRecognizer->stat_setCapture(false, true);
		InfoText->setLineOfText("Stopped capturing hand...", 30, true, 5);
		InfoText->activate();
		break;

	case 'm':
		// toggle the pointer's MinMaxFlag
		LeftHandPointer->setMinMax(!(LeftHandPointer->getMinMax()));
		RightHandPointer->setMinMax(!(RightHandPointer->getMinMax()));

	case 'c':
			// print hand contour to file
			of.open("hand_contour.txt", std::ios::out | std::ios::trunc);
			of<<"Contour values:\n";
			for(unsigned int i=0; i < RightHandRecognizer->getContourPixelsDistances().size(); i++)
			{
				of<<RightHandRecognizer->getContourPixelsDistances()[i]<<"\n";
			}
			of<<"Finger Positions:\n";
			for(unsigned int i=0; i < RightHandRecognizer->getMaximaOfContourPixelsDistances().size(); i++)
			{
				of<<RightHandRecognizer->getMaximaOfContourPixelsDistances()[i]<<"\n";
			}
			of.close();
			break;

	case 'y':
			HandleButtonEvent("BUTTON_NEW_GEST");
			break;

	case 'x':
			HandleButtonEvent("BUTTON_NEW_SET");
			break;
	*/
	}
}


// helper function for skeleton rendering
void drawJointAt(NUI_SKELETON_DATA skeleton, int joint_nr)
{					
	float x = skeleton.SkeletonPositions[joint_nr].x;
	float y = skeleton.SkeletonPositions[joint_nr].y;
	float z = skeleton.SkeletonPositions[joint_nr].z;
	glPushMatrix();
	glTranslatef(x, y, z);
	GLUquadric *quadric = gluNewQuadric();
	gluSphere(quadric, 0.04, 16, 16);	
	glPopMatrix();

	// free memory
	gluDeleteQuadric(quadric);

}


// helper function for skeleton rendering
void VertexAt(NUI_SKELETON_DATA skeleton, int joint_nr)
{					
	float x = skeleton.SkeletonPositions[joint_nr].x;
	float y = skeleton.SkeletonPositions[joint_nr].y;
	float z = skeleton.SkeletonPositions[joint_nr].z;
	glVertex3f(x, y, z);
	
}


// draw the user skeleton
void DrawSkeletonData(const NUI_SKELETON_DATA& skeleton)
{		
	glDisable(GL_DEPTH_TEST);
	glPushMatrix();
	
	//render bones
	glColor4ub(0, 170, 200, 255);
	glLineWidth(10.0f);
	glBegin(GL_LINE_STRIP);
	for (int i = 0; i < 4; i++)
		VertexAt(skeleton, i);
	glEnd();

	glBegin(GL_LINE_STRIP);
	VertexAt(skeleton, 2);
	for (int i = 4; i < 8; i++)
		VertexAt(skeleton, i);
	glEnd();

	glBegin(GL_LINE_STRIP);
	VertexAt(skeleton, 2);
	for (int i = 8; i < 12; i++)
		VertexAt(skeleton, i);
	glEnd();

	glBegin(GL_LINE_STRIP);
	VertexAt(skeleton, 0);
	for (int i = 12; i < 16; i++)
		VertexAt(skeleton, i);
	glEnd();

	glBegin(GL_LINE_STRIP);
	VertexAt(skeleton, 0);
	for (int i = 16; i < 20; i++)
		VertexAt(skeleton, i);
	glEnd();


	// render joints
	glColor4ub(0, 104, 170, 255);
	for (int i = 0; i < 20; i++)
	{
		drawJointAt(skeleton, i);
	}

	glPopMatrix();
	glEnable(GL_DEPTH_TEST);

}


// render all the kinect skeleton positions into the depth map
void renderSkeletonPointsIntoTexture(GLuint texID)
{
	glBindTexture(GL_TEXTURE_2D, texID);

	NUI_SKELETON_DATA skeleton = NuiSensor->getCurrentSkeletonData();

	BYTE marker[] = { 255, 255, 255, 255, 255,
					  255, 255, 255, 255, 255,
					  255, 255, 255, 255, 255,
					  255, 255, 255, 255, 255,
					  255, 255, 255, 255, 255
					};
	// get positions of joints in the depth map
	for (int i=0; i<20; i++)
	{
		long x_coord = 0, y_coord = 0;
		unsigned short depth;
		NuiTransformSkeletonToDepthImage(skeleton.SkeletonPositions[i], &x_coord, &y_coord, &depth, NUI_IMAGE_RESOLUTION_640x480);

		glTexSubImage2D(GL_TEXTURE_2D, 0, x_coord-2, y_coord-2, 4, 4, GL_BLUE, GL_UNSIGNED_BYTE, marker);
	}	
}


// render kinect video data or depth data into a given texture
void renderKinectFrameIntoTexture(int x_offset, int y_offset, GLuint texID, bool depthdata)
{
	glBindTexture(GL_TEXTURE_2D, texID);

	// in depth data mode we also render the skeleton joints as markers
	if (depthdata)
	{
		glTexSubImage2D(GL_TEXTURE_2D, 0, x_offset, y_offset, 640, 480, GL_BGRA_EXT, GL_UNSIGNED_BYTE, NuiSensor->getCurrentDepthFrame());
		// superimpose skeleton points 
		renderSkeletonPointsIntoTexture(texID);
	}
	else
	{
		glTexSubImage2D(GL_TEXTURE_2D, 0, x_offset, y_offset, 640, 480, GL_BGRA_EXT, GL_UNSIGNED_BYTE, NuiSensor->getCurrentVideoFrame());
	}	
}


// render the images of the hands into a texture
void renderHandImagesIntoTexture(GLuint texID)
{
	HandImage* hi_left = LeftHandRecognizer->getLastHandImageData();
	HandImage* hi_right = RightHandRecognizer->getLastHandImageData();

	glBindTexture(GL_TEXTURE_2D, texID);
	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, hi_left->width, hi_left->height, GL_BGRA_EXT, GL_UNSIGNED_BYTE, hi_left->imagedata);
	glTexSubImage2D(GL_TEXTURE_2D, 0, 640-hi_right->width, 0, hi_right->width, hi_right->height, GL_BGRA_EXT, GL_UNSIGNED_BYTE, hi_right->imagedata);
}


// draw kinect video or depth data with the help of a textured quad ( ensure that you have rendered the data into the texture before [renderKinectFrameIntoTexture()] )
void drawVideoFrame(float alpha)
{
	glPushMatrix(); //save modelview transformations
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix(); //save projection transformations
	glLoadIdentity();
	glOrtho(-1.0,1.0,-1.0,1.0,-1.0,1000.0);

	glColor4f(1.0f,1.0f,1.0f, alpha);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, VideoTexID);

	// calculate the extends of the area (of the 1024x1024 texture) where we can find the data (640x480)
	float coordx = 640.0f/1024.0f;
	float coordy = 480.0f/1024.0f;

	// set width and height of textured quad
	float height = 2.0;
	float width = height*1.333f/WindowProps.WindowAspect;

	glBegin(GL_QUADS);
		glTexCoord2f(0.0f, coordy);			glVertex3f(1.0f - width, 1.0f - height, -1000.0f);
		glTexCoord2f(coordx, coordy);		glVertex3f(1.0f, 1.0f - height, -1000.0f);
		glTexCoord2f(coordx, 0.0f);			glVertex3f(1.0f, 1.0f, -1000.0f);
		glTexCoord2f(0.0f, 0.0f);			glVertex3f(1.0f - width, 1.0f, -1000.0f);
	glEnd();

	glDisable(GL_TEXTURE_2D);

	glPopMatrix(); //restore projection transformations
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix(); //restore modelview transformations
}


// draws a histogram of the body velocities as a time serie curve 
void drawVelocityChart()
{
	glDisable(GL_DEPTH_TEST);
	glViewport(0,0,WindowProps.WindowWidth/2, WindowProps.WindowHeight/4);
	// render to left bottom side of window
	glPushMatrix(); //save modelview transformations
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix(); //save projection transformations
	glLoadIdentity();
	glOrtho(0.0, 1.0, 0.0, 1.0, -1.0, 1.0);

	float velocity_low_threshold = GRecognizer->getVelocityLowThreshold();
	float velocity_high_threshold = GRecognizer->getVelocityHighThreshold();
	int interval_length = GRecognizer->getMaxLengthOfGestures();
	float max_threshold = max(velocity_high_threshold, velocity_low_threshold);
	
	glPushMatrix();
	glScalef(1.0f/((float)interval_length), 0.7f/max_threshold, 1.0f); 

	glLineWidth(2.0f);

	int woving_window_min = interval_length;
	int woving_window_max = 0;
	// render candidate frames
	glColor3f(0.0f, 0.0f, 1.0f);
	glBegin(GL_LINES);
		for ( int i = 0; i < interval_length; i++)
		{
			if (GRecognizer->RecognitionList[i].flag)
			{
				glVertex3f( (float) i, 0.0f, 0.0f);
				glVertex3f( (float) i, 1.0f/0.7f * max_threshold, 0.0f);

				woving_window_min = min(woving_window_min, i);
				woving_window_max = max(woving_window_max, i);
			}
		}
	glEnd();

	// render moving window
	glColor4f(0.0f, 0.0f, 1.0f, 0.2f);
	glBegin(GL_QUADS);
		glVertex3f((float) woving_window_min, 0.0f, 0.0f);
		glVertex3f((float) woving_window_max, 0.0, 0.0f);
		glVertex3f((float) woving_window_max, 1.0f/0.7f * max_threshold, 0.0f);
		glVertex3f((float) woving_window_min, 1.0f/0.7f * max_threshold, 0.0f);
	glEnd();

	// render velocities
	glColor3f(1.0f, 0.0f, 0.0f);
	glBegin(GL_LINE_STRIP);
		for ( int i = 0; i < interval_length; i++)
		{
			glVertex3f( (float) i, GRecognizer->RecognitionList[i].velocity, 0.0f);
		}
	glEnd();

	
	// render low threshold
	glColor3ub(60,60,60);
	glBegin(GL_LINES);
		glVertex3f( 0.0f, velocity_low_threshold, 0.0f);
		glVertex3f( (float)interval_length, velocity_low_threshold, 0.0f);
	glEnd();
	// render high threshold
	glColor3ub(128,128,128);
	glBegin(GL_LINES);
		glVertex3f( 0.0f, velocity_high_threshold, 0.0f);
		glVertex3f( (float)interval_length, velocity_high_threshold, 0.0f);
	glEnd();
		

	glPopMatrix();
	// render frame
	glColor3ub(60,60,60);
	glBegin(GL_LINE_STRIP);
		glVertex3f(0.0f, 0.0f, 0.0f);
		glVertex3f(1.0f, 0.0, 0.0f);
		glVertex3f(1.0f, 1.0f, 0.0f);
		glVertex3f(0.0f, 1.0f, 0.0f);
		glVertex3f(0.0f, 0.0f, 0.0f);
	glEnd();
	
	glPopMatrix(); //restore projection transformations
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix(); //restore modelview transformations

	glEnable(GL_DEPTH_TEST);
	glViewport(0,0,WindowProps.WindowWidth, WindowProps.WindowHeight);

}

// draws the histogram of the hand contour + recognized fingers
void drawContourPixelDistancesChart(HandRecognizer* hr)
{
	glDisable(GL_DEPTH_TEST);
	glViewport((int) (3.0f/4.0f * WindowProps.WindowWidth), 0, (int) (WindowProps.WindowWidth/4.0), (int) (WindowProps.WindowHeight/4.0));
	// render to right bottom side of window
	glPushMatrix(); //save modelview transformations
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix(); //save projection transformations
	glLoadIdentity();
	glOrtho(0.0, 1.0, 0.0, 1.0, -1.0, 1.0);

	std::vector<float> &distances = hr->getContourPixelsDistances();
	float max_dist = 0.0f;
	for (unsigned int i = 0; i < distances.size(); i++)
	{
		max_dist = max(max_dist, distances[i]);
	}

	// we scale in respect to maximum hand pixel distance
	glScalef(1.0f/((float) distances.size()), 1.0f/max_dist, 1.0f); 

	glLineWidth(3.0f);

	// render mean
	glColor3ub(60,60,60);
	glBegin(GL_LINES);
	glVertex3f( 0.0f, hr->getContourPixelsDistancesMean(), 0.0f);
		glVertex3f( (float)distances.size(), hr->getContourPixelsDistancesMean(), 0.0f);
	glEnd();

	// get the the maxima positions
	std::vector<int> &maxima = hr->getMaximaOfContourPixelsDistances();
	int size_of_neighbourhood = hr->getSizeOfMaximaNeighbourHood();

	// render maxima and neighbourhood
	for (unsigned int i=0; i < maxima.size(); i++)
	{
		glColor4ub(255,255,255,120);
		glBegin(GL_QUADS);
			glVertex3f( (float) (maxima[i]-size_of_neighbourhood), 0.0f, 0.0f);
			glVertex3f( (float) (maxima[i]+size_of_neighbourhood), 0.0f, 0.0f);
			glVertex3f( (float) (maxima[i]+size_of_neighbourhood), distances[maxima[i]], 0.0f);
			glVertex3f( (float) (maxima[i]-size_of_neighbourhood), distances[maxima[i]], 0.0f);
		glEnd();

		glColor3ub(255,0,0);
		glBegin(GL_LINES);
			glVertex3f( (float) (maxima[i]), 0.0f, 0.0f);
			glVertex3f( (float) (maxima[i]), distances[maxima[i]], 0.0f);
		glEnd();
	}

	// render distances
	glColor3ub(60,60,60);
	glBegin(GL_LINE_STRIP);
		for (unsigned int i = 0; i < distances.size(); i++)
		{
			glVertex3f( (float) i, distances[i], 0.0f);
		}
	glEnd();
	
	glPopMatrix(); //restore projection transformations
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix(); //restore modelview transformations

	glEnable(GL_DEPTH_TEST);
	glViewport(0,0,WindowProps.WindowWidth, WindowProps.WindowHeight);
}


// draws the gui of the training window
void drawMenuGestureTraining(NUI_SKELETON_DATA &user_skeleton)
{
	glDisable(GL_DEPTH_TEST);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix(); //save projection transformations
	glLoadIdentity();
	glOrtho(0.0, WindowProps.WindowWidth, 0.0, WindowProps.WindowHeight,-1000.0, 1000.0);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix(); //save modelview transformations
	glLoadIdentity();

	// we show a window
	float margin = 0.05f;
	glColor4ub(240,255,255,255);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, TrainingBGTexID);
	glBegin(GL_QUADS);
	glTexCoord2f(0.0f, 1.0f);	
	glVertex2f(margin*WindowProps.WindowWidth, margin*WindowProps.WindowHeight);
	glTexCoord2f(1.0f, 1.0f);	
	glVertex2f(WindowProps.WindowWidth-margin*WindowProps.WindowWidth, margin*WindowProps.WindowHeight);
	glTexCoord2f(1.0f, 0.0f);	
	glVertex2f(WindowProps.WindowWidth-margin*WindowProps.WindowWidth, WindowProps.WindowHeight-margin*WindowProps.WindowHeight);
	glTexCoord2f(0.0f, 0.0f);
	glVertex2f(margin*WindowProps.WindowWidth, WindowProps.WindowHeight-margin*WindowProps.WindowHeight);
	glEnd();
	glDisable(GL_TEXTURE_2D);

	glColor4ub(0,0,0,255);
	glLineWidth(2.0f);
	glBegin(GL_LINE_STRIP);
	glVertex2f(margin*WindowProps.WindowWidth, margin*WindowProps.WindowHeight);
	glVertex2f(WindowProps.WindowWidth-margin*WindowProps.WindowWidth, margin*WindowProps.WindowHeight);
	glVertex2f(WindowProps.WindowWidth-margin*WindowProps.WindowWidth, WindowProps.WindowHeight-margin*WindowProps.WindowHeight);
	glVertex2f(margin*WindowProps.WindowWidth, WindowProps.WindowHeight-margin*WindowProps.WindowHeight);
	glVertex2f(margin*WindowProps.WindowWidth, margin*WindowProps.WindowHeight);
	glEnd();

	// set the button positions on the right...
	float radius = MenuTrainingButtonList[0]->getRadius();
	for (int i = 0; i <= 5; i++)
	{
		MenuTrainingButtonList[i]->setPosition(WindowProps.WindowWidth-margin*WindowProps.WindowWidth-1.2f*radius*(i*2+1), WindowProps.WindowHeight-margin*WindowProps.WindowHeight-1.2f*radius);
	}

	//... and on the left window side
	for (int i = 6; i <= 9; i++)
	{
		MenuTrainingButtonList[i]->setPosition(margin*WindowProps.WindowWidth+1.2f*radius*((i-6)*2+1), WindowProps.WindowHeight-margin*WindowProps.WindowHeight-1.2f*radius);
	}

	// update Buttons for right hand (right hand pointer is preferred)
	if(RightHandPointer->isActive())
	{
		math::Vec2f coords = RightHandPointer->getWindowCoordinates();
		// update buttons
		updateButtons(MenuTrainingButtonList, coords);
	}
	else
	{
		// update Buttons for left hand
		if(LeftHandPointer->isActive())
		{
			math::Vec2f coords = LeftHandPointer->getWindowCoordinates();
			// update buttons
			updateButtons(MenuTrainingButtonList, coords);
		}
	}

	// draw the menu buttons
	for (unsigned int i = 0; i < MenuTrainingButtonList.size(); i++)
	{
		MenuTrainingButtonList[i]->drawButton();
	}

	// if we are in training mode we show the user skeleton...
	glPushMatrix();
	glTranslatef(((float) WindowProps.WindowWidth) * 0.5f, ((float) WindowProps.WindowHeight) * 0.5f, 0.0f);
	glScalef(300.0f, 300.0f, 300.0f);
	if ( GRecognizer->isTrainingMode() )
	{
		DrawSkeletonData(user_skeleton);
	}
	else
	{
		// ...else we show skeleton animation for trained gestures (if there are any)
		if ( GRecognizer->TrainedGestureSets.size() > 0)
		{		
			NUI_SKELETON_DATA skeleton = GRecognizer->getCurrentTrainedGesture()->getSkeletonAnimation()->getNextFrame();
			DrawSkeletonData(skeleton);
		}
	}
	glPopMatrix();

	// we show the gesture info text
	GestureMenuText->setPosition((unsigned int) (1.2f*margin*WindowProps.WindowWidth), (unsigned int) (1.2*margin*WindowProps.WindowHeight) );
	GestureMenuText->render();

	glPopMatrix(); //restore modelview transformations
	glMatrixMode(GL_PROJECTION);
	glPopMatrix(); //restore projection transformations
	glMatrixMode(GL_MODELVIEW);
	glEnable(GL_DEPTH_TEST);
}


// MAIN PROGRAM LOOP
void update(void)	
{
	//check if button events occurred
	while (CircleButton::isEventWaiting())
	{
		HandleButtonEvent(CircleButton::getLastEvent());
	}

	//check if hand events occurred
	while (HandRecognizer::isEventWaiting())
	{
		HandleHandEvent(HandRecognizer::getLastEvent());
	}

	//check if gesture events occurred
	while (GRecognizer->isEventWaiting())
	{
		HandleGestureEvent(GRecognizer->getLastEvent());
	}

	// get new skeleton frame from kinect
	NUI_SKELETON_DATA& skeleton = NuiSensor->updateSkeletonData();

	//DATA MINING
	//RightHandCoordsWithoutSmoothing.push_back(math::Vec2f(skeleton.SkeletonPositions[10].x, skeleton.SkeletonPositions[10].y));
	
	// update video and depth frames
	NuiSensor->updateVideoFrameData();
	NuiSensor->updateDepthFrameData();

	// recognize and update current pose, we set "spot gesture" to false if we are waiting for a timer or if we are in "live mode"
	if ( Timer->hasElapsed() && HandleGestures)
	{
		CurrentPose = GRecognizer->update(skeleton, true);
	}
	else
	{
		CurrentPose = GRecognizer->update(skeleton, false);
	}
	PoseDifference = GRecognizer->getCurrentVelocity();

	// we push the current pose onto the POSES stack for statistical reasons
	POSES.push_back(CurrentPose);
	
	// try to recognize hand input
	LeftHandRecognizer->update();
	RightHandRecognizer->update();
	
	// update pointer
	RightHandPointer->update(CurrentPose);
	LeftHandPointer->update(CurrentPose);

	//DATA MINING
	//RightHandCoordsWithSmoothing.push_back(RightHandPointer->getWindowCoordinates());


	//update tools
	if(app_mode == 2)
		updateActiveTool(CurrentPose);

	
	//DATA MINING
	/*if (write_velocities_to_file)
	{
		of_velocites<<"\n"<<GRecognizer->CURRENT_VELOCITY;
		FRAME_COUNT++;
	}*/


	// render camera capture into video texture
	renderKinectFrameIntoTexture(0, 0, VideoTexID, true);
	// render hand data into video texture
	renderHandImagesIntoTexture(VideoTexID);

	// DRAW!
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// render video texture
	if ( app_mode == 0)
	{
		drawVideoFrame(0.5f);
		// draw activity chart
		drawVelocityChart();

		// draw hand distances chart
		drawContourPixelDistancesChart(RightHandRecognizer);
	}

	if(app_mode == 1)
	{
		// gesture show room overlay
		drawMenuGestureTraining(skeleton);		
	}

	//DrawSkeletonData(skeleton);

	if (!(Timer->hasElapsed()))
	{
		// if there's an active countdown we draw it
		Timer->draw(WindowProps.WindowAspect);
	}
	
	// 2D OVERLAYS
	glDisable(GL_DEPTH_TEST);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix(); //save projection transformations
	glLoadIdentity();
	glOrtho(0.0, WindowProps.WindowWidth, 0.0, WindowProps.WindowHeight,-1.0,1.0);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix(); //save modelview transformations
	glLoadIdentity();
	
	//*******************************************************
	//EASTER EGG APP
	if ( app_mode == 2)
	{
		// render painting
		paint_drawPainting();
		// render active tool
		ActiveTool->render();
	}
	//********************************************************

	// render buttons
	if(app_mode == 0)
	{
		glDisable(GL_DEPTH_TEST);
		for (unsigned int i = 0; i < ButtonList.size(); i++)
		{
			ButtonList[i]->drawButton();
		}
	}
	
	// render Pointers
	bool left_hand_active = LeftHandPointer->drawPointer(LeftHandRecognizer->getHandState() == H_CLOSED);
	bool right_hand_active = RightHandPointer->drawPointer(RightHandRecognizer->getHandState() == H_CLOSED);

	// update Buttons
	if(right_hand_active)
	{
		math::Vec2f coords = RightHandPointer->getWindowCoordinates();
		if(app_mode == 0) updateButtons(ButtonList, coords);
	}
	else 
	{
		if(left_hand_active)
		{
			math::Vec2f coords = LeftHandPointer->getWindowCoordinates();
			if(app_mode == 0) updateButtons(ButtonList, coords);
		}
	}


	// render info text
	InfoText->render();
	// render recognizer state info
	RecognizerStateText->render();

	glEnable(GL_DEPTH_TEST);
	glPopMatrix(); //restore modelview transformations
	glMatrixMode(GL_PROJECTION);
	glPopMatrix(); //restore projection transformations
	glMatrixMode(GL_MODELVIEW);

	glutSwapBuffers();

	// sync all GLUI elements
	GLUI_Master.sync_live_all();
}


int main(int argc, char** argv)
{
	WindowProps.WindowAspect = 1.6f;
	WindowProps.WindowHeight = 1000;
	WindowProps.WindowWidth =  1600;
	WindowProps.WindowTitle = "*** Gesture Recognition via MICROSOFT KINECT ***";
	
	//init GLUT & Window
	glutInit(&argc, argv);
	glutInitWindowPosition(0, 0);
	glutInitWindowSize(WindowProps.WindowWidth, WindowProps.WindowHeight);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	MainWindowID = glutCreateWindow(WindowProps.WindowTitle);
	
	WindowProps.WindowWidth = glutGet(GLUT_WINDOW_WIDTH);
	WindowProps.WindowHeight = glutGet(GLUT_WINDOW_HEIGHT);
	WindowProps.WindowAspect = ((float) WindowProps.WindowWidth) / ((float) WindowProps.WindowHeight);

	// register glut and glui callbacks
	glutDisplayFunc(update);
	GLUI_Master.set_glutIdleFunc(idle);
	GLUI_Master.set_glutReshapeFunc( reshapeWindow );
	GLUI_Master.set_glutKeyboardFunc( handleKeyEvent );
	GLUI_Master.set_glutSpecialFunc( handleSpecialKeyEvent );

	// set up kinect sensor and recognizers
	NuiSensor = new KinectSensor();
	GRecognizer = new GestureRecognizer(NuiSensor);
	LeftHandRecognizer = new HandRecognizer(NuiSensor, H_LEFT);
	RightHandRecognizer = new HandRecognizer(NuiSensor, H_RIGHT);
	LeftHandPointer = new KinectPointer(NuiSensor, P_LEFT, &WindowProps);
	LeftHandPointer->setPointerColor(0,0,255,255);
	RightHandPointer = new KinectPointer(NuiSensor, P_RIGHT, &WindowProps);
	RightHandPointer->setPointerColor(255,0,0,255);

	// load gesture sets from command line
	if (argc == 2)
	{
		// if there's one command line argument we interpret that as a filename
		GRecognizer->loadTrainedGestures(std::string(argv[1]));
	}

	// text renderer for gesture menu
	GestureMenuText = new TextureTextRenderer();

	// text render for recognition infos
	InfoText = new TextureTextRenderer();
	InfoText->setDuration(1500);
	InfoText->setPosition(0,0);
	InfoText->setTextColor(0,0,0,255);

	// text renderer for current recognizer state
	RecognizerStateText = new TextureTextRenderer();
	RecognizerStateText->setDuration(-1);
	RecognizerStateText->setPosition(WindowProps.WindowWidth/2,0);
	RecognizerStateText->setTextColor(0,0,0,255);

	// countdown timer 
	Timer = new Countdown();

	// let's create a subwindow
	GLUI *subwindow = GLUI_Master.create_glui_subwindow(MainWindowID,GLUI_SUBWINDOW_RIGHT);
	
	//subwindow->add_edittext("Hand x:", GLUI_EDITTEXT_FLOAT, &(RightHandPointer._x);
	//subwindow->add_edittext("Hand y:", GLUI_EDITTEXT_FLOAT, &RightHandPointer._y);
	//subwindow->add_edittext("Hand xmin:", GLUI_EDITTEXT_FLOAT, &RightHandPointer._xmin);
	//subwindow->add_edittext("Hand xmax:", GLUI_EDITTEXT_FLOAT, &RightHandPointer._xmax);
	//subwindow->add_edittext("Hand ymin:", GLUI_EDITTEXT_FLOAT, &RightHandPointer._ymin);
	//subwindow->add_edittext("Hand ymax:", GLUI_EDITTEXT_FLOAT, &RightHandPointer._ymax);
	//subwindow->add_separator();

	subwindow->add_spinner("Velocity Low Threshold:", GLUI_SPINNER_FLOAT, &(GRecognizer->VelocityLowThreshold));
	subwindow->add_spinner("Velocity High Threshold:", GLUI_SPINNER_FLOAT, &(GRecognizer->VelocityHighThreshold));
	subwindow->add_spinner("Max size of moving window:", GLUI_SPINNER_INT, &(GRecognizer->RecognitionWindowSize));
	subwindow->add_spinner("Min recognition interval length:", GLUI_SPINNER_INT, &(GRecognizer->MinimumRecognitionIntervalLength));
	subwindow->add_spinner("Pose Recogn. Threshold:", GLUI_SPINNER_FLOAT, &(GRecognizer->PoseThreshold));
	subwindow->add_spinner("Default Gesture Threshold:", GLUI_SPINNER_FLOAT, &(GRecognizer->DefaultGestureDistanceThreshold));

	SMListBox = subwindow->add_listbox("Similarity Measure:",NULL,1,ListBoxCB);
	SMListBox->add_item(0,"Euclidean");
	SMListBox->add_item(1,"mod. Eudlidean");
	SMListBox->add_item(2,"Pearson Corr.");
	SMListBox->add_item(3,"sqr. Pearson");
	SMListBox->add_item(4,"DTW");
	SMListBox->add_item(5,"FMDistance");
	SMListBox->set_int_val(4);

	subwindow->add_spinner("Gesture Recogn. Threshold:", GLUI_SPINNER_FLOAT, &(GRecognizer->RecognitionThreshold));
	subwindow->add_spinner("Nr. Of. Neighbours:", GLUI_SPINNER_INT, &(GRecognizer->NrOfNeighbours));

	subwindow->add_edittext("Current velocity:", GLUI_EDITTEXT_FLOAT, &PoseDifference);
	subwindow->add_edittext("Current size of moving window:", GLUI_EDITTEXT_INT, &(GRecognizer->CURRENT_SIZE_OF_MOVING_WINDOW));

	subwindow->add_separator();

	subwindow->add_edittext("Elbow Left:", GLUI_EDITTEXT_FLOAT, &CurrentPose.AngleElbowLeft);
	subwindow->add_edittext("Elbow Right:", GLUI_EDITTEXT_FLOAT, &CurrentPose.AngleElbowRight);
	subwindow->add_edittext("Elevation Left:", GLUI_EDITTEXT_FLOAT, &CurrentPose.AngleElevationArmLeft);
	subwindow->add_edittext("Elevation Right:", GLUI_EDITTEXT_FLOAT, &CurrentPose.AngleElevationArmRight);
	subwindow->add_edittext("Hor. Add. Left:", GLUI_EDITTEXT_FLOAT, &CurrentPose.AngleHorizontalAdductionArmLeft);
	subwindow->add_edittext("Hor. Add. Right:", GLUI_EDITTEXT_FLOAT, &CurrentPose.AngleHorizontalAdductionArmRight);
	subwindow->add_edittext("Rotation Left:", GLUI_EDITTEXT_FLOAT, &CurrentPose.AngleRotationUpperArmLeft);
	subwindow->add_edittext("Rotation Right:", GLUI_EDITTEXT_FLOAT, &CurrentPose.AngleRotationUpperArmRight);
	//subwindow->add_edittext("Angle Coronal Hand2Hand:", GLUI_EDITTEXT_FLOAT, &CurrentPose.AngleCoronalHandToHandVector);
	//subwindow->add_edittext("Angle Transversal Hand2Hand:", GLUI_EDITTEXT_FLOAT, &CurrentPose.AngleTransversalHandToHandVector);
	//subwindow->add_edittext("Distance Hand2Hand:", GLUI_EDITTEXT_FLOAT, &CurrentPose.DistanceHandToHand);
	//subwindow->add_edittext("Distance Hands2CoronalPLane:", GLUI_EDITTEXT_FLOAT, &CurrentPose.DistanceCenterHandsToCoronalPlane);
	
	subwindow->add_separator();

	subwindow->add_spinner("Hand depth treshold:", GLUI_SPINNER_INT, &(RightHandRecognizer->DepthThreshold));
	subwindow->add_spinner("Hand contour filter size:", GLUI_SPINNER_INT, &(RightHandRecognizer->SizeOfContourFilter));
	subwindow->add_spinner("Hand contour neighbourhood size:", GLUI_SPINNER_INT, &(RightHandRecognizer->SizeOfMaximaNeighbourHood));
	subwindow->add_spinner("Hand contour gradient:", GLUI_SPINNER_FLOAT, &(RightHandRecognizer->ContourGradient));

	// set main window
	subwindow->set_main_gfx_window(MainWindowID);
	
	// training button
	CircleButton* training_button = new CircleButton("BUTTON_TRAINING", "../images/train_button.png", WindowProps.WindowWidth - 0.25f*WindowProps.WindowWidth, WindowProps.WindowHeight - 0.25f*WindowProps.WindowHeight, 64.0f, 0, 210, 255); 
	training_button->setActive();
	ButtonList.push_back(training_button);

	// menu buttons
	CircleButton* new_gesture_button = new CircleButton("BUTTON_NEW_GEST", "../images/add_button.png", 0, 0, 32, 0, 210, 255); 
	CircleButton* remove_gesture_button = new CircleButton("BUTTON_REMOVE_GEST", "../images/remove_button.png", 0, 0, 32, 0, 210, 255); 
	CircleButton* replace_gesture_button = new CircleButton("BUTTON_REPLACE_GEST", "../images/replace_button.png", 0, 0, 32, 0, 210, 255); 
	CircleButton* prev_gesture_button = new CircleButton("BUTTON_PREV_GEST", "../images/prev_button.png", 0, 0, 32, 0, 210, 255); 
	CircleButton* next_gesture_button = new CircleButton("BUTTON_NEXT_GEST", "../images/next_button.png", 0, 0, 32, 0, 210, 255); 
	CircleButton* exit_training_button = new CircleButton("BUTTON_EXIT_GEST", "../images/exit_button.png", 0, 0, 32, 0, 210, 255); 

	CircleButton* new_gesture_set_button = new CircleButton("BUTTON_NEW_SET", "../images/new_button.png", 0, 0, 32, 0, 210, 255); 
	CircleButton* remove_gesture_set_button = new CircleButton("BUTTON_REMOVE_SET", "../images/remove_button.png", 0, 0, 32, 0, 210, 255); 
	CircleButton* prev_gesture_set_button = new CircleButton("BUTTON_PREV_SET", "../images/prev2_button.png", 0, 0, 32, 0, 210, 255); 
	CircleButton* next_gesture_set_button = new CircleButton("BUTTON_NEXT_SET", "../images/next2_button.png", 0, 0, 32, 0, 210, 255); 

	// push menu buttons into a list
	MenuTrainingButtonList.push_back(exit_training_button);
	MenuTrainingButtonList.push_back(next_gesture_button);
	MenuTrainingButtonList.push_back(prev_gesture_button);
	MenuTrainingButtonList.push_back(replace_gesture_button);
	MenuTrainingButtonList.push_back(remove_gesture_button);
	MenuTrainingButtonList.push_back(new_gesture_button);
	MenuTrainingButtonList.push_back(new_gesture_set_button);
	MenuTrainingButtonList.push_back(remove_gesture_set_button);
	MenuTrainingButtonList.push_back(prev_gesture_set_button);
	MenuTrainingButtonList.push_back(next_gesture_set_button);

	// dummy texture for Video Frame
	glGenTextures(1, &VideoTexID);
	int size = 1024;
	char * texturedata = new char[size*size*4];
	for (int i=0; i<size*size*4; i++)
	{
		texturedata[i] = 0;		
	}
	glBindTexture(GL_TEXTURE_2D, VideoTexID);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, size, size, 0, GL_BGRA_EXT, GL_UNSIGNED_BYTE, texturedata);

	delete[] texturedata;

	// load texture for training menu background
	TrainingBGTexID = SOIL_load_OGL_texture("../images/training_bg.png", SOIL_LOAD_AUTO, SOIL_CREATE_NEW_ID, SOIL_FLAG_MIPMAPS);


	// init opengl stuff
	glEnable(GL_DEPTH_TEST);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glViewport(0,0,WindowProps.WindowWidth,WindowProps.WindowHeight);
	gluPerspective(45, WindowProps.WindowAspect, 0.1, 1000.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt(0.0, 0.0, 5.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
	

	glClearColor(1.0, 1.0, 1.0, 1.0);
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	// Lighting
	glEnable(GL_COLOR_MATERIAL);  // standard settings are sufficient at the moment
	glEnable(GL_LIGHT0);

	glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
	glHint(GL_LINE_SMOOTH, GL_NICEST);

	// init value for glui live variable
	PoseDifference = 0.0f;
	

	//******* EASTER EGG ZONE - A VECTOR DRAWING APP *************
	// null tool which does nothing
	Tool* nullTool = new NullTool();

	// color picker
	Tool* colorPicker = new Colorpicker(&WindowProps);

	//2D Manipulator
	PaintManipulator = new Manipulator2D(LeftHandPointer, RightHandPointer);

	//2D brush
	Tool* brush2d = new Brush2D(RightHandPointer, &Strokes, PaintManipulator);

	//2D line tool
	Tool* line2d = new StraightLine2D(RightHandPointer, &Strokes, PaintManipulator);

	// 2D Ink
	Tool* ink2d = new Ink2D(&Strokes, &SelectedStrokeID);

	// 2D Stroke Eraser
	Tool* eraser = new StrokeEraser2D(&Strokes, &SelectedStrokeID);

	// rubber tool
	Tool* rubber = new Rubber2D(RightHandPointer, PaintManipulator, &Strokes, &SelectedStrokeID);

	ToolList.push_back(nullTool);
	ToolList.push_back(PaintManipulator);
	ToolList.push_back(colorPicker);
	ToolList.push_back(brush2d);
	ToolList.push_back(line2d);
	ToolList.push_back(ink2d);
	ToolList.push_back(rubber);
	ToolList.push_back(eraser);

	ActiveTool = ToolList[3];

	TextureFrameID = SOIL_load_OGL_texture("../images/frame.png", SOIL_LOAD_AUTO, SOIL_CREATE_NEW_ID, SOIL_FLAG_MIPMAPS);

	//***********************************************************************************************


	



	// *************************** DATA MINING *************************************************
	//stat_crossValidateGestures("test.bin");
	//exit(0);

	// write velocities to file
	//write_velocities_to_file = true;
	//of_velocites.open("Velocities.txt", std::ios::out | std::ios::trunc);
	

	//GRecognizer->loadTrainedGestures("test.bin");
	//if(GRecognizer->TrainedGestureSets.size() > 0)
	//{
	//	stat_getModDTWParameters();
	//	exit(0);
	//}

	//stat_validateRecognition("test_sets.bin","reference_sets.bin");
	//exit(0);

	//*******************************************************************************************


	// update glui live variables
	GLUI_Master.sync_live_all();
	// enter GLUT event processing cycle
	glutMainLoop();
}


// validates trained gestures against test set (we assume that the gesture set IDs correspond to each other)
void stat_validateRecognition(const char* test_file, const char* training_file)
{
	// load the test set
	GRecognizer->loadTrainedGestures(test_file);
	// copy the loaded gestures to new structure
	std::vector<std::vector<Gesture*>> gestures_to_validate;
	std::vector<Gesture*> gesture_pointers;

	for (unsigned int i = 0; i < GRecognizer->TrainedGestureSets.size(); i++)
	{
		std::vector<Gesture*> temp;

		for (unsigned int j = 0; j < GRecognizer->TrainedGestureSets[i]->Gestures.size(); j++)
		{
			Gesture* new_gest = new Gesture( *(GRecognizer->TrainedGestureSets[i]->Gestures[j]) );
			temp.push_back(new_gest);
			gesture_pointers.push_back(new_gest);
		}

		gestures_to_validate.push_back(temp);
	}

	// load the reference gestures
	GRecognizer->loadTrainedGestures(training_file);

	// validate the test set against the reference set with KNN algorithm
	std::stringstream ss;
	ss<<"Validation of "<<test_file<<"_"<<training_file<<" with ";

	GRecognizer->stat_printValidateGestures(gestures_to_validate, 1, 5, EUCLIDEAN, -1.0, (ss.str()+"EUCLIDEAN.txt").c_str());
	GRecognizer->stat_printValidateGestures(gestures_to_validate, 1, 5, MODIFIED_EUCLIDEAN, -1.0, (ss.str()+"MODIFIED_EUCLIDEAN.txt").c_str());
	GRecognizer->stat_printValidateGestures(gestures_to_validate, 1, 5, PEARSON_CORRELATION, -1.0, (ss.str()+"PEARSON_CORRELATION.txt").c_str());
	GRecognizer->stat_printValidateGestures(gestures_to_validate, 1, 5, DTW, -1.0, (ss.str()+"DTW.txt").c_str());
	GRecognizer->stat_printValidateGestures(gestures_to_validate, 1, 5, FM, -1.0, (ss.str()+"FM.txt").c_str());

	// clean memory
	for( unsigned int i = 0; i < gesture_pointers.size(); i++)
	{
		delete gesture_pointers[i];
	}
}

// writes the gathered pose data to file
void stat_printPoses()
{
	std::ofstream file ("POSES.txt", std::ios::trunc);
		
		if (file.is_open())
		{
			for (unsigned int i = 0; i < POSES.size(); i++)
			{
				std::vector<float> temp_pose = getFloatVectorFromPose(POSES[i]);
				for (unsigned int j = 0; j < temp_pose.size(); j++)
				{
					file<<temp_pose[j]<<"\t";
				}

				file<<"\n";
			}
		
		file.close();
		}
		else std::cout << "Unable to open file";
}

// cross validation test for different similarity measures
void stat_crossValidateGestures(const char* training_file)
{
	GRecognizer->loadTrainedGestures(training_file);
	
	std::ofstream of1,of2,of3,of4,of5,of6,of7,of8,of9,of10,of11,of12,of13,of14;
	
	of1.open("FM  Measure avg.txt", std::ios::out | std::ios::trunc);
	of2.open("FM  Measure best.txt", std::ios::out | std::ios::trunc);

	of3.open("DTW Measure avg.txt", std::ios::out | std::ios::trunc);
	of4.open("DTW Measure best.txt", std::ios::out | std::ios::trunc);

	of5.open("Mod DTW Measure avg.txt", std::ios::out | std::ios::trunc);
	of6.open("Mod DTW best.txt", std::ios::out | std::ios::trunc);

	of7.open("Euclidean Measure avg.txt", std::ios::out | std::ios::trunc);
	of8.open("Euclidean Measure best.txt", std::ios::out | std::ios::trunc);

	of9.open("Mod Euclidean Measure avg.txt", std::ios::out | std::ios::trunc);
	of10.open("Mod Euclidean Measure best.txt", std::ios::out | std::ios::trunc);

	of11.open("Pearson Measure avg.txt", std::ios::out | std::ios::trunc);
	of12.open("Pearson Measure best.txt", std::ios::out | std::ios::trunc);

	of13.open("Pearson SQR Measure avg.txt", std::ios::out | std::ios::trunc);
	of14.open("Pearson SQR Measure best.txt", std::ios::out | std::ios::trunc);

	for(unsigned int nr_of_set = 0; nr_of_set < GRecognizer->TrainedGestureSets.size(); nr_of_set++)
	{
		for(unsigned int nr_of_gest = 0; nr_of_gest < GRecognizer->TrainedGestureSets[nr_of_set]->Gestures.size(); nr_of_gest++)
		{
			of1<<"\ngest "<<nr_of_set<<"-"<<nr_of_gest;
			of2<<"\ngest "<<nr_of_set<<"-"<<nr_of_gest;
			of3<<"\ngest "<<nr_of_set<<"-"<<nr_of_gest;
			of4<<"\ngest "<<nr_of_set<<"-"<<nr_of_gest;
			of5<<"\ngest "<<nr_of_set<<"-"<<nr_of_gest;
			of6<<"\ngest "<<nr_of_set<<"-"<<nr_of_gest;
			of7<<"\ngest "<<nr_of_set<<"-"<<nr_of_gest;
			of8<<"\ngest "<<nr_of_set<<"-"<<nr_of_gest;
			of9<<"\ngest "<<nr_of_set<<"-"<<nr_of_gest;
			of10<<"\ngest "<<nr_of_set<<"-"<<nr_of_gest;
			of11<<"\ngest "<<nr_of_set<<"-"<<nr_of_gest;
			of12<<"\ngest "<<nr_of_set<<"-"<<nr_of_gest;
			of13<<"\ngest "<<nr_of_set<<"-"<<nr_of_gest;
			of14<<"\ngest "<<nr_of_set<<"-"<<nr_of_gest;

			for(unsigned int nr_of_other_set = 0; nr_of_other_set < GRecognizer->TrainedGestureSets.size(); nr_of_other_set++)
			{
				float fm_avg = 0.0f;
				float fm_best = 10000000000.0f;

				float dtw_avg = 0.0f;
				float dtw_best = 10000000000.0f;

				float mod_dtw_avg = 0.0f;
				float mod_dtw_best = 10000000000.0f;

				float eucl_avg = 0.0f;
				float eucl_best = 10000000000.0f;

				float mod_eucl_avg = 0.0f;
				float mod_eucl_best = 10000000000.0f;

				float pearson_avg = 0.0f;
				float pearson_best = 10000000000.0f;

				float pearson_sqr_avg = 0.0f;
				float pearson_sqr_best = 10000000000.0f;
			
				int count = 0;
				for(unsigned int nr_of_other_gest = 0; nr_of_other_gest < GRecognizer->TrainedGestureSets[nr_of_other_set]->Gestures.size(); nr_of_other_gest++)
				{
					if( !((nr_of_set == nr_of_other_set) && (nr_of_gest == nr_of_other_gest)) )
					{
						count++;
						float fm = GRecognizer->getFMGestureDistance(*(GRecognizer->TrainedGestureSets[nr_of_set]->Gestures[nr_of_gest]), *(GRecognizer->TrainedGestureSets[nr_of_other_set]->Gestures[nr_of_other_gest]));
						float dtw = GRecognizer->getDTWGestureDistance(*(GRecognizer->TrainedGestureSets[nr_of_set]->Gestures[nr_of_gest]), *(GRecognizer->TrainedGestureSets[nr_of_other_set]->Gestures[nr_of_other_gest]));
						float mod_dtw = GRecognizer->getModifiedDTWGestureDistance(*(GRecognizer->TrainedGestureSets[nr_of_set]->Gestures[nr_of_gest]), *(GRecognizer->TrainedGestureSets[nr_of_other_set]->Gestures[nr_of_other_gest]));
						
						Gesture gest1_copy = *(GRecognizer->TrainedGestureSets[nr_of_set]->Gestures[nr_of_gest]);
						Gesture gest2_copy = *(GRecognizer->TrainedGestureSets[nr_of_other_set]->Gestures[nr_of_other_gest]);

						GestureUtility::resamplePoseVector(gest1_copy.getPoses(),64);
						GestureUtility::resamplePoseVector(gest2_copy.getPoses(),64);

						float eucl = GestureUtility::getEuclideanGestureDistance(gest1_copy,gest2_copy);
						float mod_eucl = GRecognizer->getModifiedEuclideanGestureDistance(gest1_copy,gest2_copy);
						GRecognizer->setSimilarityMeasurement(PEARSON_CORRELATION);
						float pearson = GRecognizer->getPearsonGestureDistance(gest1_copy,gest2_copy,0);
						GRecognizer->setSimilarityMeasurement(PEARSON_CORRELATION_SQUARED);
						float pearson_sqr = GRecognizer->getPearsonGestureDistance(gest1_copy,gest2_copy,0);

						fm_avg += fm;
						fm_best = min(fm_best,fm);

						dtw_avg += dtw;
						dtw_best = min(dtw_best,dtw);
						
						mod_dtw_avg += mod_dtw;
						mod_dtw_best = min(mod_dtw_best,mod_dtw);

						eucl_avg += eucl;
						eucl_best = min(eucl_best,eucl);
						
						mod_eucl_avg += mod_eucl;
						mod_eucl_best = min(mod_eucl_best,mod_eucl);

						pearson_avg += pearson;
						pearson_best = min(pearson_best,pearson);
						
						pearson_sqr_avg += pearson_sqr;
						pearson_sqr_best = min(pearson_sqr_best,pearson_sqr);
					}
				}
				fm_avg /= (float) count;
				dtw_avg /= (float) count;
				mod_dtw_avg /= (float) count;
				eucl_avg /= (float) count;
				mod_eucl_avg /= (float) count;
				pearson_avg /= (float) count;
				pearson_sqr_avg /= (float) count;

				of1<<"\t"<<fm_avg;
				of2<<"\t"<<fm_best; 
				of3<<"\t"<<dtw_avg;
				of4<<"\t"<<dtw_best;
				of5<<"\t"<<mod_dtw_avg;
				of6<<"\t"<<mod_dtw_best; 
				of7<<"\t"<<eucl_avg;
				of8<<"\t"<<eucl_best; 
				of9<<"\t"<<mod_eucl_avg;
				of10<<"\t"<<mod_eucl_best; 
				of11<<"\t"<<pearson_avg;
				of12<<"\t"<<pearson_best; 
				of13<<"\t"<<pearson_sqr_avg;
				of14<<"\t"<<pearson_sqr_best; 
			}
		}
	}
	of1.close();
	of2.close();
	of3.close();
	of4.close();
	of5.close();
	of6.close();
	of7.close();
	of8.close();
	of9.close();
	of10.close();
	of11.close();
	of12.close();
	of13.close();
	of14.close();

	// calculate modified fm distance with probability model
	std::ofstream of15;
	of15.open("Mod FM scores.txt", std::ios::out | std::ios::trunc);


	for(unsigned int nr_of_set = 0; nr_of_set < GRecognizer->TrainedGestureSets.size(); nr_of_set++)
	{
		for(unsigned int nr_of_gest = 0; nr_of_gest < GRecognizer->TrainedGestureSets[nr_of_set]->Gestures.size(); nr_of_gest++)
		{
			of15<<"\ngest "<<nr_of_set<<"-"<<nr_of_gest;

			for(unsigned int nr_of_other_set = 0; nr_of_other_set < GRecognizer->TrainedGestureSets.size(); nr_of_other_set++)
			{
				int exclude = -1;
				if (nr_of_other_set == nr_of_set)
					exclude = nr_of_gest;

				std::cout<<"\r calculation of gest "<<nr_of_set<<"-"<<nr_of_gest;
				float score = GRecognizer->getModifiedFMGestureDistance( *(GRecognizer->TrainedGestureSets[nr_of_set]->Gestures[nr_of_gest]), nr_of_other_set, exclude);

				of15<<"\t"<<score;
			}
		}	
	}
	of15.close();
}