#include "KinectPointer.h"
#include "SOIL.h"
#include <iostream>
#include <fstream>

KinectPointer::KinectPointer(KinectSensor* sensor, PointerType type, WindowProperties* window_props, bool calculate_from_depth_map)
{
	NuiSensor = sensor;
	Type = type;
	State = P_INACTIVE;
	ActivationTimeOut = 500;
	X = 0.0f;
	Y = 0.0f;
	CalculateFromDepthMap = calculate_from_depth_map;
	WindowProps = window_props;
	MinMaxFlag = false;

	if (calculate_from_depth_map)
	{
		XMin = 0.0f;
		XMax = 640.0f;
		YMin = 0.0f;
		YMax = 480.0f;
	}
	else
	{
		// good min max starting values
		if(Type == P_LEFT)
		{
			// left pointer
			XMin = -0.5f;
			XMax = 0.0f;
		}
		else
		{
			//right pointer
			XMin = 0.0f;
			XMax = 0.5f;
		}

		YMin = -0.1f;
		YMax = 0.4f;

	}

	PointerTexID = SOIL_load_OGL_texture("../images/pointer_default.png", SOIL_LOAD_AUTO, SOIL_CREATE_NEW_ID, SOIL_FLAG_MIPMAPS);
	PointerClosedHandTexID = SOIL_load_OGL_texture("../images/pointer_closed_hand.png", SOIL_LOAD_AUTO, SOIL_CREATE_NEW_ID, SOIL_FLAG_MIPMAPS);
	PointerSize = 20.0f;

	RedChannel = 0;
	GreenChannal = 0;
	BlueChannel = 0;
	AlphaChannel =255;

	LastX = 0;
	LastY = 0;
}


KinectPointer::~KinectPointer(void)
{
	glDeleteTextures(1, &PointerTexID);
	glDeleteTextures(1, &PointerClosedHandTexID);
}


math::Vec2f KinectPointer::getWindowCoordinates()
{
	float x_pos = (X - XMin) / (XMax - XMin) * (float) WindowProps->WindowWidth;
	float y_pos = CalculateFromDepthMap ? ( (float) WindowProps->WindowHeight - (float) (Y - YMin) / (YMax - YMin) * (float) WindowProps->WindowHeight ) : (Y - YMin) / (YMax - YMin) * (float) WindowProps->WindowHeight;

	x_pos = max(0,x_pos);
	x_pos = min(WindowProps->WindowWidth-1, x_pos);
	y_pos = max(0,y_pos);
	y_pos = min(WindowProps->WindowHeight-1, y_pos);

	// FOR DEBUGGING PURPOSES
#ifdef MOUSEDEBUG 
	return math::Vec2f(X, Y);
#endif
	//

	return math::Vec2f(x_pos, y_pos);
}


// FOR DEBUGGING PURPOSES
void KinectPointer::setWindowCoordinates(int x, int y)
{
	X = (float) x;
	Y = (float) y;
}
///


void KinectPointer::update(Pose& pose)
{
	float x_position, y_position, z_position;
	int skeleton_index; 
	if (Type == P_LEFT)
	{
		x_position = pose.LeftHandPosition.x;
		y_position = pose.LeftHandPosition.y;
		z_position = pose.LeftHandPosition.z;
		skeleton_index = 7;
	}
	if (Type == P_RIGHT)
	{
		x_position = pose.RightHandPosition.x;
		y_position = pose.RightHandPosition.y;
		z_position = pose.RightHandPosition.z;
		skeleton_index = 11;
	}

	// calculate the hand positions
	float hand_x, hand_y;
	if (CalculateFromDepthMap)
	{
		NUI_SKELETON_DATA skeleton = NuiSensor->getCurrentSkeletonData();
		NuiTransformSkeletonToDepthImage(skeleton.SkeletonPositions[skeleton_index], &hand_x, &hand_y, NUI_IMAGE_RESOLUTION_640x480);
	}
	else
	{
		hand_x = x_position;
		hand_y = y_position;
		if(MinMaxFlag)
			setMinMaxValues();
	}	

	/* Statistics: check out average distance of hand movement and write to file (implemented very inefficient but not used later)
	static std::ofstream of;
	std::string s = "average distance";
	            s += (Type == P_RIGHT)?("_right.txt"):("_left.txt");

	of.open(s.c_str(), std::ios::out | std::ios::app);
	of<<sqrtf(  (hand_x - LastX) * (hand_x - LastX) + (hand_y - LastY) * (hand_y - LastY) )<<"\n";
	of.close();
	//*/


	// we smooth out the pointer movement in a non-linear fashion (like mouse acceleration)
	hand_x = 0.6f * LastX + 0.4f * hand_x;
	hand_y = 0.6f * LastY + 0.4f * hand_y;

	float diff_x = hand_x - LastX;
	float diff_y = hand_y - LastY;

	float scaling = 40.0f * sqrtf(diff_x * diff_x + diff_y * diff_y);

	scaling = min(scaling, 4.0f);
	scaling = max(scaling, 0.5f);

	X = X + diff_x * scaling;
	Y = Y + diff_y * scaling;

	LastX = hand_x;
	LastY = hand_y;

	X = max(XMin, X);
	X = min(XMax, X);
	Y = max(YMin, Y);
	Y = min(YMax, Y);

	// update the pointer's state
	switch (State)
	{
	case P_INACTIVE:
		// if hand is in front of body we change state to progress and set time stamp
		if (z_position > 0.2)
		{
			LastActivationTimeStamp = TimeFunctions::getTime();
			State = P_PROGRESS;
		}
		break;

	case P_PROGRESS:
		// change back to inactive if hand is not in front of body anymore...
		if (z_position < 0.2)
		{
			State = P_INACTIVE;
		}
		// ... or, otherwise, check out time stamp. if hand is long enough in front of body we switch to active state
		else
		{
			if( (TimeFunctions::getTime() - LastActivationTimeStamp) > ActivationTimeOut)
			{
				State = P_ACTIVE;
				// when we change we use our last hand coordinates for X and Y to ignore effect of non-linear scaling
				X = hand_x;
				Y = hand_y;
			}
		}
		break;

	case P_ACTIVE:
		// switch back to inactive state if hand is not in front of body anymore
		if (z_position < 0.2)
		{
			State = P_INACTIVE;
		}
		break;
	}
}

void KinectPointer::setMinMaxValues()
{
	XMin = (X < XMin)?(X):(XMin);
	XMax = (X > XMax)?(X):(XMax);
	YMin = (Y < YMin)?(Y):(YMin);
	YMax = (Y > YMax)?(Y):(YMax);	
}

bool KinectPointer::drawPointer(bool hand_closed)
{
	// if pointers state is active we draw
	if (State == P_ACTIVE)
	{
		
		glEnable(GL_TEXTURE_2D);
		glPushMatrix();

		// move pointer to coordinates
		math::Vec2f coords = getWindowCoordinates();
		glTranslatef(coords.x, coords.y, 0.0f);
		// set size...
		glScalef(PointerSize, PointerSize, 10.0);
		// ... and color
		glColor4ub(RedChannel, GreenChannal, BlueChannel, AlphaChannel);
		glBindTexture(GL_TEXTURE_2D, PointerTexID);
		
		// render open hand pointer...
		glBegin(GL_QUADS);
			glTexCoord2f(0.0, 0.0);		glVertex3f(-0.5, -0.5, 0.0);
			glTexCoord2f(1.0, 0.0);		glVertex3f(0.5, -0.5, 0.0);
			glTexCoord2f(1.0, 1.0);		glVertex3f(0.5, 0.5, 0.0);
			glTexCoord2f(0.0, 1.0);		glVertex3f(-0.5, 0.5, 0.0);
		glEnd();
	
		// ... or closed hand pointer
		if(hand_closed)
		{
			glBindTexture(GL_TEXTURE_2D, PointerClosedHandTexID);
			glBegin(GL_QUADS);
				glTexCoord2f(0.0, 0.0);		glVertex3f(-1.0, -1.0, 0.0);
				glTexCoord2f(1.0, 0.0);		glVertex3f(1.0, -1.0, 0.0);
				glTexCoord2f(1.0, 1.0);		glVertex3f(1.0, 1.0, 0.0);
				glTexCoord2f(0.0, 1.0);		glVertex3f(-1.0, 1.0, 0.0);
			glEnd();	
		}

		glPopMatrix();
		glDisable(GL_TEXTURE_2D);

		return true;
	}
	else
	{
		return false;
	}
}

void KinectPointer::setPointerShapeTexture(const char* filename)
{
	// clear old texture
	glDeleteTextures(1, &PointerTexID);
	PointerTexID = SOIL_load_OGL_texture(filename, SOIL_LOAD_AUTO, PointerTexID, SOIL_FLAG_MIPMAPS);
}

void KinectPointer::setPointerColor(unsigned char red, unsigned char green, unsigned char blue, unsigned char alpha)
{
	RedChannel = red;
	GreenChannal = green;
	BlueChannel = blue;
	AlphaChannel = alpha;
}