/*
Class providing a visual mouse pointer pendant for left or right hand with own render methods.
Open and closed hands are visualized with different textures. The Pointer are modelled via state machine.
*/
#pragma once
#include "KinectSensor.h"
#include "Pose.h"
#include "TimeFunctions.h"
#include "math\Vec2f.h"
#include "WindowProperties.h"
#include <GL\freeglut.h>

// left or right hand pointer
enum PointerType{P_LEFT, P_RIGHT};
// the pointers state
enum PointerState{P_INACTIVE, P_PROGRESS, P_ACTIVE};

class KinectPointer
{
public:
	// instantiate a pointer with different parameters. if "calculate_from_depth_map" is true we use the global hand coordinates from the kinect depth space as window coordinates (after rescaling))
	KinectPointer(KinectSensor* sensor, PointerType type, WindowProperties* window_props, bool calculate_from_depth_map=false);
	~KinectPointer(void);

	// returns the pointer's window coordinates
	math::Vec2f getWindowCoordinates();
	// updates pointer state (should be called each time in the main program loop)
	void update(Pose& pose);
	// returns if pointer shall be visible
	inline bool isActive(){return (State == P_ACTIVE);}
	// renders the pointer (returns true if pointer was rendered)
	bool drawPointer(bool hand_closed);
	// set a pointer shape texture (64x64 texture file)
	void setPointerShapeTexture(const char* filename);
	// set the size of the pointer
	inline void setPointerSize(float size){PointerSize = size;};
	// set the pointer color
	void setPointerColor(unsigned char red, unsigned char green, unsigned char blue, unsigned char alpha);	
	// set the flag to adapt the scaling
	void setMinMax(bool flag){MinMaxFlag = flag;};
	bool getMinMax(){return MinMaxFlag;};

	// FOR DEBUGGING PURPOSES
	void KinectPointer::setWindowCoordinates(int x, int y);
	//


private:
	// sets the adaptive scaling of hand coordinates
	void setMinMaxValues();
	
	// pointer coordinates (maximum/minimum interval, current and last position)
	float XMin, XMax, YMin, YMax;
	float X, Y;
	float LastX, LastY;

	// pointers type and state
	PointerType Type; 
	PointerState State;
	
	// the pointers appearance attributes
	float PointerSize;
	unsigned char RedChannel;
	unsigned char GreenChannal;
	unsigned char BlueChannel;
	unsigned char AlphaChannel;
	GLuint PointerClosedHandTexID;
	GLuint PointerTexID;

	// assigned kinect sensor
	KinectSensor* NuiSensor;

	// last time the pointers state switched from "inactive" to "in progress"
	clock_t LastActivationTimeStamp;
	// amount of time that has to elapse before we set the pointer visible
	long ActivationTimeOut;

	// flag that decides if we calculate the window coordinates from the depth map coordinates of hand
	bool CalculateFromDepthMap;
	// pointer to window properties to provide correct calculation of window coordinates
	WindowProperties *WindowProps;
	// flag to decide if we adapt min and max pointer coordinates
	bool MinMaxFlag;
};

