#include "KinectSensor.h"
#include <iostream>

KinectSensor::KinectSensor(void)
{
	VideoData = new BYTE[640*480*4];
	DepthData = new BYTE[640*480*4];
	DepthMap = new USHORT[640*480];
	ActiveSkeletonID = -1;
	init();
	CurrentSkeletonData = getDummySkeleton();
}


KinectSensor::~KinectSensor(void)
{
}


void KinectSensor::init()
{	
	BSTR m_NuiID;
	HRESULT	hr;
	
	if (NuiCreateSensorByIndex(0, &NuiSensor) == S_OK)
	{
		m_NuiID = NuiSensor->NuiDeviceConnectionId();

		// declare event handles
		HANDLE m_hNextDepthFrameEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
		HANDLE m_hNextVideoFrameEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
		HANDLE m_hNextSkeletonEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

		// initialize sensor
		DWORD nuiFlags = NUI_INITIALIZE_FLAG_USES_DEPTH_AND_PLAYER_INDEX | NUI_INITIALIZE_FLAG_USES_SKELETON |  NUI_INITIALIZE_FLAG_USES_COLOR;
		hr = NuiSensor->NuiInitialize(nuiFlags);
		if (E_NUI_SKELETAL_ENGINE_BUSY == hr)
		{
			nuiFlags = NUI_INITIALIZE_FLAG_USES_DEPTH |  NUI_INITIALIZE_FLAG_USES_COLOR;
			hr = NuiSensor->NuiInitialize(nuiFlags);
		}

		//enable skeletal tracking
		if (HasSkeletalEngine(NuiSensor))
		{	
			hr = NuiSensor->NuiSkeletonTrackingEnable(m_hNextSkeletonEvent, 0);
		}

		// open video and depth stream
		hr = NuiSensor->NuiImageStreamOpen(NUI_IMAGE_TYPE_COLOR, NUI_IMAGE_RESOLUTION_640x480, 0, 2, m_hNextVideoFrameEvent, &VideoStreamHandle);
		hr = NuiSensor->NuiImageStreamOpen(NUI_IMAGE_TYPE_DEPTH_AND_PLAYER_INDEX, NUI_IMAGE_RESOLUTION_640x480, 0, 2, m_hNextDepthFrameEvent, &DepthStreamHandle);
	}
	else
	{
		// show a message box to inform the user
		int choice = MessageBox(NULL, TEXT("Kinect Sensor nicht gefunden oder fehlerhaft initialisiert! berprfen Sie die USB-Verbindung und starten sie erneut."), NULL, MB_OK | MB_ICONERROR | MB_SETFOREGROUND | MB_TOPMOST | MB_TASKMODAL );
		exit(1);
	}
}


const BYTE* KinectSensor::updateVideoFrameData()
{
	NUI_IMAGE_FRAME imageFrame;
	
	// get current video frame
	HRESULT hr = NuiSensor->NuiImageStreamGetNextFrame(VideoStreamHandle, 100, &imageFrame );

	if(hr>=0)
	{
		NUI_LOCKED_RECT LockedRect;

		// lock the buffer to render
		imageFrame.pFrameTexture->LockRect(0, &LockedRect, NULL, 0);
		if ( LockedRect.Pitch != 0 )
		{
			//copy framedata
			memcpy(VideoData, LockedRect.pBits, LockedRect.size);
			
			// unlock the render buffer
			 imageFrame.pFrameTexture->UnlockRect( 0 );

			// release frame
			NuiSensor->NuiImageStreamReleaseFrame(VideoStreamHandle, &imageFrame );
		}	
	}
	else
	{
		std::cout<<"Something went wrong for Video Frame..."<<std::endl;
	}

	return VideoData;
}

const BYTE* KinectSensor::updateDepthFrameData()
{
	NUI_IMAGE_FRAME imageFrame;
	
	// get current depth frame
	HRESULT hr = NuiSensor->NuiImageStreamGetNextFrame(DepthStreamHandle, 100, &imageFrame );

	if(hr>=0)
	{
		NUI_LOCKED_RECT LockedRect;

		// lock the buffer to render
		imageFrame.pFrameTexture->LockRect(0, &LockedRect, NULL, 0);
		if ( LockedRect.Pitch != 0 )
		{
			USHORT *buffer = (USHORT *) LockedRect.pBits;
		
			for(int i=0; i<640*480; i++)
			{
				
				// shift operation to discard player info
				//USHORT depth = buffer[i] >> 3;
				USHORT depth = NuiDepthPixelToDepth(buffer[i]);
				// write depth into depth map
				DepthMap[i] = depth;

				// get player index
				USHORT player = NuiDepthPixelToPlayerIndex(buffer[i]);

				// prepare colors for the image of depth frame
				BYTE intensity;
				BYTE RED = 0;
				if (depth == 0) 
					RED = 255;
				// map interval 800mm - 4000mm to 0-255
				float val = (float)  (((depth - 800.0) > 0)?(depth - 800.0):(0));
		        intensity = (BYTE) (255 - 255 * val / 3200.0);

				// BGRA format!
				DepthData[4*i] = intensity - RED;
				DepthData[4*i+1] = intensity - RED;
				DepthData[4*i+2] = intensity;
				DepthData[4*i+3] = 0;

				if(player != 0)
				{
					DepthData[4*i] = 10;
					DepthData[4*i+2] = 10;
				}
			}
			// unlock the render buffer
			 imageFrame.pFrameTexture->UnlockRect( 0 );
			// release frame
			NuiSensor->NuiImageStreamReleaseFrame(DepthStreamHandle, &imageFrame );
		}		
	}
	else
	{
		std::cout<<"Something went wrong for Depth Frame..."<<std::endl;
	}

	return DepthData;
}


const BYTE* KinectSensor::getCurrentVideoFrame()
{
	return VideoData;
}


const BYTE* KinectSensor::getCurrentDepthFrame()
{
	return DepthData;
}


const USHORT* KinectSensor::getDepthMap()
{
	return DepthMap;	
}


NUI_SKELETON_DATA& KinectSensor::updateSkeletonData()
{
	NUI_SKELETON_DATA OldSkeletonData = CurrentSkeletonData;
	NUI_SKELETON_FRAME skeleton_frame; 
	int last_valid_index = -1;

	// try to get the next skeleton frame from the kinect
	if( NuiSensor->NuiSkeletonGetNextFrame(100, &skeleton_frame) >= 0)
	{
		// we have to find out which index the current user has in the current skeleton frame 
		//(if tracking is not lost occasionally this would be the same as in the last frame)
		// hence we try to find the index which equals the last one or, in the other case, we just 
		// take the last valid entry in the list.
		for( int i = 0 ; i < NUI_SKELETON_COUNT ; i++ )
		{
			if(skeleton_frame.SkeletonData[i].eTrackingState == NUI_SKELETON_TRACKED)
			{
				last_valid_index = i;
				//if i equals our last active skeleton index we break
				if ( i == ActiveSkeletonID )
				{
					break;
				}				
			}
		}

		if(last_valid_index != -1)
		{
			// we set the active skeleton index
			ActiveSkeletonID = last_valid_index;
			// smooth the tracking
			NuiSensor->NuiTransformSmooth(&skeleton_frame, NULL);
			CurrentSkeletonData = skeleton_frame.SkeletonData[ActiveSkeletonID];
		}
		else
		{
			CurrentSkeletonData = OldSkeletonData;
		}

	}
	else
	{
	// if we can't get a skeleton frame we return the last recognized skeleton
	CurrentSkeletonData = OldSkeletonData;
	}

	return CurrentSkeletonData;	
}

NUI_SKELETON_DATA& KinectSensor::getCurrentSkeletonData()
{
	return CurrentSkeletonData;
}

void KinectSensor::setCameraAngle(LONG angle)
{
	NuiSensor->NuiCameraElevationSetAngle(angle);
}


LONG KinectSensor::getCameraAngle()
{
	LONG angle;
	NuiSensor->NuiCameraElevationGetAngle(&angle);
	return angle;
}
	

NUI_SKELETON_DATA KinectSensor::getDummySkeleton()
{
	NUI_SKELETON_DATA dummy;
	dummy.dwEnrollmentIndex = 0;
	dummy.dwTrackingID = 0;
    dummy.dwEnrollmentIndex = 0;
    dummy.dwUserIndex = 0;
	dummy.Position.w = 0.0f;
	dummy.Position.x = 0.0f;
	dummy.Position.y = 0.0f;
	dummy.Position.z = 0.0f;
	dummy.dwQualityFlags = 0;
	dummy.eTrackingState = NUI_SKELETON_NOT_TRACKED;

	for (int i = 0; i < 20; i++)
	{
		dummy.SkeletonPositions[i].w = 0;
		dummy.SkeletonPositions[i].x = 0;
		dummy.SkeletonPositions[i].y = 0;
		dummy.SkeletonPositions[i].z = 0;

		dummy.eSkeletonPositionTrackingState[i] = NUI_SKELETON_POSITION_NOT_TRACKED;
	}
	return dummy;
}
 