#include "GridMesh.h"
#include <GL/freeglut.h>
#include <iostream>
#include <algorithm>
#include <set>
#include "SOIL.h"
#include "TimeFunctions.h"
#include "FileHandling.h"

// help functions
int RGBtoIndex(unsigned char R, unsigned char G, unsigned char B)
{
	return 65536 * R + 256 * G + B;
}

std::vector<unsigned char> IndexToRGB(unsigned long index)
{
	unsigned char R = (unsigned char) (index / 65536);
	index -= R * 65536;
	unsigned char G = (unsigned char) (index / 256);
	index -= G * 256;
	unsigned char B = (unsigned char)  index;

	std::vector<unsigned char> result;
	result.push_back(R);
	result.push_back(G);
	result.push_back(B);

	return result;
}



long GridMesh::getGridPointIndex(int x, int z)
{
	return z*(Width+1) + x;
}

void GridMesh::setFaceNormal(int face_index)
{
	math::Vec3f p0 = PointList[FaceList[face_index].p0].coords;
	math::Vec3f p1 = PointList[FaceList[face_index].p1].coords;
	math::Vec3f p2 = PointList[FaceList[face_index].p2].coords;

	FaceList[face_index].normal = math::normalize(math::crossProduct(p2 - p1, p0 - p1));
}

void GridMesh::setVertexNormal(int point_index)
{
	for(unsigned int i = 0; i < PointList[point_index].faces.size(); i++)
	{
		PointList[point_index].normal += FaceList[PointList[point_index].faces[i]].normal;
	}
	PointList[point_index].normal.normalize();
}

void GridMesh::recalculateGridPointNormal(GridPoint* p)
{
	PointsForNormalRecalculation.push_back(p);
}

void GridMesh::executeNormalsRecalculation()
{
	//clock_t start = TimeFunctions::getTime();
	//// first we collect all involved faces in a set to avoid duplicates 
	//std::set<long> faces;
	//for (int i = 0; i < PointsForNormalRecalculation.size(); i++)
	//{
	//	for (int j = 0; j < PointsForNormalRecalculation[i]->faces.size(); j++)
	//	{
	//		faces.insert(PointsForNormalRecalculation[i]->faces[j]);
	//	}
	//}

	//// calculate normals for the faces in the set
	//std::set<long>::iterator it;

	// for (it = faces.begin(); it != faces.end(); it++)
	// {
	//	 setFaceNormal(*it);
	// }
	// std::cout<<"\nTime elapsed using set [ms]: "<<TimeFunctions::getTime() - start;

	// ^ THE COSTS FOR CREATING A STD::SET ARE WAY TOO HIGH



	//clock_t start = TimeFunctions::getTime();
	for (unsigned int i = 0; i < PointsForNormalRecalculation.size(); i++)
	{
		for (unsigned int j = 0; j < PointsForNormalRecalculation[i]->faces.size(); j++)
		{
			 setFaceNormal(PointsForNormalRecalculation[i]->faces[j]);
		}
	}
	 //std::cout<<"\nTime elapsed w/o using set [ms]: "<<TimeFunctions::getTime() - start;

	 // calculate normals for the invovled grid points
	 for (unsigned int i = 0; i < PointsForNormalRecalculation.size(); i++)
	 {
		 setVertexNormal(PointsForNormalRecalculation[i]->index);
	 }

	 // finally we clear the list of involved grid points
	 PointsForNormalRecalculation.clear();
}

void GridMesh::calculateNormals()
{
	// calculate face normals
	for (unsigned int i = 0; i < FaceList.size(); i++)
	{
		setFaceNormal(i);
	}

	// calculate vertex normals
	for (unsigned int i = 0; i < PointList.size(); i++)
	{
		setVertexNormal(i);
	}
}

bool GridMesh::setAdjacentFace(int point_index, int face_index)
{	
	// check if face is already set
	for (unsigned int i = 0; i < PointList[point_index].faces.size(); i++ )
	{
		if ( PointList[point_index].faces[i] == face_index )
		{
			return false;
		}
	}

	PointList[point_index].faces.push_back(face_index);
	return true;
}

void GridMesh::setAdjacentGridPoint(int point_index, int neighbour_index)
{
	if(neighbour_index == point_index)
	{
		return;
	}

	bool already_in_list = false;
	for (unsigned int i = 0; i < PointList[point_index].neighbours.size(); i++ )
	{
		if ( PointList[point_index].neighbours[i] == neighbour_index )
		{
			already_in_list = true;
			return;
		}
	}

	if (!already_in_list)
	{
		PointList[point_index].neighbours.push_back(neighbour_index);
		return;
	}
}

void GridMesh::setNeighbours(int point_index, int face_index)
{
	// try to append face index to the list of faces of the grid point
	bool success = setAdjacentFace(point_index, face_index);

	// if the face was set we try to append the face's adjacent points to the neighbours list of the grid points
	if(success)
	{
		setAdjacentGridPoint(point_index, FaceList[face_index].p0);
		setAdjacentGridPoint(point_index, FaceList[face_index].p1);
		setAdjacentGridPoint(point_index, FaceList[face_index].p2);
	}	
}



bool GridMesh::ScreenToLocalCoords(math::Vec2f screen_coords, math::Vec3f& local_coords, bool snap_to_grid)
{
	// we try to find the coordinates of the pointer in object space with colorpicking
	glClearColor(1.0, 1.0, 1.0, 1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glShadeModel(GL_FLAT);
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );

	// we iterate over the faces and draw triangle strips
	for(unsigned int i = 0; i < FaceList.size(); i += Width*2)
	{
		std::vector<unsigned char> color;

		glBegin(GL_TRIANGLE_STRIP);
			
			glVertex3fv(PointList[FaceList[i].p1].coords.v); 
			glVertex3fv(PointList[FaceList[i].p0].coords.v); 

			for(int x = 0; x < Width*2; x += 2)
			{
				color = IndexToRGB(FaceList[i+x].ID);
				glColor3ub(color[0], color[1], color[2]);
				glVertex3fv(PointList[FaceList[i+x+1].p1].coords.v); 

				color = IndexToRGB(FaceList[i+x+1].ID);
				glColor3ub(color[0], color[1], color[2]);
				glVertex3fv(PointList[FaceList[i+x+1].p2].coords.v); 
			}
		glEnd();
	}

	
	unsigned char color[3];
	glReadPixels((int) screen_coords.x, (int) screen_coords.y, 1, 1, GL_RGB, GL_UNSIGNED_BYTE, color);

	//std::cout<<(int) color[0]<<"   "<<(int) color[1]<<"   "<<(int) color[2]<<"\n";


	if( (color[0] == 255) && (color[1] == 255) && (color[2] == 255) )
	{
		// this is the background color. hence we didn't find any triangle
		LastSelectedFaceIndex = -1;
		LastSelectedVertexIndex = -1;
		return false;
	}
	
	// we found a triangle, let's set the face index
	LastSelectedFaceIndex = RGBtoIndex(color[0], color[1], color[2]);

	// since we have the selected face let's find the pointer's coordinates in model space
	math::Vec3f p0 = PointList[FaceList[LastSelectedFaceIndex].p0].coords;
	math::Vec3f p1 = PointList[FaceList[LastSelectedFaceIndex].p1].coords;
	math::Vec3f p2 = PointList[FaceList[LastSelectedFaceIndex].p2].coords;

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glShadeModel(GL_SMOOTH);
	glBegin(GL_TRIANGLES);
		glColor3f(1.0,0.0,0.0);
		glVertex3fv(p0.v);
				
		glColor3f(0.0,1.0,0.0);
		glVertex3fv(p1.v);
				
		glColor3f(0.0,0.0,1.0);
		glVertex3fv(p2.v);
	glEnd();

	float bary_coords[3];
	glReadPixels((int) screen_coords.x, (int) screen_coords.y, 1, 1, GL_RGB, GL_FLOAT, &bary_coords);	
	
	math::Vec3f point;

	if (snap_to_grid)
	{
		// get the index of the nearest grid point. this is the one corresponding to the barycentric coordinate with the largest value
		if( bary_coords[0] >= std::max(bary_coords[1], bary_coords[2]) )
		{
			LastSelectedVertexIndex = FaceList[LastSelectedFaceIndex].p0;
		}
		else
		{
			if( bary_coords[1] >= std::max(bary_coords[0], bary_coords[2]) )
			{
				LastSelectedVertexIndex = FaceList[LastSelectedFaceIndex].p1;
			}
			else
			{
				LastSelectedVertexIndex = FaceList[LastSelectedFaceIndex].p2;
			}
		}

		point = PointList[LastSelectedVertexIndex].coords;
	}
	else
	{
		//point.x = (float) ((double) bary_coords[0]) * ((double) p0.x) + ((double) bary_coords[1]) * ((double) p1.x) + ((double) bary_coords[2]) * ((double) p2.x);
		// ^ leads to problems due to floating point inaccuracies

		point = bary_coords[0] * p0 + bary_coords[1] * p1 + (1.0f - bary_coords[0] - bary_coords[1]) * p2;		
	}

	local_coords.x = point.x;
	local_coords.y = point.y;
	local_coords.z = point.z;
	
	
	/*
	//std::cout<<"\r"<<bary_coords[0]<<"    "<<bary_coords[1]<<"    "<<bary_coords[2];
	std::cout<<"\r"<<point.x<<"    "<<point.y<<"    "<<point.z;
	glColor3f(.0f, .0f, .0f);
	glPointSize(5.0f);
	glBegin(GL_POINTS);
		glVertex3fv(point.v);
	glEnd();
	glutSwapBuffers();*/
	return true;
}


GridMesh::GridMesh(int width, int length)
{
	TextureID = SOIL_load_OGL_texture("../images/grey512.png", SOIL_LOAD_RGB, SOIL_CREATE_NEW_ID, SOIL_FLAG_MIPMAPS);
	Width = width;
	Length = length;
	LastSelectedFaceIndex = -1;
	LastSelectedVertexIndex = -1;
	ShowNormals = false;

	// initialize list of points
	for( long i = 0; i < (length+1)*(width+1); i++)
	{	
		int x = i % (Width + 1);
		int z = i / (Width + 1);

		math::Vec3f v = math::Vec3f((float) x, 0.0f, (float) z);

		GridPoint p;
		p.coords = v;
		p.tex_coords = math::Vec2f(v.x / width, v.z / length);
		p.normal = math::Vec3f(0.0f, 1.0f, 0.0f);
		p.index = i;

		PointList.push_back(p);
	}
	// initialize face list
	FaceList.clear();
	for( long i = 0; i < length*width; i++)
	{
		// i iterates over grid cells. we always initializes two triangles in one step
		/*
		
		i0 =  p0 __________p2  = i1
		      | \           |
		      |   \    f1   | 
		      |     \       | 
		      |  f0   \     |  
		      |         \   |
		i2 =  p1______p2_ \_p1  = i3

		*/

		long i0 = i + i / width;
		long i1 = i0 + 1;
		long i2 = i1 + width;
		long i3 = i2 + 1;


		Face f0;
		f0.ID = 2*i;
		/*f0.p1 = i + i / width;
		f0.p2 = f0.p1 + 1;
		f0.p0 = f0.p2 + width;*/
		f0.p0 = i0;
		f0.p1 = i2;
		f0.p2 = i3;
		
		f0.normal = math::Vec3f(0.0f, 1.0f, 0.0);

		Face f1;
		f1.ID = 2*i+1;
		/*f1.p1 = i + i / width + 1;
		f1.p0 = f1.p1 + width;
		f1.p2 = f1.p0 + 1;*/
		f1.p0 = i0;
		f1.p1 = i3;
		f1.p2 = i1;
		f1.normal = math::Vec3f(0.0f, 1.0f, 0.0);

		// push the faces into the list
		FaceList.push_back(f0);
		FaceList.push_back(f1);

		// we set the faces for the 4 corner points      <- TODO: a lot of overhead here, find better solution
		setNeighbours(i0, f0.ID);
		setNeighbours(i0, f1.ID);
		setNeighbours(i1, f1.ID);
		setNeighbours(i2, f0.ID);
		setNeighbours(i3, f0.ID);
		setNeighbours(i3, f1.ID);
	}
	
}

GridMesh::~GridMesh()
{
}



GridPoint* GridMesh::getGridPoint(int x, int z)
{
	return &(PointList[getGridPointIndex(x,z)]);
}

void GridMesh::render()
{
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
	glEnable(GL_LIGHTING);
	glShadeModel(GL_SMOOTH);
	

	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, TextureID);
	glColor3f(1.0f, 1.0f, 1.0f);

	// we iterate over the faces and draw triangle strips
	for(unsigned int i = 0; i < FaceList.size(); i += Width*2)
	{
		glBegin(GL_TRIANGLE_STRIP);
			
			
			glNormal3fv(PointList[FaceList[i].p1].normal.v);
			glTexCoord2f(PointList[FaceList[i].p1].coords.x/((float) Width), PointList[FaceList[i].p1].coords.z/((float) Length));
			glVertex3fv(PointList[FaceList[i].p1].coords.v); 
		
			glNormal3fv(PointList[FaceList[i].p0].normal.v);
			glTexCoord2f(PointList[FaceList[i].p0].coords.x/((float) Width), PointList[FaceList[i].p0].coords.z/((float) Length));
			glVertex3fv(PointList[FaceList[i].p0].coords.v); 

			for(int x = 0; x < Width*2; x += 2)
			{
				
				glNormal3fv(PointList[FaceList[i+x+1].p1].normal.v);
				glTexCoord2f(PointList[FaceList[i+x+1].p1].coords.x/((float) Width), PointList[FaceList[i+x+1].p1].coords.z/((float) Length));
				glVertex3fv(PointList[FaceList[i+x+1].p1].coords.v); 
				
				glNormal3fv(PointList[FaceList[i+x+1].p2].normal.v);
				glTexCoord2f(PointList[FaceList[i+x+1].p2].coords.x/((float) Width), PointList[FaceList[i+x+1].p2].coords.z/((float) Length));
				glVertex3fv(PointList[FaceList[i+x+1].p2].coords.v); 
			}
		glEnd();
	}

	// show vertex normals
	if(ShowNormals)
	{
		for (unsigned int i = 0; i < PointList.size(); i++)
		{
			glBegin(GL_LINES);
			glColor3ub(255, 0, 0);
			math::Vec3f p1 = PointList[i].coords;
			math::Vec3f p2 = p1 + PointList[i].normal;
				glVertex3fv(p1.v);
				glVertex3fv(p2.v);
			glEnd();
		}	
	}
	
	
		
	

	// show selected face
	/*if(LastSelectedFaceIndex != -1)
	{
		glTranslatef(0.0f, 0.1f, 0.0f);
		glColor3ub(255,0,0);
		glBegin(GL_TRIANGLES);
			glVertex3fv(PointList[FaceList[LastSelectedFaceIndex].p0].coords.v);
			glVertex3fv(PointList[FaceList[LastSelectedFaceIndex].p1].coords.v);
			glVertex3fv(PointList[FaceList[LastSelectedFaceIndex].p2].coords.v);
		glEnd();
		glTranslatef(0.0f, -0.3f, 0.0f);
	}*/

	// show selected grid point
	/*if(LastSelectedVertexIndex != -1)
	{
		glTranslatef(0.0f, 0.2f, 0.0f);
		glColor3ub(255,0,0);
		glPointSize(30.f);
		glBegin(GL_POINTS);
			glVertex3fv(PointList[LastSelectedVertexIndex].coords.v);
		glEnd();
		glTranslatef(0.0f, -0.3f, 0.0f);
	}*/

	glDisable(GL_TEXTURE_2D);

	return;


	// DEBUGGING: grow selected point
	if(LastSelectedVertexIndex != -1)
	{
		PointList[LastSelectedVertexIndex].coords.y += 0.1f;
		// recalculate normals
		for(unsigned int i = 0; i < PointList[LastSelectedVertexIndex].faces.size(); i++)
		{
			setFaceNormal(PointList[LastSelectedVertexIndex].faces[i]);
		}
		setVertexNormal(LastSelectedVertexIndex);
	}

	
}

void GridMesh::drawColorMap()
{
	// draw color texture to quad
	glBindTexture(GL_TEXTURE_2D, TextureID);
	glColor3f(1.0f, 1.0f, 1.0f);
	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 0.0);		glVertex3f(0.0, 0.0, 0.0);
		glTexCoord2f(0.0, 1.0);		glVertex3f(0.0, 512.0, 0.0);
		glTexCoord2f(1.0, 1.0);		glVertex3f(512.0, 512.0, 0.0);
		glTexCoord2f(1.0, 0.0);		glVertex3f(512.0, 0.0, 0.0);
	glEnd();
}

std::vector<unsigned char> HeightToRGB(float height, float min, float max, bool high_precision = false)
{
	if(high_precision)
	{
		// variant 1: encode height value into base-256 represantation with 16777216 distinctive values
		if((max - min) != 0.0f)
		{
			return IndexToRGB((unsigned long) (16777215 * (height - min) / (max - min)));
		}
		else 
		{
			return IndexToRGB( 0 );
		}
	}
	else
	{
		// variant 2: encode height value in heatmap style (R->G->B) with 765 distinctive values, poor resolution but better to comprehend for humans
		unsigned char R = 0;
		unsigned char G = 0;
		unsigned char B = 0;

		if((max - min) != 0.0f)
		{
			int scaled_height = (int) (765 * (height - min) / (max - min));

			if(scaled_height < 256)
			{
				R = scaled_height;
				G = 0;
				B = 0;
			}
			else
			{
				if(scaled_height < 511)
				{
					R = 255;
					G = scaled_height - 255;
					B = 0;
				}
				else
				{
					R = 255;
					G = 255;
					B = scaled_height - 510;
				}
			}	
		}

		std::vector<unsigned char> result;
		result.push_back(R);
		result.push_back(G);
		result.push_back(B);

		return result;
	}
}

void GridMesh::saveColorAndHeightMap()
{
	std::string filename = FileHandling::save("just insert a name without file extension, tga files will be generated anyway");
	
	// abort if no file name was specified
	if(filename == "") return;
	
	// save color texture
	glBindTexture(GL_TEXTURE_2D, TextureID);
	unsigned char* imagedata = new unsigned char[512*512*3];
	glGetTexImage(GL_TEXTURE_2D, 0, GL_RGB, GL_UNSIGNED_BYTE, imagedata); 
	SOIL_save_image((filename + "_colormap.tga").c_str(), SOIL_SAVE_TYPE_TGA, 512, 512, 3, imagedata);
	delete[] imagedata;

	// get max and min height value
	float max_height = PointList[0].coords.y;
	float min_height = max_height;

	for(unsigned int i = 0; i < PointList.size(); i++)
	{
		max_height = (PointList[i].coords.y > max_height) ? (PointList[i].coords.y):(max_height); 
		min_height = (PointList[i].coords.y < min_height) ? (PointList[i].coords.y):(min_height); 
	}

	std::vector<std::vector<unsigned char>> rgb_heights_high_precision;
	std::vector<std::vector<unsigned char>> rgb_heights_low_precision;

	
	for(unsigned int i = 0; i < PointList.size(); i++)
	{
		rgb_heights_high_precision.push_back( HeightToRGB(PointList[i].coords.y, min_height, max_height, true) ); 
		rgb_heights_low_precision.push_back( HeightToRGB(PointList[i].coords.y, min_height, max_height, false) ); 
	}

	unsigned char* height_data_hp = new unsigned char[PointList.size()*3];
	unsigned char* height_data_lp = new unsigned char[PointList.size()*3];
	
	for (unsigned int i = 0; i < PointList.size(); i++)
	{
		height_data_hp[3*i]   = rgb_heights_high_precision[i][0];
		height_data_hp[3*i+1] = rgb_heights_high_precision[i][1];
		height_data_hp[3*i+2] = rgb_heights_high_precision[i][2];

		height_data_lp[3*i]   = rgb_heights_low_precision[i][0];
		height_data_lp[3*i+1] = rgb_heights_low_precision[i][1];
		height_data_lp[3*i+2] = rgb_heights_low_precision[i][2];
	}

	SOIL_save_image((filename + "_heightmap_high_precision.tga").c_str(), SOIL_SAVE_TYPE_TGA, Width+1, Length+1, 3, height_data_hp);
	SOIL_save_image((filename + "_heightmap_low_precision.tga").c_str(), SOIL_SAVE_TYPE_TGA, Width+1, Length+1, 3, height_data_lp);
	delete[] height_data_hp;
	delete[] height_data_lp;
}