#include "TextureTextRenderer.h"
#include <algorithm>
#include <iostream>

float TextureTextRenderer::TextureAspect = 21.875;
GLuint TextureTextRenderer::CapitalsTextureID;
GLuint TextureTextRenderer::MinusculesTextureID;
std::map<char, LetterTexInfo> TextureTextRenderer::FontTexInformations;
bool TextureTextRenderer::TexturesLoaded = false;


TextureTextRenderer::TextureTextRenderer(void)
{
	TextDisplayList = glGenLists(1);
	BoxDisplayList = glGenLists(1);

	// set up all the letter positions and load textures
	if(!TexturesLoaded)
	{
		setupLetters();
		TexturesLoaded = true;
	}
	
	// init values
	Text = "";
	TextHeight = 0; 
	LetterSpacing = 0;
	XPosition = 0;
	YPosition = 0;
	RedChannel = 0;
	GreenChannal = 0;
	BlueChannel = 0;
	AlphaChannel = 0;
	BoxedText = 0;
	TextDuration = 0;
	TextFadeDuration = 0;
}


TextureTextRenderer::~TextureTextRenderer(void)
{
}

void TextureTextRenderer::setupLetters()
{
	// we load the textures
	CapitalsTextureID = SOIL_load_OGL_texture("../images/capitals.png", SOIL_LOAD_AUTO, SOIL_CREATE_NEW_ID, SOIL_FLAG_POWER_OF_TWO); // SOIL_FLAG_MIPMAPS);
	MinusculesTextureID = SOIL_load_OGL_texture("../images/minuscules.png", SOIL_LOAD_AUTO, SOIL_CREATE_NEW_ID, SOIL_FLAG_POWER_OF_TWO); // SOIL_FLAG_MIPMAPS);

	// the letter boundaries in the texture files
	float positions_capitals_tex[] = {.0f, .0358f, .0606f, .0925f, .1253f, .1499f, .1712f, .2034f, .2336f, .2430f, .2623f, .2923f, .3170f, .3533f, .3837f, .42f, .4436f, .4799f, 
									  .5071f, .5325f, .5655f, .5946f, .6265f, .6689f, .7001f, .7310f, .7592f, .7713f, .7923f, .8047f, .8147f, .8295f, .8416f, .8759f, .9019f, .9239f, .9458f, .9773f, .9873f};

	float positions_minuscules_tex[] = {.0f, .0257f, .0528f, .0764f, .1056f, .1314f, .1491f, .1779f, .2045f, .2158f, .2357f, .2608f, .2721f, .3134f, .3401f, .3686f, .3958f, .425f, 
										.4419f, .4643f, .4825f, .5095f, .5354f, .5732f, .6022f, .6283f, .6546f, .6834f, .6984f, .7237f, .7486f, .7777f, .8013f, .8296f, .8565f, .8843f, .9124f, .9269f, .9428f, .9630f};

	char capitals[] = {'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z','!','?','.',',',':',';','+','-','*','/','#','\''};
	char minuscules[] = {'a','b','c','d','e','f','g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y','z','0','1','2','3','4','5','6','7','8','9','(',')',' '};

	// insert all the letter informations in the map
	for (int i = 0; i < 38; i++)
	{
		insertLetter(capitals[i], positions_capitals_tex[i], positions_capitals_tex[i+1], CapitalsTextureID);
	}

	for (int i = 0; i < 39; i++)
	{
		insertLetter(minuscules[i], positions_minuscules_tex[i], positions_minuscules_tex[i+1], MinusculesTextureID);
	}
}

void TextureTextRenderer::insertLetter(char letter, float x_begin, float x_end, GLuint texID)
{
	LetterTexInfo temp;
	temp.letter = letter;
	temp.x_begin = x_begin;
	temp.x_end = x_end;
	temp.texID = texID;

	FontTexInformations[letter] = temp;
}

void TextureTextRenderer::setLineOfText(const char* text, unsigned int text_height, bool boxed_text, unsigned int letter_spacing)
{
	std::string str = "";
	str.append(text);
	setLineOfText(str, text_height, boxed_text, letter_spacing);
}

void TextureTextRenderer::setLineOfText(std::string text, unsigned int text_height, bool boxed_text, unsigned int letter_spacing)
{
	Text = text;	
	TextHeight = text_height;
	BoxedText = boxed_text;
	LetterSpacing = letter_spacing;
}

void TextureTextRenderer::setDuration(int duration, unsigned int fade_time)
{
	TextDuration = duration;
	TextFadeDuration = fade_time;
}

void TextureTextRenderer::setPosition(unsigned int x_position, unsigned int y_position)
{
	XPosition = x_position;
	YPosition = y_position;
}

void TextureTextRenderer::setTextColor(unsigned char color_red, unsigned char color_green, unsigned char color_blue, unsigned char alpha)
{
	RedChannel = color_red;
	GreenChannal = color_green;
	BlueChannel = color_blue;
	AlphaChannel = alpha;
}


void TextureTextRenderer::activate()
{
	float x_margin = 0.0f;
	float y_margin = 0.0f;

	if (BoxedText) 
	{
		x_margin = 0.1f;
		y_margin = 0.3f;
	}

	// we set up the text display list
	glNewList(TextDisplayList, GL_COMPILE);
	glEnable(GL_TEXTURE_2D);
	glPushMatrix();
	// set the scaling
	glScalef((float)TextHeight * TextureAspect, (float) TextHeight, 0.0f);

	float width = 0.0f;

	// we translate the text respective margins
	glTranslatef(x_margin, y_margin, .0f);

	// parse the string
	for(unsigned int i=0; i < Text.size(); i++)
	{
		// for every letter we look for the texture infos
		std::map<char,LetterTexInfo>::iterator it = FontTexInformations.find(Text[i]);
		
		if(it != FontTexInformations.end())
		{
			// we've found a tex info 
			LetterTexInfo info = it->second;
			// let's put the letter into the display list
			glBindTexture(GL_TEXTURE_2D, info.texID);
			glBegin(GL_QUADS);
				glTexCoord2f(info.x_begin, 1.0f);		glVertex3f(0.0, 0.0, 0.0);
				glTexCoord2f(info.x_end, 1.0f);			glVertex3f(info.x_end - info.x_begin, 0.0, 0.0);
				glTexCoord2f(info.x_end, 0.0f);			glVertex3f(info.x_end - info.x_begin, 1.0, 0.0);
				glTexCoord2f(info.x_begin, 0.0f);		glVertex3f(0.0, 1.0, 0.0);
			glEnd();
			// translate to the next letter position
			float increment = info.x_end - info.x_begin + ((float) LetterSpacing)/(TextureAspect*TextHeight);
			width += increment;
			glTranslatef(increment, 0.0f, 0.0f);
		}
	}
	glDisable(GL_TEXTURE_2D);
	glPopMatrix();
	glEndList();

	// we build a separate display list for the box
	if (BoxedText) 
	{
		glNewList(BoxDisplayList, GL_COMPILE);
		glDisable(GL_TEXTURE_2D);
		glPushMatrix();
		// set the scaling
		glScalef((float)TextHeight * TextureAspect, (float) TextHeight, 0.0f);
		glBegin(GL_QUADS);
				glVertex3f(-0, -0, 0.0);
				glVertex3f(width + 2.0f * x_margin, 0.0f, 0.0f);
				glVertex3f(width + 2.0f * x_margin, 1.0f + 2.0f * y_margin, 0.0f);
				glVertex3f(-x_margin, 1.0f + 2.0f * y_margin, 0.0f);
		glEnd();
		glPopMatrix();
		glEndList();

	}

	StartTime = TimeFunctions::getTime();
}


void TextureTextRenderer::render()
{
	float alpha_scaling = 1.0f;
	float size_scaling = 1.0f;

	// if duration is not set to infinite we fade the text
	if (TextDuration != -1)
	{
		// get the elapsed time
		clock_t elapsed_time = TimeFunctions::getTime() - StartTime;
	
		if (elapsed_time > (long) (TextDuration + 2*TextFadeDuration))
		{
			// if time has elapsed we return without doing anything
			return;
		}
		else
		{
			// check for fade in
			if ( elapsed_time < (long) TextFadeDuration)
			{
				alpha_scaling = ((float) elapsed_time)/((float) TextFadeDuration);
				size_scaling = 1.0f/alpha_scaling;
			}
			// check for fade out
			if (elapsed_time > (long) (TextDuration + TextFadeDuration))
			{
				alpha_scaling = ((float) (TextDuration + 2 * TextFadeDuration - elapsed_time))/((float) TextFadeDuration);
				size_scaling = 1.0f/alpha_scaling;
			}		
		}
	}


		
	glDisable(GL_DEPTH_TEST);
	glPushMatrix();
	// set the position
	glTranslatef((float) XPosition, (float) YPosition, 0.0f);
	// set a scaling
	glScalef(size_scaling, size_scaling, 0.0f);
	// render box and text
	if(BoxedText)
	{
		glColor4ub(255, 255, 255, std::max(0, (int) ( (AlphaChannel - 150) * alpha_scaling)));
		glCallList(BoxDisplayList);
	}
	glColor4ub(RedChannel, GreenChannal, BlueChannel, (unsigned char) (AlphaChannel * alpha_scaling));
	glCallList(TextDisplayList);
	glEnable(GL_DEPTH_TEST);
	glPopMatrix();
}
