#include "BulldozeBrush.h"
#include "SOIL.h"

BulldozeBrush::BulldozeBrush(KinectPointer* right_hand, GridMesh* mesh)
{
	RightHandPointer = right_hand;
	BrushRadius = 4.0f;
	BrushHeight = 1.0f;
	Mesh = mesh;
	showBrush = false;
	Scale = 1.0f;
	TextureID = SOIL_load_OGL_texture("../images/bulldoze_brush.png", SOIL_LOAD_AUTO, SOIL_CREATE_NEW_ID, SOIL_FLAG_MIPMAPS);
}



BulldozeBrush::~BulldozeBrush(void)
{
}


void BulldozeBrush::updateLeftHandInput(Pose &pose)
{
	setRadiusAndHeight(pose);
	LastPose = Pose(pose);
}

void BulldozeBrush::updateRightHandInput(Pose &pose)
{
	processMesh();
	LastPose = Pose(pose);
}

void BulldozeBrush::updateBothHandInput(Pose &pose)
{
	setRadiusAndHeight(pose);
	processMesh();
	LastPose = Pose(pose);
}

void BulldozeBrush::update(Pose &pose)
{
	showBrush = (RightHandPointer->isActive())?(Mesh->ScreenToLocalCoords(RightHandPointer->getWindowCoordinates(), BrushCoordinates)):(false);
	LastPose = Pose(pose);
}

void BulldozeBrush::render()
{
	// render brush
	if (showBrush)
	{
		GLUquadric* quadric = gluNewQuadric();
		glPushMatrix();
		glTranslatef(BrushCoordinates.x, BrushHeight + 0.05f, BrushCoordinates.z);
		
		glColor3f(1.0f, 0.0f, 0.0f);
		
		// brush height line
		float trans_val = min(0.0f, -BrushHeight + BrushCoordinates.y);

		glBegin(GL_LINES);
			glVertex3f(0.0f, 0.0f, 0.0f);
			if(BrushHeight < BrushCoordinates.y)
			{
				glVertex3f(0.0f, -BrushHeight + BrushCoordinates.y + 1.0f, 0.0f);
			}
			else
			{
				glVertex3f(0.0f, trans_val, 0.0f);
			}
		glEnd();

		gluSphere(quadric,0.2,10,10);
		
		// brush radius
		glRotatef(-90, 1.0f, 0.0f, 0.0f);
		gluDisk(quadric, BrushRadius - 0.2f, BrushRadius + 0.2f, 20, 1);

		// brush shadow
		if(trans_val < -0.3f)
		{
			glColor3f(0.0f, 0.0f, 0.0f);
			glTranslatef(0.0f, 0.0f, trans_val); //<- Attention! we have to translate along the z-axis instead of the y-axis (because the disc is oriented in x-y-direction in its coordinate system)
			gluDisk(quadric, BrushRadius - 0.2f, BrushRadius + 0.2f, 20, 1);
		}

		gluDeleteQuadric(quadric);
		glPopMatrix();
	}
}

void BulldozeBrush::setRadiusAndHeight(Pose& pose)
{
	float diff_x = pose.LeftHandPosition.x - LastPose.LeftHandPosition.x;
	float diff_y = pose.LeftHandPosition.y - LastPose.LeftHandPosition.y;

	if ( abs(diff_x) > abs(diff_y))
	{
		BrushRadius += 50 * diff_x;
		BrushRadius = max(0.0f, BrushRadius);
	}
	else 
	{
		BrushHeight += 50 * diff_y;
	}
}

void BulldozeBrush::processMesh()
{
	showBrush = Mesh->ScreenToLocalCoords(RightHandPointer->getWindowCoordinates(), BrushCoordinates);
	// if brush is inside grid area
	if(showBrush)
	{
		int radius = (int) abs(BrushRadius);
		int brush_x = (int) BrushCoordinates.x;
		int brush_z = (int) BrushCoordinates.z;

	
		// iterate over all the grid points in the radius area
		for (int x = brush_x -radius; x <= brush_x + radius; x++)
		{
			if ( (x >= 0) && (x <= Mesh->getWidth()) )
			{
				for (int z = brush_z - radius; z <= brush_z + radius; z++)
				{
					if ( (z >= 0) && (z <= Mesh->getLength()) )
					{
						GridPoint* p = Mesh->getGridPoint(x, z); 
						// we check if grid point is inside the radius
						float v1 = p->coords.x - BrushCoordinates.x;
						float v2 = p->coords.z - BrushCoordinates.z;
						float dist = sqrt(v1 * v1 + v2 * v2);
						if( dist < radius )
						{
							// for every grid point we set the its height to BrushHeight if actual height is greater than BrushHeight (with a slope et the edge)
							if(p->coords.y > BrushHeight) 
							{
								p->coords.y = (1 - dist/BrushRadius) * BrushHeight + dist/BrushRadius * p->coords.y;
							}

							// put the grid point into the normal recalculation list
							Mesh->recalculateGridPointNormal(p);
						}
						
					}	
				}
			}
		}
		// recalculate the involved grid point normals
		Mesh->executeNormalsRecalculation();
	}
}