#include "Colorpicker.h"
#include <iostream>
#include "SOIL.h"

#define INTERVALSIZE (1.0472f) // = 2*PI/6
#define PI_180 (0.017453f)  //  = PI/180

// initial color value
math::Vec3f Colorpicker::CurrentColor = math::Vec3f(0.0f, 0.55f, 0.75f);

// helper function to convert from HSV parameters to RGB color
math::Vec3f HSVToRGB(float hue, float saturation, float value)
{
	float f = hue / INTERVALSIZE;
	int i = (int) f;
	f -= i;
	float p = value * (1.0f - saturation);
	float q = value * (1.0f - saturation * f);
	float t = value * (1.0f - saturation * (1 - f));

	switch(i)
	{
	case 0:
	case 6:
		return math::Vec3f(value, t, p);
	case 1:
		return math::Vec3f(q, value, p);
	case 2:
		return math::Vec3f(p, value, t);
	case 3:
		return math::Vec3f(p, q, value);
	case 4:
		return math::Vec3f(t, p, value);
	case 5:
		return math::Vec3f(value, p, q);
	default:
		return math::Vec3f(0.0f, 0.0f, 0.0f);
	}
}

// clamps to [0.0f...1.0f]
inline void clamp(float &val)
{
	if (val > 1.0f)
	{
		val = 1.0f;
	}
	else if (val < 0.0f)
    {
        val = 0.0f;
    }
}

// clamps to [0.0f...360.0f]
inline void Colorpicker::clampHue()
{
	while (Hue > 360.0f)
	{
		Hue -= 360.0f;
	}
	while (Hue < 0.0f)
    {
        Hue += 360.0f;
    }
}


Colorpicker::Colorpicker(WindowProperties* window_props)
{
	WindowProps = window_props;
	Value = 0.8f;
	Saturation = 1.0f;
	Hue = 200.0f;
	TextureID = SOIL_load_OGL_texture("../images/colorpicker.png", SOIL_LOAD_AUTO, SOIL_CREATE_NEW_ID, SOIL_FLAG_MIPMAPS);
}


Colorpicker::~Colorpicker(void)
{
}

void Colorpicker::setHueAndSaturation(Pose& pose)
{
	float diff_x = pose.RightHandPosition.x - LastPose.RightHandPosition.x;
	float diff_y = pose.RightHandPosition.y - LastPose.RightHandPosition.y;

	if ( abs(diff_x) > abs(diff_y))
	{
		diff_x *= 200;
		Hue += diff_x*abs(diff_x); 
		clampHue();
	}
	else 
	{
		Saturation += 2 * diff_y; 
		clamp(Saturation);
	}
}

void Colorpicker::setValue(Pose& pose)
{
	Value += 2 * (pose.LeftHandPosition.y - LastPose.LeftHandPosition.y); 
	clamp(Value);
}

void Colorpicker::updateLeftHandInput(Pose &pose)
{
	setValue(pose);
	update(pose);
}

void Colorpicker::updateRightHandInput(Pose &pose)
{
	setHueAndSaturation(pose);
	update(pose);
}

void Colorpicker::updateBothHandInput(Pose &pose)
{
	setHueAndSaturation(pose);
	setValue(pose);
	update(pose);
}

void Colorpicker::update(Pose &pose)
{
	LastPose = Pose(pose);
	// set the color
	CurrentColor = HSVToRGB(Hue * PI_180 , Saturation, Value);
}


void Colorpicker::render()
{
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix(); //save projection transformations
	glLoadIdentity();
	glOrtho(0.0, WindowProps->WindowWidth, 0.0, WindowProps->WindowHeight,-1.0, 1.0);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix(); //save modelview transformations
	glLoadIdentity();

	glTranslatef((float) WindowProps->WindowWidth/2, (float) WindowProps->WindowHeight/2, 0.0f);
	glScalef((float) WindowProps->WindowHeight/3, (float) WindowProps->WindowHeight/3, 1.0f);

	// render a rotated disc
	glPushMatrix();
	glRotatef(90, 0.0f, 0.0f, 1.0f);
	glRotatef(-Hue, 0.0f, 0.0f, 1.0f);
	glBegin(GL_TRIANGLE_FAN);
		glColor3f(Value, Value, Value);
		glVertex3f(0.0f, 0.0f, 0.0f);
		for(int i = 0; i <= 361; i++)
		{
			float angle = ((float) i) * PI_180;
			glColor3fv(HSVToRGB(angle, 1.0f, Value).v);
			glVertex3f(cosf(angle), sinf(angle), 0.0f);
		}
	glEnd();
	glPopMatrix();

	glTranslatef(0.0f, Saturation, 0.0f);
	glScalef(0.2f, 0.2f, 0.2f);
	// render the picker overlay
	glBegin(GL_QUADS);
		glColor3f(0.0f, 0.0f, 0.0f);
		glVertex3f(-1.05f, -1.05f, 0.0f); 
		glVertex3f(1.05f, -1.05f, 0.0f);
		glVertex3f(1.05f, 1.05f, 0.0f); 
		glVertex3f(-1.05f, 1.05f, 0.0f);
		
		glColor3fv(CurrentColor.v);
		glVertex3f(-1.0f, -1.0f, 0.0f); 
		glVertex3f(1.0f, -1.0f, 0.0f);
		glVertex3f(1.0f, 1.0f, 0.0f); 
		glVertex3f(-1.0f, 1.0f, 0.0f);
	glEnd();

	glPopMatrix(); //restore modelview transformations
	glMatrixMode(GL_PROJECTION);
	glPopMatrix(); //restore projection transformations
	glMatrixMode(GL_MODELVIEW);
	glEnable(GL_DEPTH_TEST);
}