#pragma once

#include "mat.h"
#include "diag_mat.h"

#include <limits>
#include <algorithm>

namespace cgv {
namespace math {

//helper funcition for svd
template<class T>
T pythag(const T a, const T b)
{
	T absa,absb;

	absa=fabs(a);
	absb=fabs(b);
	if (absa > absb) return (T)absa*sqrt(1+(absb*absb/(absa*absa)));
	else return (T)(absb == 0.0 ? 0.0 : absb*sqrt(1.0+(absa*absa/(absb*absb))));
}

//internal helper function to sort singular values
template<typename T>
struct compare_sv
		{
			const diag_mat<T> &mw;
			compare_sv(const diag_mat<T> &win):mw(win){}

		bool operator()(const int & _Left, const int& _Right) const
			{	
				return (mw(_Left) > mw(_Right));
			}
		};

///computes the singular value decomposition of an MxN matrix a = u* w * v^t where u is an MxN matrix, w is a diagonal
///NxN matrix and v is a NxN square matrix. If the algorithm can't achieve a convergence after 20 iteration false is returned other wise true.
///The resulting matrices are stored in the parameters u,w,v. Attention: v is returned not v^T ! So to compute the original matrix a from the 
///decomposed matrices you have to multiply u*w*transpose(v). 
///It is possible to store u directly into a to save memory, just put the same reference into a and u.
///If ordering is true the singular values are sorted in descending order.
///To ensure that u*w*v^t remains equal to the matrix a the algorithm also exchanges the columns of u and v. 
template <typename T>
bool svd(const mat<T> &a, mat<T> &u, diag_mat<T> &w,  mat<T> &v,bool ordering=true,unsigned int maxiter=20)
{
	int M = a.nrows();
	int N = a.ncols();
	diag_mat<T> W(N);
	mat<T> U=a;
	mat<T> V(N,N);
	u.resize(M,N);
	v.resize(N,N);
	w.resize(N);

	
	bool flag;
	int i,its,j,jj,k,l=0,nm=0;
	T anorm,c,f,g,h,s,scale,x,y,z;


	const T EPS=std::numeric_limits<T>::epsilon();
	vec<T> rv1(N);
	g=scale=anorm=0.0;
	for (i=0;i<N;i++) {
		l=i+2;
		rv1(i)=scale*g;
		g=s=scale=0;
		if (i < M) {
			for (k=i;k<M;k++) scale += fabs(U(k,i));
			if (fabs(scale) > EPS) {
				for (k=i;k<M;k++) {
					U(k,i) /= scale;
					s += U(k,i)*U(k,i);
				}
				f=U(i,i);
				g = -(f >= 0 ? (sqrt(s) >= 0 ? sqrt(s): -sqrt(s)) : (sqrt(s) >= 0 ? -sqrt(s) : sqrt(s)));
				h=f*g-s;
				U(i,i)=f-g;
				for (j=l-1;j<N;j++) {
					for (s=0,k=i;k<M;k++) s += U(k,i)*U(k,j);
					f=s/h;
					for (k=i;k<M;k++) U(k,j) += f*U(k,i);
				}
				for (k=i;k<M;k++) U(k,i) *= scale;
			}
		}
		W(i)=scale *g;
		g=s=scale=0.0;
		if (i+1 <= M && i+1 != N)  {
			for (k=l-1;k<N;k++) scale += fabs(U(i,k));
			if (fabs(scale) > EPS ) {
				for (k=l-1;k<N;k++) {
					U(i,k) /= scale;
					s += U(i,k)*U(i,k);
				}
				f=U(i,l-1);
				g = -(f >= 0 ? (sqrt(s) >= 0 ? sqrt(s): -sqrt(s)) : (sqrt(s) >= 0 ? -sqrt(s) : sqrt(s)));
				h=f*g-s;
				U(i,l-1)=f-g;
				for (k=l-1;k<N;k++) rv1(k)=U(i,k)/h;
				for (j=l-1;j<M;j++) {
					for (s=0.0,k=l-1;k<N;k++) s += U(j,k)*U(i,k);
					for (k=l-1;k<N;k++) U(j,k) += s*rv1(k);
				}
				for (k=l-1;k<N;k++) U(i,k) *= scale;
			}
		}
		T aaa =(fabs(W(i))+fabs(rv1(i)));
		anorm=anorm>aaa?anorm:aaa;
	}
	for (i=N-1;i>=0;i--) {
		if (i < N-1) {
			if (fabs(g) > EPS) {
				for (j=l;j<N;j++)
					V(j,i)=(U(i,j)/U(i,l))/g;
				for (j=l;j<N;j++) {
					for (s=0,k=l;k<N;k++) s += U(i,k)*V(k,j);
					for (k=l;k<N;k++) V(k,j) += s*V(k,i);
				}
			}
			for (j=l;j<N;j++) V(i,j)=V(j,i)=0.0;
		}
		V(i,i)=1.0;
		g=rv1(i);
		l=i;
	}
	int BBB = M < N ? M:N;
	for (i=BBB-1;i>=0;i--) {
		l=i+1;
		g=W(i);
		for (j=l;j<N;j++) U(i,j)=0;
		if (fabs(g) > EPS) {
			g=1/g;
			for (j=l;j<N;j++) {
				for (s=0,k=l;k<M;k++) s += U(k,i)*U(k,j);
				f=(s/U(i,i))*g;
				for (k=i;k<M;k++) U(k,j) += f*U(k,i);
			}
			for (j=i;j<M;j++) U(j,i) *= g;
		} else for (j=i;j<M;j++) U(j,i)=0;
		++U(i,i);
	}
	for (k=N-1;k>=0;k--) {
		for (its=0;its<30;its++) {
			flag=true;
			for (l=k;l>=0;l--) {
				nm=l-1;
				
				if (fabs(rv1(l)) <= EPS*anorm) {
					flag=false;
					break;
				}
				
				if (fabs(W(nm)) <= EPS*anorm) break;
			}
			if (flag) {
				c=0.0;
				s=1.0;
				for (i=l;i<k+1;i++) {

					f=s*rv1(i);
					rv1(i)=c*rv1(i);
					
					if (fabs(f) <= EPS*anorm) break;
					g=W(i);
					h=pythag(f,g);
					W(i)=h;
					h=1/h;
					c=g*h;
					s = -f*h;
					for (j=0;j<M;j++) {
						y=U(j,nm);
						z=U(j,i);
						U(j,nm)=y*c+z*s;
						U(j,i)=z*c-y*s;
					}
				}
			}
			z=W(k);
			if (l == k) {
				if (z < 0.0) {
					W(k) = -z;
					for (j=0;j<N;j++) V(j,k) = -V(j,k);
				}
				break;
			}
			if (its == maxiter) return false;
			x=W(l);
			nm=k-1;
			y=W(nm);
			g=rv1(nm);
			h=rv1(k);
			f=((y-z)*(y+z)+(g-h)*(g+h))/(2*h*y);
			T one=1;
			g=pythag(f,one);
			f=((x-z)*(x+z)+h*((y/(f+(f >= 0 ? (g >= 0 ? g : -g) : (g >= 0 ? -g : g))))-h))/x;
			c=s=1.0;
			for (j=l;j<=nm;j++) {
				i=j+1;
				g=rv1(i);
				y=W(i);
				h=s*g;
				g=c*g;
				z=pythag(f,h);
				rv1(j)=z;
				c=f/z;
				s=h/z;
				f=x*c+g*s;
				g=g*c-x*s;
				h=y*s;
				y *= c;
				for (jj=0;jj<N;jj++) {
					x=V(jj,j);
					z=V(jj,i);
					V(jj,j)=x*c+z*s;
					V(jj,i)=z*c-x*s;
				}
				z=pythag(f,h);
				W(j)=z;
				if (z) {
					z=1/z;
					c=f*z;
					s=h*z;
				}
				f=c*g+s*y;
				x=c*y-s*g;
				for (jj=0;jj<M;jj++) {
					y=U(jj,j);
					z=U(jj,i);
					U(jj,j)=y*c+z*s;
					U(jj,i)=z*c-y*s;
				}
			}
			rv1(l)=0.0;
			rv1(k)=f;
			W(k)=x;
		}
	}
	if(ordering)
	{
		//unsure if this always works correct

			
		int *order = new int[N];
		for(int i = 0; i < N; i++)
			order[i]=i;
		
		std::sort((int*)order,(int*)(order+N),compare_sv<T>(W));
		
		
	
		for(int j = 0; j < N;j++)
		{
			w(j) =W(order[j]);
			for(int i = 0; i < M; i++)
				u(i,j) = U(i,order[j]);
				
			for(int i = 0; i < N;i++)
				v(i,j) = V(i,order[j]);
		}
		delete []order;
	
	}
	else
	{
		w=W;
		u=U;
		v=V;
	}
	return true;
}





}

}
